(* allocator_code.v

   This contains the code for the two different allocator modules from
   
   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import programs.

(* ---------
   Constants
   --------- *)


Definition retreg := 0.    (* return address *)
Definition argreg := 1.    (* call argument *)
Definition workreg := 2.   (* working space *)
Definition spreg := 5.     (* stack pointer *)
Definition envreg := 6.    (* environmnet head *)


(*******************************************)
(*   Trivial allocator with no free list   *)
(*******************************************)
  
Definition initinstructions_list (init : nat)  :=
  [(i_move (d_imm 10) (s_cst 11)) ;
   (i_branch (s_imm retreg))].

Definition initinstructions (init : nat) := progfromlist (initinstructions_list init).

Definition allocinstructions (alloc:nat) :=  
  [(i_move (d_imm workreg) (s_imm retreg)) ;
   (i_move (d_imm retreg) (s_imm 10)) ;
   (i_op plus (d_imm 10) (s_imm 10) (s_imm argreg)) ;
   (i_branch (s_imm workreg))].

Definition deallocinstructions (dealloc : nat) :=
  [i_branch (s_imm retreg)].

(*******************************************)
(* Non-trivial allocator, with a free list *)
(*******************************************)

Definition ntinitinstructions (init : nat) :=
  [(i_move (d_imm 10) (s_cst 11)) ;
   (i_move (d_ind 10) (s_cst 0));
   ( i_branch (s_imm retreg))].

Definition ntdeallocinstructions (dealloc : nat) :=
  [(i_brz (s_imm argreg) (s_cst (dealloc + 8))) ;
   (i_move (d_imm 3) (s_imm argreg)) ;
   (i_op minus (d_imm 3) (s_imm 3) (s_cst 1)) ;
   (i_brz (s_imm 3) (s_cst (dealloc + 8))) ;
   (i_move (d_imm 3) (s_imm 10)) ;
   (i_move (d_imm 10) (s_imm workreg)) ;
   (i_move (d_ind 10) (s_imm argreg)) ;
   (i_move (d_indo 1 10) (s_imm 3)) ;
   (i_branch (s_imm retreg))].

Definition ntallocinstructions (alloc : nat) :=
  (* store the continuation in 2*)
  [(i_move (d_imm workreg) (s_imm retreg)) ;
  (* test if [1] == 0 *)
   (i_brz (s_imm argreg) (s_cst (alloc + 14))) ;
  (* first branch, [4] <- 10 and [0] <- [10] =*)
   (i_move (d_imm 4) (s_cst 10)) ;
   (i_move (d_imm retreg) (s_imm 10)) ;
  (* test if [0] == 0 which means that listP is empty*)
   (i_brz (s_ind retreg) (s_cst (alloc + 9))) ;
  (* main loop *)
   (i_eqtst (s_ind retreg) (s_imm argreg) (s_cst (alloc + 13))) ;
   (i_op plus (d_imm 4) (s_imm retreg) (s_cst 1)) ;
   (i_move (d_imm retreg) (s_ind 4)) ;
   (i_branch (s_cst (alloc + 4))) ;
  (* final branch [[4]] <- [1] +[0] and [3]<- [[4]] and [[3]] <- 0 *)
   (i_op plus (d_ind 4) (s_imm 1) (s_imm retreg)) ;
   (i_move (d_imm 3) (s_ind 4)) ;
   (i_move (d_ind 3) (s_cst retreg)) ;
   (i_branch (s_cst (alloc + 14))) ;
  (* branch when [1] == 0 we just do [[4]] <- [0] + 1 *) 
   (i_move (d_ind 4) (s_indo 1 retreg))	;
  (* final step, branching to the continuation *)
   (i_branch (s_imm workreg))].



