(* allocator_spec.v

   This contains the *abstract* version of the specification for allocators for

   Abstracting Allocation : The New New Thing
   Nick Benton

   (c) 2006 Microsoft Research
*)

Require Import allocator_code.
Require Import listP.
Require Import program_semantics.

(* Definition of the registers that are not modified for each program *)

Definition unchange_stack n5 n5' n6 n6' := (spreg|->{n5,n5'}) ## (envreg|->{n6,n6'}) ## unused_space.

(* Preconditions for init *)

Definition SPost_init (Ra:stateRel) (n5 n5' n6 n6': nat) := Ra ## (retreg |-> -) ##  (argreg|->-) ## (workreg|->-) ## (3|->-) ## (4|->-) ## unchange_stack n5 n5' n6 n6'.

Definition SPre_init (Ra:stateRel) (n5 n5' n6 n6' l l': nat) := (retreg |-> {l,l'}) 
##  (argreg|->-) ## (workreg|->-) ## (3|->-) ## (4|->-) ## unchange_stack n5 n5' n6 n6' ## Topfrom 10 10.

(* Preconditions for alloc *)

Definition SPost_alloc (Ra Rc:stateRel) (n n' n5  n5' n6 n6': nat) := (Ex fb, Ex fb', ((retreg |-> {fb,fb'}) ## (Block fb fb' n n'))) ##
Ra ## Rc ## (argreg |->-) ## (workreg|->-) ## (3|->-) ## (4|->-) ## unchange_stack n5 n5' n6 n6'.

Definition SPre_alloc (Ra Rc:stateRel) (n n' n5 n5' n6 n6' l l': nat) := (retreg |-> {l,l'}) ## 
Ra ## Rc ## (argreg |-> {n,n'}) ## (workreg|->-) ## (3|->-) ## (4|->-) ## unchange_stack n5 n5' n6 n6'.

(* Preconditions for dealloc *)

Definition SPost_dealloc (Ra Rc:stateRel) (n5 n5' n6 n6': nat) := (retreg|->-) ## (argreg |->-) ## 
(workreg |->-)  ## (3 |->-) ## (4 |->-) ## Ra ## Rc ## unchange_stack n5 n5' n6 n6'.

Definition SPre_dealloc (Ra Rc:stateRel) (k k' n n' n5 n5' n6 n6' l l': nat) := (retreg |-> {l,l'}) 
## (argreg |-> {n,n'}) ## (workreg |-> {k,k'}) ## Block k k' n n' ## (3 |->-) ## (4 |->-) ## Ra ## Rc 
## unchange_stack n5 n5' n6 n6'.

Definition total_judgment Spre Spost p p' label label' := 
  forall k l l' s1 s1' k', k'<=k -> Interp (Spre l l') s1 s1' p p' k' -> 
    (exists j1, exists j1', exists s2 ,exists s2',
         j1 > 0 /\ j1' >0 
      /\ kstepreduce j1 p s1 s2 label l 
      /\ kstepreduce j1' p' s1' s2' label' l' 
      /\ (min j1 j1' > k' \/ (min j1 j1' <= k' /\ Interp Spost s2 s2' p p' (min j1 j1' - k')))).

Lemma terminate_reduce : forall p s s' l l' k,
  terminates p s' l' -> kstepreduce k p s s' l l' -> terminates p s l.
Proof.
  unfold terminates; intros.
  destruct H as (k',H).
  exists (k+k').
  generalize dependent l; generalize dependent  s; induction k.
  intros; simpl in *; destruct H0; subst; assumption.
  simpl in *; intros; destruct (sem_instr (p l) s l); trivial;  destruct p0.
  apply (IHk s0 n H0).
Qed.

Lemma kstep_induction : forall j k p s l, k <= j -> kstepterm k p s l -> kstepterm j p s l.
Proof.
  induction j.
  intros; assert (k=0); try omega; subst; simpl; inversion H0.
  destruct k.
  intros ; inversion H0.
  unfold kstepterm; simpl; fold kstepterm; intros.
  destruct (sem_instr (p l) s l); trivial.
  destruct p0; simpl.
  eapply IHj; try eassumption.
  omega.
Qed.

Lemma stepterm_reduce : forall k1 k k2 p s s' l l',
  k1 + k2 = k -> kstepreduce k1 p s s' l l' -> 
  ( kstepterm k2 p s' l' <-> kstepterm k p s l).
Proof.
  induction k1.
  simpl; intuition; destruct H1; subst;
  eapply kstep_induction; try eassumption; omega.
  intuition.
  destruct k.
  assert (k2 = 0) ; try omega.
  subst ; inversion H1.
  unfold kstepterm; fold kstepterm;
  unfold kstepreduce in H0; fold kstepreduce in H0;
  destruct (sem_instr (p l) s l); trivial.
  destruct p0; simpl;
  specialize (IHk1 k k2 p s0 s' n l').
  destruct IHk1 as (Hs0,Hs'); try assumption.
  omega.
  apply (Hs0 H1).
  destruct k.
  assert (k2 = 0) ; try omega.
  subst ; inversion H1.
  unfold kstepterm in H1; fold kstepterm in H1;
  unfold kstepreduce in H0; fold kstepreduce in H0.
  destruct (sem_instr (p l) s l); try inversion H0.
  destruct p0; simpl;
  specialize (IHk1 k k2 p s0 s' n l').
  destruct IHk1 as (Hs0,Hs'); try assumption.
  omega.
  apply (Hs' H1).  
Qed.

Lemma moveimm : forall pf pf' p p' start start' m Rpre Rpost P src src' Q,
  pf start = Some(i_move (d_imm m) src) ->
  pf' start' = Some(i_move (d_imm m) src') ->
  (forall s s' p p' k ll ll',  Interp ((m|=>P) ## Rpre ll ll') s s' p p'  k ->  
    (exists l,exists l', natrel Q p p' l l' k /\  sem_src src s = l /\  sem_src src' s' = l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  total_judgment (fun l => fun l' => (m|=> Q) ## Rpre l l') Rpost p p' (start+1) (start'+1) -> 
  total_judgment (fun l => fun l' => (m|=> P) ## Rpre l l') Rpost p p' start start'.
Proof.
  intros until Q;  rewrite_start_succ start start'.
  intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog' Hnext.
  unfold total_judgment in *.
  intros until k'; intros Hk' Hinterp.
  hreplace Hinstr2 with (Hinstr2 s1 s1' p p' k' l l' Hinterp).
  destruct Hinstr2 as (ll & ll' & HQ & Hsem & Hsem').
  pose (next_state := update s1 m (sem_src src s1)).
  pose (next_state' := update s1' m (sem_src src' s1')).
  assert (Hkk : k'-1 <= k-1); try omega.
  specialize (Hnext (k-1) l l' next_state next_state' (k'-1) Hkk).
  destruct Hnext as (j1 & j1' & s2 & s2' & Hj1 & Hj1' & Hstep & Hstep' & Hnext).
  unfold Interp in *; eapply stateRel_cond;
  [idtac | apply subPfunction_refl | apply subPfunction_refl | idtac].
  unfold next_state, next_state'; apply indepupdate ; try eassumption.
  subst; eapply natRel_cond; try eassumption; try omega; try trivial.
  rewrite <- (ptsto_blank_leq m m P).
  assumption.
  omega.
  exists (S j1); exists (S j1'); exists s2; exists s2'.
  (* exists j1; exists j1'; exists s2; exists s2'. *)
  destruct Hnext.
  intuition.
  assert (Hsubprog_start := (Hsubprog start));
  rewrite Hinstr1 in Hsubprog_start.
  simpl. rewrite Hsubprog_start.
  simpl.
  assumption.
  assert (Hsubprog_start := (Hsubprog' start'));
  rewrite Hinstr1' in Hsubprog_start.
  simpl. rewrite Hsubprog_start.
  simpl.
  assumption.
  left. 
  simpl.
  omega.
  intuition.
  assert (Hsubprog_start := (Hsubprog start));
  rewrite Hinstr1 in Hsubprog_start.
  simpl. rewrite Hsubprog_start.
  simpl.

  assumption.
  assert (Hsubprog_start := (Hsubprog' start'));
  rewrite Hinstr1' in Hsubprog_start.
  simpl. rewrite Hsubprog_start.
  simpl.
  assumption.
  right.
  assert (Hj1j1':= minGt j1 j1' 0 Hj1 Hj1').
  split; simpl.
  omega.
  assert (Hk : k'>0).
  omega.
  destruct k'.
  inversion Hk.
  replace (min j1 j1' - (S k' - 1)) with (min j1 j1' - k') in H1 by omega.
  assumption.
Qed.



(* Definition allocator_spec init alloc dealloc:= *)
(*  fun p =>  *)
(*    exists Ra, forall n Rc n5 n6 n6' base size, *)
(*       total_judgment (SPre_init Ra n5 n6 n6')  *)
(*             (SPost_init Ra n5 n6 n6') p init /\ *)
(*       total_judgment (SPre_alloc Ra Rc n n5 n6 n6')  *)
(*             (SPost_alloc Ra Rc n n5 n6 n6') p alloc /\  *)
(*       total_judgment (SPre_dealloc Ra Rc base size n5 n6 n6') *)
(*             (SPost_dealloc Ra Rc n5 n6 n6') p dealloc. *)

Definition Ra_simple := Ex x, ((10 |-> x) ## Topfrom x x).

Definition Ra_nt := (Ex xs, (Ex m, ((10 |-> m) ## listP xs m))).
