(* allocator_spec.v

   This contains the *abstract* version of the specification for allocators for
   
   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import allocator_code.
Require Import listP.
Require Import program_semantics.

(* Definition of the registers that are not modified for each program *)

Definition unchange_stack n5 n5' n6 n6' := (spreg|->{n5,n5'}) ## (envreg|->{n6,n6'}) ## unused_space.

(* Preconditions for init *)

Definition SPost_init (Ra:stateRel) (n5 n5' n6 n6': nat) := Ra ## (retreg |-> -) ##  (argreg|->-) ## (workreg|->-) ## (3|->-) ## (4|->-) ## unchange_stack n5 n5' n6 n6'.

Definition SPre_init (Ra:stateRel) (n5 n5' n6 n6' l l': nat) := (retreg |-> {l,l'}) 
##  (argreg|->-) ## (workreg|->-) ## (3|->-) ## (4|->-) ## unchange_stack n5 n5' n6 n6' ## Topfrom 10 10.

(* Preconditions for alloc *)

Definition SPost_alloc (Ra Rc:stateRel) (n n' n5  n5' n6 n6': nat) := (Ex fb, Ex fb', ((retreg |-> {fb,fb'}) ## (Block fb fb' n n'))) ##
Ra ## Rc ## (argreg |->-) ## (workreg|->-) ## (3|->-) ## (4|->-) ## unchange_stack n5 n5' n6 n6'.

Definition SPre_alloc (Ra Rc:stateRel) (n n' n5 n5' n6 n6' l l': nat) := (retreg |-> {l,l'}) ## 
Ra ## Rc ## (argreg |-> {n,n'}) ## (workreg|->-) ## (3|->-) ## (4|->-) ## unchange_stack n5 n5' n6 n6'.

(* Preconditions for dealloc *)

Definition SPost_dealloc (Ra Rc:stateRel) (n5 n5' n6 n6': nat) := (retreg|->-) ## (argreg |->-) ## 
(workreg |->-)  ## (3 |->-) ## (4 |->-) ## Ra ## Rc ## unchange_stack n5 n5' n6 n6'.

Definition SPre_dealloc (Ra Rc:stateRel) (k k' n n' n5 n5' n6 n6' l l': nat) := (retreg |-> {l,l'}) 
## (argreg |-> {n,n'}) ## (workreg |-> {k,k'}) ## Block k k' n n' ## (3 |->-) ## (4 |->-) ## Ra ## Rc 
## unchange_stack n5 n5' n6 n6'.


Definition AllocSpec Ra init init' alloc alloc' dealloc dealloc':=
 fun (p p' : program) => 
   (forall n5 n5' n6 n6',  total_judgment (SPre_init Ra n5 n5' n6 n6') 
     (SPost_init Ra n5 n5' n6 n6') p p' init init') /\
   (forall Rc_alloc size size' n5 n5' n6 n6',  total_judgment (SPre_alloc Ra Rc_alloc size size' n5 n5' n6 n6') 
     (SPost_alloc Ra Rc_alloc size size' n5 n5' n6 n6') p p' alloc alloc') /\
   (forall Rc_alloc size size' base base' n5 n5' n6 n6', total_judgment (SPre_dealloc Ra Rc_alloc base base' size size' n5 n5' n6 n6')
     (SPost_dealloc Ra Rc_alloc n5 n5' n6 n6') p p' dealloc dealloc').

Definition Ra_simple := Ex x, ((10 |-> x) ## Topfrom x x).

Definition Ra_nt := (Ex xs, (Ex m, ((10 |-> m) ## listP xs m))).

