(* correction_app.v

   This is the partial correctness proof for application for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import correction_compiler_definition.

Typeclasses eauto := dfs 10000.

Lemma correction_of_app : forall base base' Ra alloc alloc' dealloc dealloc' env a b e1 e2 (t1 : env|-- e1 ::: a--->b) (t2:env|--e2:::a), 
  (forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env (a--->b) e1 t1 Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
  (forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env a e2 t2 Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
  forall Rc Rc_cloud start  start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env b (APP e1 e2) (TAPP t1 t2) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3'.
Proof with normal_form.
  intros until t2; intros HypRecE1 HypRecE2; intros. 
  unfold statement_of_compiler_correctness in *.
  unfold prog_extends_code, prog_extends_auxcode, extract_code_from_globalcode, extract_stacksize_from_code in *;
  intros until global_code'; intros hprog hcode Hstack hprog' hcode' Hstack' Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc.
  split ; [idtac | intros e' Hlambda; inversion Hlambda]; intro hyp.
  
  (* recover the actual piece of code *)

  simpl in hprog; surjective_pairing_two_arg e1 e2 aux_code base alloc dealloc in hprog;
  rewrite <- hcode in Hsubprog; rewrite <- hcode' in Hsubprog';        
  rewrite hprog in hcode; simpl in hcode;
  surjective_pairing_two_arg e1 e2 code1_gen code2_gen start in hcode.
  simpl in hprog'; surjective_pairing_two_arg e1 e2 aux_code' base' alloc' dealloc' in hprog';
  rewrite hprog' in hcode'; simpl in hcode';
  rename code2_gen0 into code2_gen'; rename code1_gen0 into code1_gen';
  surjective_pairing_two_arg e1 e2 code1_gen' code2_gen' start' in hcode'.

  unfold pair_code in hcode; simpl in hcode.
  rewrite hprog in Hsubprog_aux; simpl in Hsubprog_aux; clear hprog.
  rewrite hcode in Hstack; simpl in Hstack.
  unfold pair_code in hcode'; simpl in hcode'.
  rewrite hprog' in Hsubprog_aux'; simpl in Hsubprog_aux'; clear hprog'.
  rewrite hcode' in Hstack'; simpl in Hstack'.
  rename code0 into code2'; rename code3 into code1'.
  replace (snd(fst code2)+0) with (snd(fst code2)) in *; try omega.
  replace (snd(fst code2')+0) with (snd(fst code2')) in *; try omega.
  assert (Hmin : min (snd(fst code2)) 1 = 1); try (apply min_r;  unfold code2,code2_gen; apply stack_size_positive).
  rewrite Hmin in Hstack.
  assert (max (snd (fst code2)) (snd(fst code1) +1) <= stack_free); [eapply le_trans ; try apply Hstack; apply le_max_l | idtac].
  clear Hmin Hstack; rename H into Hstack.
  assert (Hmin' : min (snd(fst code2')) 1 = 1); try (apply min_r;  unfold code2',code2_gen'; apply stack_size_positive).
  rewrite Hmin' in Hstack'.
  assert (max (snd (fst code2')) (snd(fst code1') +1) <= stack_free'); [eapply le_trans ; try apply Hstack'; apply le_max_l | idtac].
  clear Hmin' Hstack'; rename H into Hstack'.

  (* apply each argument *)

  apply_arg_e2 code2_gen code1_gen code2 code2_gen' code1_gen' code2' hcode hcode' stack_free stack_free' p p' start base start' base' Rc Rc_cloud stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' stack_list  Halloc Hdealloc Hsubprog Hsubprog_aux Hsubprog' Hsubprog_aux' HypRecE2 Hstack Hstack'.

  intros ptr_a ptr_a'.

  apply_arg_e1 code2_gen code1_gen code2 code1 code2_gen' code1_gen'
  code2' code1' hcode hcode' stack_free stack_free' p p' start base
  start' base' Ra Rc Rc_cloud stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code
  aux_code' stack_list Halloc Hdealloc Hsubprog
  Hsubprog_aux Hsubprog' Hsubprog_aux' a ptr_a ptr_a' HypRecE1.

  match goal with | h : code = (?l1++?l,?stack,?next) |- _ => set (code_app := l) end; rewrite hcode in Hsubprog.
  match goal with | h : code' = (?l1++?l,?stack,?next) |- _ => set (code_app' := l) end; rewrite hcode' in Hsubprog'.

  assert (HsubprogApp : PfunctionExtendsFunction p (fragfromprog (progfromlist (code_app)) (start+length (fst (fst code2))+length (fst (fst code1))) )).
  eapply subPfunction_extends; try apply Hsubprog.
  rewrite <- sub_is_extends; simpl;rewrite <- plus_assoc; rewrite <- app_length;  apply suffix_code.
  unfold code_app in HsubprogApp ; clear Hsubprog code_app.
  assert (HsubprogApp' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (code_app')) (start'+length (fst (fst code2'))+length (fst (fst code1'))) )).
  eapply subPfunction_extends; try apply Hsubprog'.
  rewrite <- sub_is_extends; simpl;rewrite <- plus_assoc; rewrite <- app_length;  apply suffix_code.
  unfold code_app' in HsubprogApp' ; clear Hsubprog' code_app'.

  unfold memory_specification; simpl.
  intros fcode fcode'; unfold code2, code2_gen, code2', code2_gen'; repeat rewrite compile_length_main; fold code2_gen code2 code2_gen' code2'.
  set (new_start := start+length (fst (fst code2))+length (fst (fst code1))) in *.
  set (new_start' := start'+length (fst (fst code2'))+length (fst (fst code1'))) in *.

  simpl; unfold storing_space...
  (* [[env]+4] <- label + 3  save application return label *)
  focus 5;focus 3.
  unfold application_code in *.
  
  instr_moveindo new_start new_start' 0 HsubprogApp HsubprogApp' with nat (snd code1 +3) (snd code1' +3) on values envreg envreg n n'.
  (* [wk] <- [[sp]] now wkreg points to the function code*)
  focus 6;focus 3;focus 9.  
  instr_moveimmind new_start new_start' 1 HsubprogApp HsubprogApp' with nat fcode fcode' on values spreg spreg (stack_ptr+1+1) (stack_ptr'+1+1).
  focus 9; focus_conj 1.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack; rewrite <- existsJudgment; intro Pprivate ...
  normal_form_conj.
  omega_replace (stack_ptr +1+1-1-1) with stack_ptr.
  omega_replace (stack_ptr' +1+1-1-1) with stack_ptr'.
  (* place the program in the predicate not to "lose" it after the jump*)
  rewrite judgment_internalization_ptsto_stack.
  rewrite judgment_internalization.
  normal_form;  focus 1.
  (*  jmp [[wk]]              jump to function code *)  
  instr_jumpind new_start new_start' 2 HsubprogApp HsubprogApp'.
  apply continuation_simplification_stack.
  (* hard work to prove be able to apply hyp on the continuation*)
  rewrite ptsto_blank_leq.
  focus 1; rewrite cloudRelConj_comm; rewrite cloudRelConj_weakening.
  focus 17; focus 6.
  do 2 rewrite stateRelTensor_assoc; hide_tail.   
  do 2 rewrite <- stateRelTensor_assoc.
  rename Hstack into Hstack_free; rename Hstack' into Hstack_free'.
  explicit_stack stack_list stack_ptr stack_ptr'.
  focus 2.
  rewrite (Hstack (Pprivate //\\ ([|a|]) Ra ptr_a ptr_a' //\\ ([||env||]) Ra n n' //\\ Rc_cloud)).
  unfold Pre_arrow.
  omega_replace (stack_ptr +1 +1 -1) with (stack_ptr +1);  omega_replace (stack_ptr +1 +1 +1) with (stack_ptr +3); omega_replace (stack_ptr  +1 +1 ) with (stack_ptr  +2).
  omega_replace (stack_ptr' +1 +1 -1) with (stack_ptr' +1);  omega_replace (stack_ptr' +1 +1 +1) with (stack_ptr' +3); omega_replace (stack_ptr'  +1 +1 ) with (stack_ptr'  +2). 
  normal_form.
  set (new_Rc := ({n+4,n'+4}|->-) ## ({n+2,n'+2}|->{n2,n2'}) ## ({n+3,n'+3}|->{n3,n3'}) ##Pstack##Block (stack_ptr+3)(stack_ptr'+3)(stack_free-1-1)(stack_free'-1-1) ## Rc).
  apply stateExistsLeq; exists new_Rc.
  set (new_Rc_cloud := Qstack//\\ [||env||] Ra n n' //\\ Rc_cloud).
  apply stateExistsLeq; exists new_Rc_cloud.
  apply stateExistsLeq; exists n.
  apply stateExistsLeq; exists n'.
  apply stateExistsLeq; exists ptr_a.
  apply stateExistsLeq; exists ptr_a'.
  apply stateExistsLeq; exists (stack_ptr+1).
  apply stateExistsLeq; exists (stack_ptr'+1).
  omega_replace (stack_ptr +1 +1 -1) with (stack_ptr +1);  omega_replace (stack_ptr +1 +1 +1) with (stack_ptr +3); omega_replace (stack_ptr +1 +1 ) with (stack_ptr +2).
  set (Postn4 := ({n+4,n'+4}||=>Later
    (Perp (Post_arrow ([|b|]) Ra new_Rc new_Rc_cloud n n' (stack_ptr+1)(stack_ptr'+1))))). 
  omega_replace (stack_ptr' +1 +1 -1) with (stack_ptr' +1);  omega_replace (stack_ptr' +1 +1 +1) with (stack_ptr' +3); omega_replace (stack_ptr' +1 +1 ) with (stack_ptr' +2).
  normal_form.
  focus' 3;  unfold new_Rc, storing_space.
  rewrite <- linearDistributivity.
  set (smallRc := ({n+2,n'+2}|->{n2,n2'}) ## ({n+3,n'+3}|->{n3,n3'}) ##Pstack##Block (stack_ptr+3)(stack_ptr'+3)(stack_free-1-1)(stack_free-1-1) ## Rc).
  normal_form.
  isolate_conj'.
  rewrite stateCloudPtsot in Block; unfold Postn4 in Block; rewrite <- ptsto_blank_elim in Block.
  rewrite cloudStatePtsot in Block.
  rebuilt_conj Conj_Block Block.
  clear Postn4;   
    set (Postn4 := ({n + 4, n' + 4}|=> Later(Perp(Post_arrow ([|b|]) Ra new_Rc new_Rc_cloud n n' (stack_ptr+1)(stack_ptr'+1))))).
  focus 2 ; focus 3; focus' 8; focus' 1; focus' 18.
  hide_tail.
  do 2 rewrite stateRelTensor_assoc.
  unfold tail0; clear tail0;   do 2 rewrite stateRelTensor_assoc.
  unfold new_Rc, tail,smallRc; clear tail.
  omega_replace (stack_ptr+1+1) with (stack_ptr+2); omega_replace (stack_ptr+2-1) with (stack_ptr+1); omega_replace (stack_ptr+2+1) with (stack_ptr+3).
  omega_replace (stack_ptr'+1+1) with (stack_ptr'+2); omega_replace (stack_ptr'+2-1) with (stack_ptr'+1); omega_replace (stack_ptr'+2+1) with (stack_ptr'+3) .
  eapply ppredtensorcompat2; try ineq_ring.
  normal_form; unfold new_Rc_cloud.
  focus' 2;  isolate_conj'.
  focus_conj_hyp 2.
  rebuilt_conj Conj_Block Block.
  focus 1; focus' 1;focus' 2.
  unfold Postn4;  apply continuation_deduction.
  unfold Post_arrow; unfold memory_specification in hyp.
  clear - hyp hcode hcode' Hstack Hstack_free Hstack_free'.
  rewrite <- existsJudgment; intro ptr_b.
  rewrite <- existsJudgment; intro ptr_b'.
  hreplace hyp with (hyp ptr_b ptr_b').
  unfold code1, code1_gen, code2, code2_gen, code1', code1_gen', code2', code2_gen'.
  rewrite compile_length_aux; rewrite compile_length_main;  rewrite compile_length_main;  rewrite <- compile_length_aux.
  fold code2_gen;  unfold code2_gen at 4;  rewrite <- compile_length_main;  fold code2_gen code1_gen code2 code1.
  rewrite compile_length_aux; rewrite compile_length_main;  rewrite compile_length_main;  rewrite <- compile_length_aux.
  fold code2_gen';  unfold code2_gen' at 4;  rewrite <- compile_length_main;  fold code2_gen' code1_gen' code2' code1'.
  rewrite hcode in hyp;rewrite hcode' in hyp; simpl in hyp; do 4 rewrite app_length in hyp; simpl in hyp.
  replace (start+(length (fst (fst code2))+length (fst (fst code1)) +3)) with (start+length (fst (fst code2))+length (fst (fst code1)) +3 ) in hyp; try omega.  
  set (new_start := start+length (fst (fst code2))+length (fst (fst code1)) +3 ) in *.
  replace (start'+(length (fst (fst code2'))+length (fst (fst code1')) +3)) with (start'+length (fst (fst code2'))+length (fst (fst code1')) +3 ) in hyp; try omega.  
  set (new_start' := start'+length (fst (fst code2'))+length (fst (fst code1')) +3 ) in *.
  replace (stack_ptr +1 -1) with (stack_ptr) in hyp ;try omega.
  replace (stack_ptr +1 +1 ) with (stack_ptr+2) in hyp ; try omega.
  replace (stack_ptr' +1 -1) with (stack_ptr') in hyp ;try omega.
  replace (stack_ptr' +1 +1 ) with (stack_ptr'+2) in hyp ; try omega.
  unfold new_Rc, new_Rc_cloud; clear new_Rc new_Rc_cloud...
  focus 11;focus 2.
  omega_replace (stack_ptr+3) with (S(stack_ptr+2));  omega_replace (stack_ptr+1+1) with (stack_ptr+2).
  omega_replace (stack_ptr'+3) with (S(stack_ptr'+2));  omega_replace (stack_ptr'+1+1) with (stack_ptr'+2). 
  rewrite stateRelTensor_assoc; rewrite <- block_unfold_start.
  unfold ge in Hstack_free; assert (2<=stack_free).
  eapply le_trans ; try apply Hstack_free;  apply le_trans with (m := snd(fst code1)+1) ; try apply le_max_r.
  assert (snd(fst code1) >= 1); [ unfold code1, code1_gen, code2_gen ;  rewrite compile_length_aux;
  apply stack_size_positive | omega].
  unfold ge in Hstack_free'; assert (2<=stack_free').
  eapply le_trans ; try apply Hstack_free';  apply le_trans with (m := snd(fst code1')+1) ; try apply le_max_r.
  assert (snd(fst code1') >= 1); [ unfold code1', code1_gen', code2_gen' ;  rewrite compile_length_aux;
  apply stack_size_positive | omega].
  omega_replace (S(stack_free-1-1)) with (stack_free-1).
  omega_replace (S(stack_free'-1-1)) with (stack_free'-1).
  focus 10;focus 3; focus_conj 1.
  rewrite stateRelTensor_assoc; rewrite <- (Hstack (([|b|]) Ra ptr_b ptr_b' //\\ ([||env||]) Ra n n' //\\ Rc_cloud)).
  eapply judgment_leq; try apply hyp.
  unfold storing_space; ineq_ring.
Qed.
