(* correction_basic_ops.v

   This is the partial correctness proof for basic types (Int, Bool and Var) for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)


Require Import correction_compiler_definition.

Typeclasses eauto := dfs 10000.

Lemma correction_of_int : forall base base' Ra alloc alloc' dealloc dealloc' env (P:nat-> Prop) n (Pn : P n) Rc Rc_cloud
  start start' stack_free stack_free' stack_ptr stack_ptr' ptr_env n2 n3 ptr_env' n2' n3',
  statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env (Int P) (INT n) (TINT env P n Pn) Rc Rc_cloud start start' stack_free stack_free' stack_ptr  stack_ptr' ptr_env n2 n3 ptr_env' n2' n3'.
Proof with normal_form.

  unfold statement_of_compiler_correctness; clean_judgment_basic.

  instr_opimmleft plus start start' 0 Hsubprog Hsubprog' (stack_ptr) (stack_ptr') 1 1.

  assert (Hstack : stack_free = S(stack_free-1)); try omega.
  assert (Hstack' : stack_free' = S(stack_free'-1)); try omega.

  rewrite Hstack; rewrite Hstack'; rewrite block_unfold_start...
  instr_moveind start start' 1 Hsubprog Hsubprog' with nat n n.
    
  unfold memory_specification in hyp; hreplace hyp with (hyp n n).
  simpl in hyp; normal_form in hyp.
  eapply judgment_leq; try apply hyp.
  omega_replace (S(stack_ptr+1)) with (stack_ptr+1+1).
  omega_replace (stack_ptr+1-1) with stack_ptr.
  omega_replace (S(stack_ptr'+1)) with (stack_ptr'+1+1).
  omega_replace (stack_ptr'+1-1) with stack_ptr'.
  focus 5;focus' 5.
  eapply ppredtensorcompat2.
  setoid_replace (P n) with True using relation iff; try intuition.
  setoid_replace (n=n) with True using relation iff; try intuition.
  setoid_replace (True /\ True /\ True) with True using relation iff; [idtac | intuition].
  fold cloudPtrue; repeat rewrite cloudRelConj_unit; reflexivity.
  ineq_ring.
Qed.

Lemma correction_of_bool : forall base base' Pa alloc alloc' dealloc dealloc' env (P:bool-> Prop) b (Pb : P b)
  Pc Pc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' ptr_env n2 n3 ptr_env' n2' n3',
  statement_of_compiler_correctness base base' Pa alloc alloc' dealloc dealloc' env (Bool P) (BOOL b) (TBOOL env P b Pb) Pc Pc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' ptr_env n2 n3  ptr_env' n2' n3'.
Proof with normal_form.
  unfold statement_of_compiler_correctness; clean_judgment_basic.
  instr_opimmleft plus start start' 0 Hsubprog Hsubprog' stack_ptr stack_ptr' 1 1.

  assert (Hstack : stack_free = S(stack_free-1)); try omega.
  assert (Hstack' : stack_free' = S(stack_free'-1)); try omega.
  rewrite Hstack; rewrite Hstack'; rewrite block_unfold_start...

  let not_to_do_twice b :=  instr_moveind start start' 1 Hsubprog Hsubprog' with nat b b;
  unfold memory_specification in hyp ; hreplace hyp with (hyp b b);
  simpl in hyp; normal_form in hyp;
  eapply judgment_leq; try apply hyp;
  omega_replace (S(stack_ptr+1)) with (stack_ptr+1+1);
  omega_replace (stack_ptr+1-1) with stack_ptr;
  omega_replace (S(stack_ptr'+1)) with (stack_ptr'+1+1);
  omega_replace (stack_ptr'+1-1) with stack_ptr';
  focus 5;focus' 5; eapply ppredtensorcompat2; [
  setoid_replace (P (bool_of_nat b)) with True using relation iff; try intuition;
  setoid_replace (bool_of_nat b= bool_of_nat b) with True using relation iff; try intuition;
  setoid_replace (True /\ True /\ True) with True using relation iff; [idtac | intuition];
  fold cloudPtrue; repeat rewrite cloudRelConj_unit; reflexivity |
  ineq_ring ]
  in induction b ; [not_to_do_twice 1 | not_to_do_twice 0].
Qed.


Lemma correction_of_var :  forall base base' Pa alloc alloc' dealloc dealloc' m a env (h: nth_error env m = Some a) 
  Pc Pc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' ptr_env n2 n3 ptr_env' n2' n3',  
  statement_of_compiler_correctness base base' Pa alloc alloc' dealloc dealloc' env a (VAR m) (TVAR env m h) Pc Pc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' ptr_env n2 n3 ptr_env' n2' n3'.
Proof with normal_form.

  unfold statement_of_compiler_correctness; clean_judgment_basic.
  rewrite length_list in hyp.
  match goal with | h: PfunctionExtendsFunction ?p (fragfromprog (progfromlist ?l) ?start) |- _ =>  set (ugly_list := l) in * end.
  assert (ugly_list = (fuse (takedrop 2 nil ugly_list))); try (simpl;trivial).
  Opaque fuse. simpl in H. Transparent fuse.
  unfold fuse in H; rewrite H in Hsubprog; rewrite H in Hsubprog'; clear H ugly_list.
  rewrite hsize in Hstack_free.
  rewrite hsize' in Hstack_free'.
  instr_opimmleft plus start start' 0 Hsubprog Hsubprog' stack_ptr stack_ptr' 1 1.

  focus 2; focus 5.
  instr_moveimm start start' 1 Hsubprog Hsubprog' with nat ptr_env ptr_env'.
  
(* preparing induction hypothesis *)

  set (H := dismember_program _ _ Hsubprog); destruct H as (Hsubprog_done,Hsubprog_loop); clear Hsubprog_done.
  simpl in Hsubprog_loop; set (new_start:=start+2) in *.
  clear Hsubprog.
  set (H := dismember_program _ _ Hsubprog'); destruct H as (Hsubprog'_done,Hsubprog'_loop); clear Hsubprog'_done.
  simpl in Hsubprog'_loop; set (new_start':=start'+2) in *.
  clear Hsubprog'.
  match goal with | H:PfunctionExtendsFunction ?p (fragfromprog (progfromlist (?l)) ?start) |- _ => set (code_loop := l) in * end.
  assert (forall headenv tailenv, forall start start' x0 x0' p p' n n' stack_list stack_free stack_free', 
     env = headenv ++ tailenv -> nth_error tailenv m = Some a ->
     stack_free >= 1 -> stack_free' >= 1 -> 
     PfunctionExtendsFunction p (fragfromprog (progfromlist code_loop) start) ->
     PfunctionExtendsFunction p' (fragfromprog (progfromlist code_loop) start') ->
     (forall ptr_a ptr_a', 
       |= start+m+1,start'+m+1 : 
         memory_specification Pa Pc Pc_cloud env (([|a|] Pa, (ptr_a,ptr_a')) :: stack_list) (stack_free-1) (stack_free'-1) (stack_ptr+1) (stack_ptr'+1) n n2 n3 n' n2' n3';
     p,p') ->
     |= start,start' : 
       (spreg |-> {stack_ptr+1,stack_ptr'+1}) ## (envreg|-> {n,n'}) ## Block (stack_ptr+1) (stack_ptr'+1) stack_free stack_free' ## (workreg |-> {x0,x0'}) ## stack_description stack_list stack_ptr stack_ptr' ([||tailenv||] Pa x0 x0' //\\ unfold_env headenv n n' x0 x0' Pa //\\ Pc_cloud) ## Pa ## Pc ## storing_space n n2 n3 n' n2' n3' ## (argreg|->-)##(retreg|->-) ## (3|->-)##(4|->-) ## unused_space; p,p').
  clear h hyp Hsubprog_loop Hsubprog'_loop new_start new_start';  generalize dependent m.    
  induction m.

(*  case  m =0 *)
  clear; intros until stack_free'; intros enveq h Hstack Hstack' Hsubprog Hsubprog' hyp.
  simpl in code_loop.
  unfold memory_specification in hyp; rewrite nth_error_0 in h.
  destruct h as [env' h];  rewrite h; simpl.
  unfold code_loop in Hsubprog; unfold code_loop in Hsubprog'.
  focus 4. 
  repeat (rewrite <- cloudExP_conj; trivial;  rewrite <- stateExP_stack;
  rewrite <- existsJudgment; intro).
  normal_form_conj.
  focus 3;  assert (exists m , S m = stack_free).
  apply strictPosSucc; omega.
  clear Hstack; destruct H as [stack_free0 Hstack].
  assert (exists m , S m = stack_free').
  apply strictPosSucc; omega. 
  clear Hstack'; destruct H as [stack_free0' Hstack'].
  rewrite <- Hstack; rewrite <- Hstack'; rewrite block_unfold_start...
  focus 2;focus 5;focus 2;focus 4.
  instr_moveind start start' 0 Hsubprog Hsubprog' with nat x x1.

  replace (start +0 +1) with (start +1) in hyp ; try omega.
  replace (start' +0 +1) with (start' +1) in hyp ; try omega.
  hreplace hyp with (hyp x x1).
  rewrite enveq in hyp; simpl in hyp; normal_form in hyp.
  focus_hyp 5.
  rewrite <- (unfold_env_semantic headenv tailenv Pa n n' x0 x0') in hyp.
  rewrite h in hyp; simpl in hyp.
  normal_form_conj in hyp.
  focus_conj_hyp 2.
  setoid_rewrite <- cloudExP_conj in hyp; rewrite <- stateExP_stack in hyp.
  rewrite <- existsJudgment in hyp; hreplace hyp with (hyp x).
  setoid_rewrite <- cloudExP_conj in hyp; rewrite <- stateExP_stack in hyp.
  rewrite <- existsJudgment in hyp; hreplace hyp with (hyp x1).
  setoid_rewrite <- cloudExP_conj in hyp; rewrite <- stateExP_stack in hyp.
  rewrite <- existsJudgment in hyp; hreplace hyp with (hyp x2).
  setoid_rewrite <- cloudExP_conj in hyp; rewrite <- stateExP_stack in hyp.
  rewrite <- existsJudgment in hyp; hreplace hyp with (hyp x3).
  normal_form_conj in hyp.
  focus_conj_hyp 2; focus_conj_hyp 4.
  rewrite cloudRelConj_assoc in hyp.
  rewrite <- (cloudConjContraction ([|a|] Pa x x1)([|a|] Pa x x1)) in hyp; try ineq_ring_conj.
  focus_conj_hyp 4.
  omega_replace stack_free0 with (stack_free-1).
  omega_replace stack_free0' with (stack_free'-1).
  omega_replace (S (stack_ptr+1)) with (stack_ptr+1+1).
  omega_replace (S (stack_ptr'+1)) with (stack_ptr'+1+1).
  rewrite (ptsto_blank_leq workreg).
  eapply judgment_leq; try apply hyp.
  omega_replace (stack_ptr+1-1) with stack_ptr.
  omega_replace (stack_ptr'+1-1) with stack_ptr'.
  focus 3;  eapply ppredtensorcompat2; try (apply stackdescriptioncompat2;trivial;ineq_ring_conj).
  focus 1; focus' 4; focus 2; focus' 5; focus 6; focus 6; focus' 9 ; focus' 9.
  reflexivity.
  
(* induction case*)

  clear -IHm; intros until stack_free'; intros enveq h Hstack Hstack' Hsubprog Hsubprog' hyp.
  replace (start + S m +1 ) with (start + 2 +m) in hyp ; try omega.
  replace (start' + S m +1 ) with (start' + 2 +m) in hyp ; try omega.
  rewrite nth_error_Sn in h;  destruct h as [env' [a0 (h1,h2)]].
  rewrite h1;  simpl in code_loop;  unfold code_loop  in Hsubprog;unfold code_loop  in Hsubprog'.
  simpl;  focus 4.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack ;  rewrite <- existsJudgment; intro ptr.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack ;  rewrite <- existsJudgment; intro ptr'.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack ;  rewrite <- existsJudgment; intro next.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack ;  rewrite <- existsJudgment; intro next'.
  normal_form_conj.
  focus_conj 1.
  focus 4; rewrite split_list in Hsubprog; rewrite split_list in Hsubprog'. 
  instr_moveimmindo start start' 0 Hsubprog Hsubprog' with nat next next' on values workreg workreg x0 x0'.

  assert (enveq' : env =  (headenv++[a0]) ++ env').
  rewrite app_ass;  rewrite  <- app_comm_cons; simpl.
  rewrite <- h1; assumption.
  set (H := dismember_program _ _ Hsubprog).
  destruct H as (foo,Hsubprog_end); clear foo.
  set (H := dismember_program _ _ Hsubprog').
  destruct H as (foo,Hsubprog'_end); clear foo.
  hreplace IHm with (IHm (headenv++[a0]) env' (start+1) (start'+1) next next' p p' n n' 
  stack_list stack_free stack_free' enveq' h2 Hstack Hstack' Hsubprog_end Hsubprog'_end).
  eapply judgment_leq; [idtac | apply IHm;omega_replace (start+ 1 + m+ 1) with (start +2+m);
    omega_replace (start'+ 1 + m+ 1) with (start' +2+m); try assumption].
  focus 1; focus' 4.
  eapply ppredtensorcompat2. 
  rewrite <- (tail_unfold_env_semantic headenv a0 Pa n n' x0 x0' ptr ptr' next next').
  apply stackdescriptioncompat2; trivial;  ineq_ring_conj.
  focus' 3; reflexivity.
 
(* applying the induction hypothesis*)

  assert (envNil : env = nil ++ env);simpl;trivial.
  replace (start+S(S(m+1))) with (start+2+m+1) in hyp; try omega.
  replace (start'+S(S(m+1))) with (start'+2+m+1) in hyp; try omega.
  hreplace H with (H nil env (start+2) (start'+2) ptr_env ptr_env' p p' ptr_env ptr_env' stack_list stack_free stack_free' envNil h Hstack_free Hstack_free' Hsubprog_loop Hsubprog'_loop hyp).
  eapply judgment_leq; try apply H.
  focus 5;focus' 4.
  eapply ppredtensorcompat2.
  simpl; setoid_replace (ptr_env=ptr_env) with True using relation iff; try (intuition; omega).
  simpl; setoid_replace (ptr_env'=ptr_env') with True using relation iff; try (intuition; omega).
  setoid_replace (True /\ True) with True using relation iff; [idtac | intuition].
  fold cloudPtrue ; rewrite cloudRelConj_unit; reflexivity.
  ineq_ring.
Qed.


