(* correction_gt.v

   This is the partial correctness proof for coercion for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import correction_compiler_definition.

Typeclasses eauto := dfs 10000.

(* the first lemma prove the compatibilty of coercion and cloudPredLeq *)

Lemma coerce_leq : forall a b Pa (p: a <-< b) ptr ptr', 
      cloudRelLeq (([|a|]) Pa ptr ptr') (([|b|]) Pa ptr ptr').
Proof.
  intros until Pa; intro Pcoerce;  induction Pcoerce.
  intros; simpl; apply cloudRellift_Leq; intuition.
  intros; simpl; apply cloudRellift_Leq; intuition.
  intros; simpl.
  apply cloudExP_mono; intros ptrb.
  apply cloudExP_mono; intros ptra.
  apply cloudExP_mono; intros ptrb'.
  apply cloudExP_mono; intros ptra'.
  rewrite (IHPcoerce1 ptra ptra');  rewrite (IHPcoerce2 ptrb ptrb'); ineq_ring_conj.
  intros; simpl.
  apply cloudExP_mono; intro Rprivate.
  apply ppredconjcompat2; try ineq_ring_conj.
  apply ptstoCovariant_c; [ reflexivity | reflexivity | 
  apply latercompat2; apply perpContravariant].
  unfold flip.
  unfold Pre_arrow.
  repeat(
  match goal with
    | |- stateRelLeq ?P ?Q =>  set (Conj_Block := Q)
  end;
  repeat(rewrite <- stateExP_tensor);
    unfold Conj_Block; clear Conj_Block;
      repeat (rewrite <- stateExP_tensor);
        apply stateExP_mono; intro).
  normal_form.
  focus 2; focus 3;focus' 2;focus' 3.
  apply ppredtensorcompat2.
  apply cloudtostatecompat2.
  apply ppredconjcompat2; try reflexivity.
  apply ptstoCovariant_c; [reflexivity | reflexivity|
  apply latercompat2; apply perpContravariant].
  unfold flip, Post_arrow.
  repeat(
  match goal with
    | |- stateRelLeq ?P ?Q =>  set (Conj_Block := Q)
  end;
  repeat(rewrite <- stateExP_tensor);
    unfold Conj_Block; clear Conj_Block;
      repeat (rewrite <- stateExP_tensor);
        apply stateExP_mono; intro).
  rewrite (IHPcoerce2 x7 x8); try ineq_ring.
  apply ppredtensorcompat2; try ineq_ring.
  rewrite (IHPcoerce1 x3 x4); reflexivity.
Qed.

Lemma correction_of_coerce : forall base base' Ra alloc alloc' dealloc dealloc' env a b e (Pcoercion : a <-< b)  (ta : env |-- e ::: a), 
(forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env a e ta Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') -> 
  forall Rc Rc_cloud start  start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env b e (TCOERCE Pcoercion ta) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3'.
Proof with normal_form.
  intros until e; intros Pcoercion ta HypRec;  intros.
  unfold statement_of_compiler_correctness;
  intros until global_code'; intros hprog hcode Hstack_free hprog' hcode' Hstack_free' Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc.

  hreplace HypRec with (HypRec Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' code code' p p' stack_list global_code global_code' hprog hcode Hstack_free hprog' hcode' Hstack_free' Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc).
  destruct HypRec as (HypRec,HypRec_extra).
  split.
  
  (* just apply coerce_leq for the main property *)
  intro hyp; eapply HypRec; try eassumption.
  intros; unfold memory_specification in *;  simpl in *...
  focus 5.
  rewrite (coerce_leq a b Ra Pcoercion).
  eapply judgment_leq; try apply (hyp ptr ptr'); ineq_ring.

  (* the extra code is more compicated as we need to do an induction on the coercion*)
  clear - HypRec_extra Pcoercion; intros e' He'.
  induction Pcoercion; try (simpl in HypRec_extra; destruct (HypRec_extra e' He')).
  (* case a --> b <-< a' --> b'*)
  simpl in HypRec_extra; simpl.
  intros; hreplace HypRec_extra with (HypRec_extra e' He' ptr ptr' fb fb').
  eapply judgment_leq ; try apply HypRec_extra; clear HypRec_extra.
  (* We need to unfold Pre_arrow and Post_arrow to rewrite *)
  unfold Pre_arrow in *.
  repeat(
  match goal with
    | |- stateRelLeq ?P ?Q =>  set (Conj_Block := Q)
  end;
  repeat(rewrite <- stateExP_tensor);
    unfold Conj_Block; clear Conj_Block;
      repeat (rewrite <- stateExP_tensor);
        apply stateExP_mono; intro).
  normal_form; focus 2.
  rewrite (coerce_leq a' a Ra Pcoercion1).
  focus' 3; focus 3.
  apply ppredtensorcompat2; try ineq_ring.
  apply cloudtostatecompat2; apply ppredconjcompat2; try reflexivity; apply ptstoCovariant_c;
    [ reflexivity | reflexivity | apply latercompat2; apply perpContravariant].
  unfold Post_arrow, flip.
    repeat(
  match goal with
    | |- stateRelLeq ?P ?Q =>  set (Conj_Block := Q)
  end;
  repeat(rewrite <- stateExP_tensor);
    unfold Conj_Block; clear Conj_Block;
      repeat (rewrite <- stateExP_tensor);
        apply stateExP_mono; intro).
  rewrite (coerce_leq b b' Ra Pcoercion2).
  ineq_ring.
Qed.