(* correction_lambda.v

   This is the partial correctness proof for the compiler for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)


Require Import correction_compiler_definition.


Require Import correction_basic_ops.
Require Import correction_pair.
Require Import correction_first.
Require Import correction_second.
Require Import correction_op.
Require Import correction_app.
Require Import correction_ifThenElse.
Require Import correction_gt.
Require Import correction_coercion.
Require Import correction_lambda.
Require Import correction_fixpoint.


Typeclasses eauto := dfs 10000.

(* This is just gluing all the lemmas together *)

Lemma compiler_sound_rec : forall base base' Ra alloc alloc' dealloc dealloc' env a e (t:env |-- e ::: a)
  Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' ptr_env n2 n3 ptr_env' n2' n3',
 statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env a e t Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' ptr_env n2 n3 ptr_env' n2' n3'.
Proof.
  induction t;  [apply correction_of_int | apply correction_of_bool | apply correction_of_var |
  apply correction_of_lambda; assumption | apply correction_of_fixpoint; assumption |
  apply correction_of_app; assumption | apply correction_of_op; assumption |
  apply correction_of_gt; assumption | apply correction_of_pair; assumption |
  apply correction_of_first; assumption | apply correction_of_second; assumption |
  apply correction_of_ifThenElse; assumption | apply correction_of_coerce; assumption].
Qed.

(********* here is the proof of compiler correctness for open terms  ****)

Theorem compiler_sound : forall base base' Ra init init' alloc alloc' dealloc dealloc' env a e (t:env |-- e ::: a) Rc Rc_cloud start start' stack_ptr stack_ptr' ptr_env n2 n3 ptr_env' n2' n3' p p' stack_list,
  let global_code := compile e nil base alloc dealloc in
  let code := extract_code_from_globalcode global_code start in
  let stack_free := extract_stacksize_from_code code in
  let global_code' := compile e nil base' alloc' dealloc' in
  let code' := extract_code_from_globalcode global_code' start' in
  let stack_free' := extract_stacksize_from_code code' in
    prog_extends_auxcode global_code p base ->
    prog_extends_code global_code p start ->
    prog_extends_auxcode global_code' p' base' ->
    prog_extends_code global_code' p' start' ->
    AllocSpec Ra init init' alloc alloc' dealloc dealloc' p p' ->
    (forall ptr ptr',
      |= start+length (instruct_of code), start'+length (instruct_of code') :
        memory_specification Ra Rc Rc_cloud env (([|a|] Ra, (ptr,ptr')) :: stack_list)
         (stack_free-1) (stack_free'-1) (stack_ptr+1) (stack_ptr'+1) ptr_env n2 n3
         ptr_env'  n2' n3' ; p, p') ->
    |= start,start': memory_specification Ra Rc Rc_cloud env stack_list stack_free stack_free' stack_ptr stack_ptr' ptr_env n2 n3 ptr_env' n2' n3'; p,p' .
Proof.
  intros until stack_free'.
  intros Hprog_aux Hprog Hprog_aux' Hprog' Halloc.
  unfold AllocSpec in Halloc.
  destruct Halloc as (Hinit & Halloc & Hdealloc).
  assert (HypRec := compiler_sound_rec base base' Ra alloc alloc' dealloc dealloc' env a e (t:env |-- e ::: a)
  Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' ptr_env n2 n3 ptr_env' n2' n3' nil nil code code' p p' stack_list global_code global_code').
  assert (hmain : global_code = compile e nil base alloc dealloc); trivial.
  assert (hmain' : global_code' = compile e nil base' alloc' dealloc'); trivial.
  assert (hcode : code = extract_code_from_globalcode global_code start); trivial.
  assert (hcode' : code' = extract_code_from_globalcode global_code' start'); trivial.
  destruct HypRec as [HypRec HypExtra]; trivial;
  try first [simpl; replace (base+0) with base by omega;
      rewrite <- hmain; assumption |
        simpl; replace (base'+0) with base' by omega;
          rewrite <- hmain'; assumption].
Qed.


(********* here is the proof of compiler correctness for closed terms  ****)

Theorem compiler_sound_closed_terms : forall Ra init alloc dealloc init' alloc' dealloc' a e (t:nil|--e:::a) base base' p p',
  let code := compile' e base init alloc dealloc in
  let code' := compile' e base' init' alloc' dealloc' in
    prog_extends_maincode code p base ->
    prog_extends_maincode code' p' base' ->
    AllocSpec Ra init init' alloc alloc' dealloc dealloc' p p' ->
    |= base+length (snd code), base'+length (snd code') :
      Ex ptr,Ex ptr',((retreg |-> {ptr,ptr'}) ## cloudToState (([|a|]) Ra ptr ptr'));
      p,p' ->
    |=  fst code , fst code' : Topfrom 0 0 ; p, p'.
Proof with normal_form.
  intros until code'.
  intros Hprog Hprog' Halloc hyp.
  unfold AllocSpec in Halloc.
  destruct Halloc as (Hinit & Halloc & Hdealloc).
  assert (hcode : code = compile' e base init alloc dealloc); trivial.
  assert (hcode' : code' = compile' e base' init' alloc' dealloc'); trivial.
  set (global_code_gen := compile e nil base alloc dealloc).
  set (global_code_gen' := compile e nil base' alloc' dealloc').
  assert (hmain : global_code_gen = compile e nil base alloc dealloc); trivial.
  assert (hmain' : global_code_gen' = compile e nil base' alloc' dealloc'); trivial.
  rewrite hcode; rewrite hcode'.
  setoid_rewrite aux_next_in_compile.
  rewrite <- hmain; rewrite <- hmain'.  
  unfold compile' in hcode;  simpl in hcode.
  rewrite <- hmain in hcode.
  unfold compile' in hcode';  simpl in hcode'.
  rewrite <- hmain' in hcode'.
  do 10 (rewrite topfromBlock).
  setoid_replace ((7 |->-) ## (8 |->-) ## (9 |->-) ## Topfrom 10 10) with 
    (((7 |->-) ## (8 |->-) ## (9 |->-)) ## Topfrom 10 10) using relation stateRelEq by ring.  fold unused_space.
  fold envreg spreg workreg argreg retreg.
  unfold workreg at 3; unfold workreg at 4;  unfold workreg at 3 ; unfold workreg at 3.
  unfold envreg at 3; unfold envreg at 3.
  (* extracting the real piece of code *)
  rewrite (surjective_pairing global_code_gen) in hcode;
  rewrite (surjective_pairing (fst global_code_gen)) in hcode.
  rewrite (surjective_pairing global_code_gen') in hcode';
  rewrite (surjective_pairing (fst global_code_gen')) in hcode'.
  set (main:= fst (fst global_code_gen)) in *.
  set (aux_next:= snd global_code_gen) in *.
  set (main':= fst (fst global_code_gen')) in *.
  set (aux_next':= snd global_code_gen') in *.
  rewrite (surjective_pairing (main (aux_next + intro_code_size))) in hcode.
  rewrite (surjective_pairing (main (aux_next + intro_code_size))) in hcode; simpl in hcode.  rewrite (surjective_pairing (fst(main (aux_next + intro_code_size)))) in hcode.

  rewrite (surjective_pairing (main' (aux_next' + intro_code_size))) in hcode'.
  rewrite (surjective_pairing (main' (aux_next' + intro_code_size))) in hcode'; simpl in hcode'.
  rewrite (surjective_pairing (fst(main' (aux_next' + intro_code_size)))) in hcode'.
  set (global_code := fst (fst (main (aux_next + intro_code_size)))) in *;  simpl in hcode.
  set (aux_code := snd(fst global_code_gen)) in *.
  set (global_code' := fst (fst (main' (aux_next' + intro_code_size)))) in *;  simpl in hcode'.
  set (aux_code' := snd(fst global_code_gen')) in *.
  assert (size_aux_next : aux_next = base + length aux_code). 
  unfold aux_next; rewrite hmain;  replace base with (base + length (@nil instruction)); try (simpl;omega).
  rewrite compile_length_aux; simpl; omega_replace (base+0) with base;  rewrite <- hmain; fold aux_code; reflexivity.
  assert (size_aux_next' : aux_next' = base' + length aux_code'). 
  unfold aux_next'; rewrite hmain';  replace base' with (base' + length (@nil instruction)); try (simpl;omega).
  rewrite compile_length_aux; simpl; omega_replace (base'+0) with base';  rewrite <- hmain'; fold aux_code'; reflexivity.
  rewrite hcode in Hprog ; rewrite hcode in hyp ; simpl in Hprog; simpl in hyp ; clear hcode.
  rewrite hcode' in Hprog' ; rewrite hcode' in hyp ; simpl in Hprog'; simpl in hyp ; clear hcode'.
  hreplace Hprog with (dismember_program _ _ Hprog).
  hreplace Hprog' with (dismember_program _ _ Hprog').
  destruct Hprog as (Hprog,Hintro).
  destruct Hprog' as (Hprog',Hintro').
  match goal with | h: PfunctionExtendsFunction p (fragfromprog (progfromlist ?l) ?b) |- _ =>  
     set (ugly_list := l) in * end.
  match goal with | h: PfunctionExtendsFunction p' (fragfromprog (progfromlist ?l) ?b) |- _ =>  
     set (ugly_list' := l) in * end.
  assert (ugly_list = (fuse (takedrop 7 nil ugly_list))); try (simpl;trivial).
  Opaque fuse. simpl in H. Transparent fuse.
  unfold fuse in H; rewrite H in Hintro; rewrite H in hyp; clear H ugly_list.
  assert (ugly_list' = (fuse (takedrop 7 nil ugly_list'))); try (simpl;trivial).
  Opaque fuse. simpl in H. Transparent fuse.
  unfold fuse in H; rewrite H in Hintro'; rewrite H in hyp; clear H ugly_list'.
  hreplace Hintro with (dismember_program _ _ Hintro).
  destruct Hintro as (Hintro,Hmain); simpl in Hmain.
  hreplace Hmain with (dismember_program _ _ Hmain).
  destruct Hmain as (Hmain,Hreturn); simpl in Hreturn.
  hreplace Hintro' with (dismember_program _ _ Hintro').
  destruct Hintro' as (Hintro',Hmain'); simpl in Hmain'.
  hreplace Hmain' with (dismember_program _ _ Hmain').
  destruct Hmain' as (Hmain',Hreturn'); simpl in Hreturn'.
  
  (* unfolding the intro code *)

  (* [ret] <- label + 2 *)
  intros.
  rewrite size_aux_next; set (new_start := base + length aux_code) in *.
  rewrite size_aux_next'; set (new_start' := base' + length aux_code') in *.
  instr_moveimm new_start new_start' 0 Hintro Hintro' with nat (aux_next+2) (aux_next'+2).
  (*jmp init*)
  focus 6; rewrite existsBlank;rewrite <- existsJudgment; intro n.
  rewrite <- existsJudgment; intro n'.
  focus 6; rewrite existsBlank; rewrite <- existsJudgment; intro ptr.
  rewrite <- existsJudgment; intro ptr'.
  do 5 focus 6.
  instr_jumpcst new_start new_start' 1 Hintro Hintro'.
  unfold SPre_init , unchange_stack in Hinit; hreplace Hinit with (Hinit ptr ptr' n n').
  assert (Hinit2 := total_to_partial _ _ _ _ _ _ _ Hinit).
  normal_form in Hinit2.
  eapply continuation_enrichment; try apply Hinit2.
  (* [arg] <- stack_size + 5 *)
  unfold SPost_init, unchange_stack.
  set (stack_size := snd(fst (main (aux_next+intro_code_size)))) in *.
  set (stack_size' := snd(fst (main' (aux_next'+intro_code_size)))) in *.
  focus 2; rewrite size_aux_next; fold new_start. 
  rewrite size_aux_next'; fold new_start'. 
  instr_moveimm new_start new_start' 2 Hintro Hintro' with nat (stack_size+5) (stack_size'+5).
  (* [ret] <- label + 5*)
  focus 2; instr_moveimm new_start new_start' 3 Hintro Hintro' with nat (aux_next+5) (aux_next'+5).
  (* jmp alloc *)
  instr_jumpcst new_start new_start' 4 Hintro Hintro'.
  unfold SPre_alloc, unchange_stack in Halloc.
  assert (Halloc' := (Halloc statePtrue (stack_size+5) (stack_size'+5) ptr ptr' n n')).
  assert (Halloc2 := total_to_partial _ _ _ _ _ _ _ Halloc').
  eapply continuation_enrichment  with (Post := SPost_alloc Ra statePtrue (stack_size+5) (stack_size'+5) ptr ptr' n n');
  [idtac | eapply judgment_leq; try apply Halloc2 ; ineq_ring].
  unfold SPost_alloc, unchange_stack.
  rewrite <- existsJudgment; intro fb.
  rewrite <- existsJudgment; intro fb'...
  (*  [env] <- [ret] *)
  rewrite size_aux_next; fold new_start.
  rewrite size_aux_next'; fold new_start' ; focus 8.
  instr_moveimm new_start new_start' 5 Hintro Hintro' with nat fb fb'.
  (* [sp] <- [ret]+5 *)
  focus 1;focus 8.
  instr_opimmleft plus new_start new_start' 6 Hintro Hintro' fb fb' 4 4.
  clear Hintro Hintro' Halloc2 Hinit2 Halloc'.

(* applying the main code *)

  focus 3.
  omega_replace (stack_size+5) with (S (stack_size+4)).
  omega_replace (stack_size'+5) with (S(stack_size'+4)).
  rewrite block_unfold_start ; normal_form; rewrite stateRelTensor_comm; rewrite stateRelTensor_weakening.
  omega_replace (stack_size+4) with (S(stack_size+3)).
  omega_replace (stack_size'+4) with (S(stack_size'+3)).
  rewrite block_unfold_start ; normal_form; rewrite stateRelTensor_comm; rewrite stateRelTensor_weakening.
  omega_replace (stack_size+3) with (S(stack_size+2)).
  omega_replace (stack_size'+3) with (S(stack_size'+2)).  
  rewrite block_unfold_start ;normal_form; rewrite existsBlank; rewrite <- existsJudgment; intro n2.
  rewrite <- existsJudgment; intro n2'.
  focus 1.
  omega_replace (stack_size+2) with (S(stack_size+1)).
  omega_replace (stack_size'+2) with (S(stack_size'+1)).
  rewrite block_unfold_start ;normal_form; rewrite existsBlank; rewrite <- existsJudgment; intro n3.
  rewrite <- existsJudgment; intro n3'.
  focus 1.
  omega_replace (stack_size+1) with (S(stack_size)).
  omega_replace (stack_size'+1) with (S(stack_size')).
  rewrite block_unfold_start ...
  omega_replace (S fb) with (fb+1); omega_replace (S(fb+1)) with (fb+2); omega_replace (S(fb+2)) with (fb+3); omega_replace (S(fb+3)) with (fb+4); omega_replace (S(fb+4)) with (fb+4+1).
  omega_replace (S fb') with (fb'+1); omega_replace (S(fb'+1)) with (fb'+2); omega_replace (S(fb'+2)) with (fb'+3); omega_replace (S(fb'+3)) with (fb'+4); omega_replace (S(fb'+4)) with (fb'+4+1).
  focus 5; rewrite ptsto_blank_leq.
  set (hypRec := compiler_sound_rec base base' Ra alloc alloc' dealloc dealloc' nil a e t statePtrue cloudPtrue (new_start+7) (new_start'+7) stack_size stack_size' (fb+4) (fb'+4) fb n2 n3 fb' n2' n3' nil nil (main (new_start+7)) (main' (new_start'+7)) p p' nil global_code_gen global_code_gen').
  simpl in hypRec.
  replace (base +0) with base in hypRec; try omega; fold main in hypRec.
  replace (base' +0) with base' in hypRec; try omega; fold main' in hypRec.
  unfold stack_size in hypRec; rewrite size_aux_next in hypRec; unfold intro_code_size in hypRec; fold aux_code in hypRec.
  unfold stack_size' in hypRec; rewrite size_aux_next' in hypRec; fold aux_code' in hypRec.
  unfold global_code in Hmain ; rewrite size_aux_next in Hmain; unfold intro_code_size in Hmain.
  unfold global_code' in Hmain' ; rewrite size_aux_next' in Hmain'; unfold intro_code_size in Hmain'.
  hreplace hypRec with (hypRec hmain (refl_equal (main(new_start+7))) (le_refl (snd(fst(main(new_start+7))))) hmain' (refl_equal (main'(new_start'+7))) (le_refl (snd(fst(main'(new_start'+7))))) Hprog Hmain Hprog' Hmain' Halloc Hdealloc).
  destruct hypRec as (hypRec,Huseless); clear Huseless.
  unfold statement_of_compiler_correctness in hypRec.
  eapply judgment_leq; [idtac | eapply hypRec]; try assumption; clear hypRec.
  unfold memory_specification, storing_space; simpl.
  focus' 4;rewrite cloudRelConj_unit.
  rewrite cloudToStatePtrue.
  unfold stack_size; rewrite size_aux_next;unfold intro_code_size.
  unfold stack_size'; rewrite size_aux_next';unfold intro_code_size.
  ineq_ring.

(* unfolding the return value *)
  
  clear Hinit Hmain ptr ptr'; intros ptr ptr'.
  unfold memory_specification; rewrite <- size_aux_next; rewrite <- size_aux_next'; fold intro_code_size global_code global_code'.
  rewrite size_aux_next; unfold intro_code_size.
  rewrite size_aux_next'; unfold intro_code_size.
  set (new_start2 := new_start +7+length global_code) in *.
  set (new_start2' := new_start' +7+length global_code') in *; simpl ...
  focus 4;focus 1;focus 8.
  instr_moveimmind new_start2 new_start2' 0 Hreturn Hreturn' with nat ptr ptr' on values spreg spreg (fb+4+1) (fb'+4+1).
  focus 6; isolate_conj.
  rewrite cloudRelConj_comm in Block; normal_form_conj in Block.
  rebuilt_conj Conj_Block Block.
  focus 1; rewrite stateRelTensor_assoc; rewrite stateRelTensor_weakening.
  repeat rewrite app_length in hyp; repeat rewrite plus_assoc in hyp; simpl in hyp.
  rewrite <- existsJudgment_single in hyp.
  hreplace hyp with (hyp ptr).
  rewrite <- existsJudgment_single in hyp.
  hreplace hyp with (hyp ptr').
  unfold new_start2, new_start2', new_start, new_start'.
  apply hyp.
Qed.
