(* statement_of_compiler_correctness.v

   Define the shape of the correctness lemma.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export compiler.
Require Export theory_of_types.
Require Export program_semantics.
Require Export allocator_spec.

Typeclasses eauto := dfs 10000.

Fixpoint extra_property a env Ra p p' e aux_code aux_code' base base' alloc alloc' dealloc dealloc' {struct a} : Prop :=
match a with 
  | a ---> b => forall ptr ptr' fb fb', 
    |=  (snd (compile e aux_code (base+length aux_code) alloc dealloc)),
      (snd (compile e aux_code' (base'+length aux_code') alloc' dealloc')) :
      (Pre_arrow ({ptr+1,ptr'+1}||->{fb,fb'} //\\ ([||env||]) Ra fb fb') ptr ptr'  
        Ra ([|a|]) ([|b|])); p,p'
  | _ => False
end.

Definition statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env a e (t : env |-- e:::a) 
  Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3':=
     forall aux_code aux_code' code code' p p' stack_list global_code global_code'
     (hprog : global_code = compile e aux_code (base+length aux_code) alloc dealloc)
     (hcode : code = extract_code_from_globalcode global_code start)
     (hstack : extract_stacksize_from_code code <= stack_free)
     (hprog' : global_code' = compile e aux_code' (base'+length aux_code') alloc' dealloc')
     (hcode' : code' = extract_code_from_globalcode global_code' start')
     (hstack' : extract_stacksize_from_code code' <= stack_free'),
     prog_extends_auxcode global_code p base ->
     prog_extends_code global_code p start ->
     prog_extends_auxcode global_code' p' base' ->
     prog_extends_code global_code' p' start' ->
     (forall Rc_alloc size size' n5 n5' n6 n6', total_judgment (SPre_alloc Ra Rc_alloc size size' n5 n5' n6 n6') 
            (SPost_alloc Ra Rc_alloc size size' n5 n5' n6 n6') p p' alloc alloc') ->
     (forall Rc_dealloc size size' base base' n5 n5' n6 n6', total_judgment (SPre_dealloc Ra Rc_dealloc base base' size size' n5 n5' n6 n6')
            (SPost_dealloc Ra Rc_dealloc n5 n5' n6 n6') p p' dealloc dealloc') ->
     ((forall ptr ptr', 
       |= start+List.length (fst(fst code)),start'+List.length (fst(fst code')) :
        memory_specification Ra Rc Rc_cloud env (([|a|] Ra, (ptr,ptr')) :: stack_list)
        (stack_free-1) (stack_free'-1) (stack_ptr+1) (stack_ptr'+1) n n2 n3 n' n2' n3' ;
        p,p')->
     |=  start, start' : (memory_specification Ra Rc Rc_cloud env stack_list stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3'); p,p')
     /\ 
     (forall e' , e = LAMBDA e' -> extra_property a env Ra p p' e' aux_code aux_code' 
       base base' alloc alloc' dealloc dealloc').

Ltac clean_judgment_basic := 
  unfold prog_extends_code, prog_extends_auxcode, extract_code_from_globalcode, extract_stacksize_from_code;
  intros until global_code'; intros hprog hcode Hstack_free hprog' hcode' Hstack_free'
    Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc;
  try (split ; [idtac | intros e' Hlambda; inversion Hlambda]);
  intro hyp; 
  rewrite <- hcode in Hsubprog; rewrite <- hcode' in Hsubprog';
  rewrite hprog in hcode; rewrite hprog' in hcode'; 
  clear hprog hprog' Hsubprog_aux Hsubprog_aux' ;simpl in hcode; simpl in hcode';
  unfold append_code_table, int_code, bool_code,var_code, code_table_nil in hcode; simpl in hcode;
  unfold append_code_table, int_code, bool_code,var_code, code_table_nil in hcode'; simpl in hcode';
  rewrite <- triple_destruction in hcode; rewrite <- triple_destruction in hcode;
  simpl in hcode; destruct hcode as ((hcode,hsize),hfoo);  clear hfoo;
  rewrite <- triple_destruction in hcode'; rewrite <- triple_destruction in hcode';
  simpl in hcode'; destruct hcode' as ((hcode',hsize'),hfoo');  clear hfoo';
  rewrite hcode,hcode' in hyp; simpl in hyp; rewrite hcode in Hsubprog;rewrite hcode' in Hsubprog';
  unfold memory_specification;  clear hcode hcode'.

Ltac apply_arg code1_gen code1_gen' code1 code1' hcode hcode' stack_free stack_free' p p' start start' base base' Rc Rc_cloud stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' stack_list
  Halloc Hdealloc Hsubprog Hsubprog' Hsubprog_aux Hsubprog_aux' type ptr ptr' Ra HypRecE1:= 
  assert (Hstack_free1 : snd(fst code1)  <= stack_free); [  eapply le_trans; try eassumption;
  eapply le_trans; try apply (le_max_l (snd(fst code1) +0)); omega | idtac];
  assert (Hstack_free1' : snd(fst code1')  <= stack_free'); [  eapply le_trans; try eassumption;
  eapply le_trans; try apply (le_max_l (snd(fst code1') +0)); omega | idtac];
  assert (Hsubprog1 : PfunctionExtendsFunction p (fragfromprog (progfromlist (fst(fst code1))) start)); 
  [eapply subPfunction_extends; try apply Hsubprog;  rewrite <- sub_is_extends;  rewrite hcode;
  simpl;  apply prefix_code | idtac];
  assert (Hsubprog1' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (fst(fst code1'))) start')); 
  [eapply subPfunction_extends; try apply Hsubprog';  rewrite <- sub_is_extends;  rewrite hcode';
  simpl;  apply prefix_code | idtac];
  hreplace HypRecE1 with (HypRecE1 Rc Rc_cloud start start'  stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' code1 code1' p p' stack_list code1_gen code1_gen');
  fold code1 code1_gen in HypRecE1;
  fold code1' code1_gen' in HypRecE1;
  hreplace HypRecE1 with (HypRecE1 (refl_equal code1_gen) (refl_equal code1) Hstack_free1 (refl_equal code1_gen') (refl_equal code1') Hstack_free1' Hsubprog_aux Hsubprog1 Hsubprog_aux' Hsubprog1' Halloc Hdealloc);
  destruct HypRecE1 as (HypRecE1,Huseless); clear Huseless;
  apply HypRecE1;  clear HypRecE1 Hstack_free1 Hsubprog1 Hstack_free1' Hsubprog1'.


Ltac apply_arg_e2 code2_gen code1_gen code2 code2_gen' code1_gen' code2' hcode hcode' stack_free stack_free' p p' start base start' base' Rc Rc_cloud stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' stack_list Halloc Hdealloc Hsubprog Hsubprog_aux Hsubprog' Hsubprog_aux' HypRecE2 Hstack Hstack' := 
  assert (Hstack_free : snd(fst code2)  <= stack_free) ;
  [eapply le_trans ; try apply Hstack ; apply le_max_l | idtac];
  assert (Hsubprog2 : PfunctionExtendsFunction p (fragfromprog (progfromlist (fst(fst code2))) start));
  [eapply subPfunction_extends; try apply Hsubprog; instantiate ; rewrite <- sub_is_extends;  rewrite hcode ;
  simpl; rewrite app_ass; apply prefix_code | idtac];
  assert (Hsubprog2_aux : PfunctionExtendsFunction p (fragfromprog (progfromlist (snd(fst code2_gen))) base));
  [eapply subPfunction_extends; try apply Hsubprog_aux;  unfold code1_gen;
  rewrite <- sub_is_extends; unfold code2_gen; rewrite compile_length_aux; fold code2_gen;  apply auxcode_monotony | idtac];
  assert (Hstack_free' : snd(fst code2')  <= stack_free') ;
  [eapply le_trans ; try apply Hstack' ; apply le_max_l | idtac];
  assert (Hsubprog2' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (fst(fst code2'))) start'));
  [eapply subPfunction_extends; try apply Hsubprog'; instantiate ; rewrite <- sub_is_extends;  rewrite hcode' ;
  simpl; rewrite app_ass; apply prefix_code | idtac];
  assert (Hsubprog2_aux' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (snd(fst code2_gen'))) base'));
  [eapply subPfunction_extends; try apply Hsubprog_aux';  unfold code1_gen';
  rewrite <- sub_is_extends; unfold code2_gen'; rewrite compile_length_aux; fold code2_gen';  apply auxcode_monotony | idtac];
  hreplace HypRecE2 with (HypRecE2 Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' code2 code2' p p' stack_list code2_gen code2_gen' );
  fold code2 code2_gen in HypRecE2;
  fold code2' code2_gen' in HypRecE2;
  hreplace HypRecE2 with (HypRecE2 (refl_equal code2_gen) (refl_equal code2) Hstack_free (refl_equal code2_gen') (refl_equal code2') Hstack_free' Hsubprog2_aux Hsubprog2 Hsubprog2_aux' Hsubprog2' Halloc Hdealloc);
  destruct HypRecE2 as (HypRecE2,Huseless); clear Huseless;
  apply HypRecE2;  clear HypRecE2 Hstack_free Hstack_free' Hsubprog2 Hsubprog2_aux Hsubprog2' Hsubprog2_aux'.

Ltac apply_arg_e1 code2_gen code1_gen code2 code1 code2_gen' code1_gen' code2' code1'
  hcode hcode' stack_free stack_free' p p' start base start' base' Ra Rc Rc_cloud 
  stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' stack_list Halloc Hdealloc 
  Hsubprog Hsubprog_aux Hsubprog' Hsubprog_aux' type ptr ptr' HypRecE1 :=
   assert (foo : snd(fst code1) +1 <= stack_free); 
    [eapply le_trans ; try eassumption ; apply le_max_r | idtac];
  assert (Hstack1 : snd(fst code1)  <= stack_free-1); try omega; clear foo;
  assert (Hsubprog1 : PfunctionExtendsFunction p (fragfromprog (progfromlist (fst(fst code1))) (start+length (fst (fst code2)))));
  [eapply subPfunction_extends; try apply Hsubprog;  rewrite <- sub_is_extends;  rewrite hcode;
  simpl; rewrite app_ass; apply sub_code | idtac];
   assert (foo' : snd(fst code1') +1 <= stack_free'); 
    [eapply le_trans ; try eassumption ; apply le_max_r | idtac];
  assert (Hstack1' : snd(fst code1')  <= stack_free'-1); try omega; clear foo';
  assert (Hsubprog1' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (fst(fst code1'))) (start'+length (fst (fst code2')))));
  [eapply subPfunction_extends; try apply Hsubprog';  rewrite <- sub_is_extends;  rewrite hcode';
  simpl; rewrite app_ass; apply sub_code | idtac];
  hreplace HypRecE1 with (HypRecE1 Rc Rc_cloud (start+length (fst (fst code2))) (start'+length (fst (fst code2'))) (stack_free-1) (stack_free'-1) (stack_ptr+1) (stack_ptr'+1) n n2 n3 n' n2' n3'
    (snd(fst code2_gen)) (snd(fst code2_gen')) code1 code1' p p' (([|type|] Ra , (ptr,ptr')) :: stack_list) code1_gen code1_gen');
  unfold code2_gen in HypRecE1; rewrite <- compile_length_aux in HypRecE1;  fold code2_gen code1_gen in HypRecE1;
  unfold code2, code2_gen in HypRecE1; rewrite <- compile_length_main in HypRecE1;  fold code2 code1 in HypRecE1; fold code2_gen code2 code1 in HypRecE1;
  unfold code2, code2_gen in Hsubprog1; rewrite <- compile_length_main in Hsubprog1;  fold code2 code1 in HypRecE1; fold code2_gen code2 code1 in Hsubprog;
  unfold code2_gen' in HypRecE1; rewrite <- compile_length_aux in HypRecE1;  fold code2_gen' code1_gen' in HypRecE1;
  unfold code2', code2_gen' in HypRecE1; rewrite <- compile_length_main in HypRecE1;  fold code2' code1' in HypRecE1; fold code2_gen' code2' code1' in HypRecE1;
  unfold code2', code2_gen' in Hsubprog1'; rewrite <- compile_length_main in Hsubprog1';  fold code2' code1' in HypRecE1; fold code2_gen' code2' code1' in Hsubprog';
  hreplace HypRecE1 with (HypRecE1 (refl_equal code1_gen) (refl_equal code1) Hstack1 (refl_equal code1_gen') (refl_equal code1') Hstack1' Hsubprog_aux Hsubprog1 Hsubprog_aux' Hsubprog1' Halloc Hdealloc);
  unfold code2, code2_gen, code2', code2_gen'; repeat rewrite <- compile_length_main;  fold code2_gen code2 code2_gen' code2';
  destruct HypRecE1 as (HypRecE1,Huseless); clear Huseless;
  apply HypRecE1;  clear HypRecE1 Hstack1 Hstack1' Hsubprog1 Hsubprog1'.
  
Ltac compute_code_size code1_gen code2_gen code1 code2 :=
  unfold code1, code1_gen, code2, code2_gen;
  rewrite compile_length_aux; rewrite compile_length_main;  rewrite compile_length_main;  rewrite <- compile_length_aux;
  fold code2_gen;  unfold code2_gen at 4;  rewrite <- compile_length_main;  fold code2_gen code1_gen code2 code1.
