(* correction_fixpoint.v

   This is the partial correctness proof for recursive functions for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import correction_compiler_definition.
Require Import correction_functions.

Typeclasses eauto := dfs 10000.

(* the two first lemmas are here because we want to unfold FIX and not LAMBDA from the expression
   "FIX LAMBDA e". Maybe there uis a clever to do that be this one works *)

Lemma FIX_unfold : forall e aux_code base alloc dealloc start,
let Code_gen := compile e aux_code (base +length aux_code) alloc dealloc in 
let Code := fst (fst Code_gen) (start+5) in 
fst (fst (compile (FIX e) aux_code (base+length aux_code) alloc dealloc)) start = 
( closure_rec_make start alloc base ++ fst (fst Code) ++ closure_rec_make2 (snd Code) alloc (snd Code_gen),
 max (snd (fst Code)) (min (snd (fst Code)) 1), snd Code + 2) .
Proof.
  intros; simpl.
  surjective_pairing_one_arg e aux_code base alloc dealloc in goal.
  surjective_pairing_one_arg e code1_gen (start+5) in goal.
  omega_replace (snd(fst code1) +0) with (snd(fst code1)).
  unfold Code,Code_gen,code1,code1_gen;reflexivity.
Qed.

Lemma FIX_unfold_aux : forall e aux_code base alloc dealloc,
let Code_gen := compile e aux_code (base +length aux_code) alloc dealloc in 
snd (fst (compile (FIX e) aux_code (base+length aux_code) alloc dealloc)) = 
snd (fst Code_gen).
Proof.
  intros; simpl.
  surjective_pairing_one_arg e aux_code base alloc dealloc in goal.
  unfold Code_gen,code1_gen; reflexivity.
Qed.

Lemma correction_of_fixpoint :  forall base base' Ra alloc alloc' dealloc dealloc' env a e  (t : a :: env |-- LAMBDA e :::a), 
(forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' (a::env) a (LAMBDA e) t Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
  forall Rc Rc_cloud start  start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env a (FIX (LAMBDA e)) (TFIX t) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3'.
Proof with normal_form.
  intros until e; intros t HypRec;  intros.
  unfold statement_of_compiler_correctness in *.
  unfold prog_extends_code, prog_extends_auxcode, extract_code_from_globalcode, extract_stacksize_from_code in *;
  intros until global_code'; intros hprog hcode Hstack_free hprog' hcode' Hstack_free' Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc.
  split ; [idtac | intros e' Hlambda; inversion Hlambda]; intro hyp.

  (* recovering the actual piece of code *)

  rewrite <- hcode in Hsubprog; rewrite <- hcode' in Hsubprog';        
  rewrite hprog in hcode; rewrite hprog in Hsubprog_aux; clear hprog.
  rewrite FIX_unfold in hcode.
  rewrite FIX_unfold_aux in Hsubprog_aux.
  set (code1_gen := compile (LAMBDA e) aux_code (base +length aux_code) alloc dealloc) in  *.
  set (code1 := fst (fst code1_gen) (start+5)) in *.
  rewrite hprog' in hcode'; rewrite hprog' in Hsubprog_aux'; clear hprog'.
  rewrite FIX_unfold in hcode'.
  rewrite FIX_unfold_aux in Hsubprog_aux'.
  set (code1_gen' := compile (LAMBDA e) aux_code' (base' +length aux_code') alloc' dealloc') in  *.
  set (code1' := fst (fst code1_gen') (start'+5)) in *.

  rewrite hcode in Hstack_free; simpl in Hstack_free.
  rewrite hcode' in Hstack_free'; simpl in Hstack_free'.

  rewrite hcode in Hsubprog; unfold fst in Hsubprog.
  set (H := dismember_program _ _ Hsubprog). 	 
  destruct H as (Hsubprog_env,Hsubprog_main); clear Hsubprog.
  rename Hsubprog_main into Hsubprog.
  set (foo := fst (fst code1));  assert (hfoo : foo = foo) ; try reflexivity.
  unfold foo at 1 in hfoo;  unfold fst in hfoo; unfold foo in hfoo.
  rewrite hfoo in Hsubprog; clear hfoo foo.
  simpl in Hsubprog;  set (H := dismember_program _ _ Hsubprog). 	 
  destruct H as (Hsubprog_lambda,Hsubprog_main); clear Hsubprog.
  rename Hsubprog_main into Hsubprog_end.

  rewrite hcode' in Hsubprog'; unfold fst in Hsubprog'.
  set (H := dismember_program _ _ Hsubprog'). 	 
  destruct H as (Hsubprog_env',Hsubprog_main); clear Hsubprog'.
  rename Hsubprog_main into Hsubprog'.
  set (foo := fst (fst code1'));  assert (hfoo : foo = foo) ; try reflexivity.
  unfold foo at 1 in hfoo;  unfold fst in hfoo; unfold foo in hfoo.
  rewrite hfoo in Hsubprog'; clear hfoo foo.
  simpl in Hsubprog';  set (H := dismember_program _ _ Hsubprog'). 	 
  destruct H as (Hsubprog_lambda',Hsubprog_main'); clear Hsubprog'.
  rename Hsubprog_main' into Hsubprog_end'.

  (* [arg] <- 2                 allocate heap *)
  intros; unfold memory_specification; simpl. 
  focus 6.
  instr_moveimm start start' 0 Hsubprog_env Hsubprog_env' with nat 2 2.
  (* [ret] <- label + 3 *)
  focus 7.
  instr_moveimm start start' 1 Hsubprog_env Hsubprog_env' with nat (start+3) (start'+3).
  (* jmp alloc *)
  instr_jumpcst start start' 2 Hsubprog_env Hsubprog_env'.
  (* reorganized to apply the allocator*)
  focus 9;focus 7;focus 7;focus 6.
  do 3 rewrite stateRelTensor_assoc; isolate_conj; clear Block; rename Conj_Block into Rc_alloc.
  unfold SPre_alloc, unchange_stack in Halloc.
  assert (Halloc' := Halloc Rc_alloc 2 2 stack_ptr stack_ptr' n n').
  assert (Halloc2 := total_to_partial _ _ _ _ _ _ _ Halloc').
  focus 1;eapply continuation_enrichment  with (Post := SPost_alloc Ra Rc_alloc 2 2 stack_ptr stack_ptr' n n');
  [idtac | eapply judgment_leq; try apply Halloc2;ineq_ring ]; clear Halloc' Halloc2.
  unfold SPost_alloc, unchange_stack.
  rewrite <- existsJudgment; intro fb.
  rewrite <- existsJudgment; intro fb'...
  focus 1 ...
  focus 1.
  omega_replace (fb+0) with fb; omega_replace (fb'+0) with fb'.
  unfold Rc_alloc ;clear Rc_alloc...
  unfold closure_rec_make in *.
  focus 13;focus 2;focus 3.
  (* [[ret]+1] <- [env] *)
  instr_moveindo start start' 3 Hsubprog_env Hsubprog_env' with nat n n' on values 0 0 0 0.
  (* [env] <- [ret] *)
  focus 2.
  instr_moveimm start start' 4 Hsubprog_env Hsubprog_env' with nat fb fb'.
  (* apply the lemma giving the shape of the heap after closure_make *)
  focus 1; rewrite ptsto_blank_leq.
  assert (Hstack_free1 : snd(fst code1)  <= stack_free); [  eapply le_trans; try eassumption;
  eapply le_trans; try apply (le_max_l (snd(fst code1))); trivial | idtac].
  assert (Hstack_free1' : snd(fst code1')  <= stack_free'); [  eapply le_trans; try eassumption;
  eapply le_trans; try apply (le_max_l (snd(fst code1'))); trivial | idtac].
  assert (HypRec1 := correction_of_lambda_main base base' Ra alloc alloc' dealloc dealloc' env e).
  unfold statement_of_compiler_correctness in *.
  set (code_of_e := compile e aux_code (base+length aux_code) alloc dealloc).
  set (code_of_e' := compile e aux_code' (base'+length aux_code') alloc' dealloc').
  hreplace HypRec1 with (HypRec1 (({fb+1,fb'+1}|->{n,n'})##({fb,fb'}|->-)##Rc) Rc_cloud (start+5) (start'+5) stack_free stack_free' stack_ptr stack_ptr' n fb n2 n3 n' fb' n2' n3' aux_code aux_code' code1 code1' p p' stack_list code_of_e code_of_e' code1_gen code1_gen');
  fold code1 code1_gen code_of_e code1' code1_gen' code_of_e' in HypRec1.
  hreplace HypRec1 with (HypRec1 (refl_equal code1_gen) (refl_equal code1) (refl_equal code_of_e) Hstack_free1 (refl_equal code1_gen') (refl_equal code1') (refl_equal code_of_e') Hstack_free1' Hsubprog_lambda Hsubprog_lambda' Halloc).
  unfold memory_specification in HypRec1; simpl in HypRec1.
  eapply judgment_leq; try apply HypRec1;  clear HypRec1; try ineq_ring.
  
  (* [[env]] <- [[sp]]          push the ptr to the function  into the hole  *)
  intros.
  set (new_start := start +5+length(fst (fst code1))) in  *...
  set (new_start' := start' +5+length(fst (fst code1'))) in  *...
  focus 2;focus 3;focus 13;focus 6.
  instr_moveind new_start new_start' 0 Hsubprog_end Hsubprog_end' with nat ptr ptr'.
  (* [env] <- [[env]+1]         restore the former environment *)
  focus 13;focus 1.
  instr_moveimmindo new_start new_start' 1 Hsubprog_end Hsubprog_end' with nat n n' on values envreg envreg fb fb'.

  (* induction on the shape of the type of lambda *)
  assert (Hlambda := shape_of_lambda t (refl_equal (LAMBDA e))).
  hreplace HypRec with (HypRec Rc Rc_cloud (start+5) (start'+5) stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' code1 code1' p p' stack_list code1_gen code1_gen').
  fold code1 code1_gen code1' code1_gen' in HypRec.
  hreplace HypRec with (HypRec (refl_equal code1_gen) (refl_equal code1) Hstack_free1 (refl_equal code1_gen') (refl_equal code1') Hstack_free1' Hsubprog_aux Hsubprog_lambda Hsubprog_aux' Hsubprog_lambda' Halloc Hdealloc).
  destruct HypRec as (HypRec_useless,HypRec); clear HypRec_useless.
  hreplace HypRec with (HypRec e (refl_equal (LAMBDA e))).
  clear - HypRec Hlambda t hyp hcode hcode' code_of_e code_of_e'.
  destruct Hlambda.

  (* dealing with the ARROW case, this is just applying the extra property *)

  unfold memory_specification in hyp; rewrite hcode in hyp; rewrite hcode' in hyp; simpl in hyp.
  repeat rewrite app_length in hyp; unfold closure_rec_make in hyp; simpl in hyp.
  replace (start+S(S(S(S(S(length (fst(fst code1)) +2)))))) with (new_start +2) in hyp; try (unfold new_start; omega).
  replace (start'+S(S(S(S(S(length (fst(fst code1')) +2)))))) with (new_start' +2) in hyp; try (unfold new_start'; omega).
  hreplace hyp with (hyp ptr ptr'); normal_form in hyp.
  replace (stack_ptr +1+1) with (stack_ptr+2) in *; try omega.
  replace (stack_ptr'+1+1) with (stack_ptr'+2) in hyp; try omega.
  simpl in hyp.
  focus_hyp 5; rewrite <- cloudExP_conj in hyp;  rewrite <- stateExP_stack in hyp;  rewrite <- existsJudgment  in hyp.
  set (fcode := snd code_of_e) in *.
  set (fcode' := snd code_of_e') in *.
  hreplace hyp with (hyp (({ptr,ptr'}||->{fcode,fcode'}) //\\ ({ptr+1,ptr'+1}||->{fb,fb'}) //\\ ({fb,fb'} ||-> {ptr,ptr'}) //\\ ({fb+1,fb'+1}||->{n,n'}) //\\  [||env||] Ra n n')).
  normal_form_conj in hyp.
  focus 9;focus 2.
  rewrite stateRelTensor_assoc; rewrite stack_description_strength; rewrite stateCloudPtsot.
  focus 2; rewrite stateRelTensor_assoc; rewrite stack_description_strength; rewrite stateCloudPtsot.
  focus 4;rewrite stateRelTensor_assoc; rewrite stack_description_strength; rewrite stateCloudPtsot.
  focus 4;rewrite stateRelTensor_assoc; rewrite stack_description_strength; rewrite stateCloudPtsot.
  isolate_conj.
  focus_conj_hyp 4.
  rewrite (cloudConjContraction ([||env||] Ra n n') ([||env||] Ra n n')) in Block; try reflexivity.
  focus_conj_hyp 1.
  rewrite (cloudConjContraction ({ptr,ptr'}||-> {fcode,fcode'}) ({ptr,ptr'}||-> {fcode,fcode'})) in Block; try reflexivity.
  normal_form_conj in Block.
  focus_conj_hyp 6;focus_conj_hyp 6;focus_conj_hyp 6.
  focus_conj_hyp 4;focus_conj_hyp 4.
  normal_form_conj in Block.
  rebuilt_conj Conj_Block Block.
  replace (stack_ptr+1-1) with stack_ptr in hyp; try omega.
  replace (stack_ptr'+1-1) with stack_ptr' in hyp; try omega.
  eapply continuation_enrichment_stack;
  [idtac| rewrite ptsto_later_c; eapply judgment_leq; try apply hyp ; eapply ppredtensorcompat2 ;try apply stateRelLeq_refl; ineq_ring].
  clear hyp; apply recursion_continuation.
  unfold fcode, code_of_e, fcode', code_of_e'.
  eapply judgment_leq; [idtac |  eapply HypRec with (ptr:=ptr)(ptr':=ptr')(fb:=fb)(fb':=fb'); try reflexivity].
  apply prearrowcompat2; unfold semantics_of_env; fold semantics_of_env;try reflexivity.
  match goal with
    | |- cloudRelLeq ?q (?p1 //\\ ?p2) =>
      setoid_rewrite (cloudRelConj_comm p1 p2)
    end.
  rewrite <- cloudExP_conj; apply cloudExistsLeq_single; exists ptr.
  rewrite <- cloudExP_conj; apply cloudExistsLeq_single; exists ptr'.
  rewrite <- cloudExP_conj; apply cloudExistsLeq_single; exists n.
  rewrite <- cloudExP_conj; apply cloudExistsLeq_single; exists n'.
  ineq_ring_conj.
Qed.
