(* correction_functions.v

   This file contains the two splitted properties of a function.
   The first one says that what is the shape of the stack 
   after executing closure_make.
   The second (splitted in two again but this time because of coq efficiency)
   certifies that the code the stack is pointing to satisfies the good property.

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import correction_compiler_definition.

Typeclasses eauto := dfs 10000.

Lemma correction_of_lambda_main : forall base base' Ra alloc alloc' dealloc dealloc' env e Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n fb n2 n3 n' fb' n2' n3',
     forall aux_code aux_code' code code' p p' stack_list code_of_e code_of_e' global_code global_code'
     (hprog : global_code = compile (LAMBDA e) aux_code (base+length aux_code) alloc dealloc)
     (hcode : code = (fst(fst global_code)) start)
     (hsubcode : code_of_e = compile e aux_code (base+length aux_code) alloc dealloc)
     (hstack : snd(fst code) <= stack_free)
     (hprog' : global_code' = compile (LAMBDA e) aux_code' (base'+length aux_code') alloc' dealloc')
     (hcode' : code' = (fst(fst global_code')) start')
     (hsubcode' : code_of_e' = compile e aux_code' (base'+length aux_code') alloc' dealloc')
     (hstack' : snd(fst code') <= stack_free'),
     PfunctionExtendsFunction p (fragfromprog (progfromlist (fst (fst code))) start)  ->
     PfunctionExtendsFunction p' (fragfromprog (progfromlist (fst (fst code'))) start')  ->
     (forall Rc_alloc size size' n5 n5' n6 n6', total_judgment (SPre_alloc Ra Rc_alloc size size' n5 n5' n6 n6')
            (SPost_alloc Ra Rc_alloc size size' n5 n5' n6 n6') p p' alloc alloc') ->
     (forall ptr ptr', judgment (({ptr+1,ptr'+1}|->{fb,fb'}) ## ({ptr,ptr'}|-> {(snd code_of_e), (snd code_of_e')})
       ## ({stack_ptr + 1,stack_ptr' + 1} |-> {ptr,ptr'})
       ## (spreg|->{stack_ptr+1,stack_ptr'+1}) ## Block (stack_ptr+2) (stack_ptr'+2) (stack_free-1) (stack_free'-1) ## (envreg|-> {fb,fb'}) ##  storing_space n n2 n3 n' n2' n3'
       ## stack_description stack_list stack_ptr stack_ptr' ([||env||] Ra n n' //\\ Rc_cloud)
       ## (workreg |-> -) ## (argreg |-> -) ## (retreg |-> -) ## Ra ## Rc
       ## (3|->-) ## (4|->-) ## unused_space)
     p p' (start+List.length (fst(fst code))) (start'+List.length (fst(fst code'))))->
     judgment ((spreg|->{stack_ptr,stack_ptr'}) ## Block (stack_ptr+1) (stack_ptr'+1) stack_free stack_free' ## (envreg|-> {fb,fb'}) ##
             storing_space n n2 n3 n' n2' n3' ## stack_description stack_list stack_ptr stack_ptr' ([||env||] Ra n n' //\\ Rc_cloud)
          ## (workreg |-> -) ## (argreg |-> -) ## (retreg |-> -) ## Ra ## Rc
	  ## (3|->-) ## (4|->-) ## unused_space) p p' start start'.
Proof with normal_form.
  intros until global_code'.
  intros hprog hcode hsubcode Hstack hprog' hcode' hsubcode' Hstack' Hsubprog Hsubprog' Halloc hyp.

  simpl in hprog; simpl in hprog'.
  (* recover the actual piece of code *)
  surjective_pairing_function e aux_code base alloc dealloc in hprog.
  surjective_pairing_function e code1_gen in hprog.
  rewrite hprog in hcode; simpl in hcode.
  rewrite hsubcode in hyp; fold code1_gen in hyp; clear hsubcode code_of_e.
  match goal with | h:global_code = ?code |- _ =>
    set (code_next := snd code) end.
  simpl in code_next; clear hprog.
  unfold closure_make in hcode; simpl in hcode.
  unfold memory_specification.
  rewrite hcode in Hstack; simpl in Hstack.
  rewrite hcode in Hsubprog; simpl in Hsubprog.

  (* Duplicated code, not very nice *)
  surjective_pairing_function e aux_code' base' alloc' dealloc' in hprog'.
  surjective_pairing_function e code1_gen0 in hprog'.
  rewrite hprog' in hcode'; simpl in hcode'.
  rewrite hsubcode' in hyp; fold code1_gen0 in hyp; clear hsubcode' code_of_e'.
  match goal with | h:global_code' = ?code |- _ =>
    set (code_next' := snd code') end.
  simpl in code_next'; clear hprog'.
  unfold closure_make in hcode'; simpl in hcode'.
  unfold memory_specification.
  rewrite hcode' in Hstack'; simpl in Hstack'.
  rewrite hcode' in Hsubprog'; simpl in Hsubprog'.
  (* [arg] <- 2*)
  focus 6.
  instr_moveimm start start' 0 Hsubprog Hsubprog' with nat 2.
  (* [ret] <- label + 3 *)
  focus 7.
  instr_moveimm start start' 1 Hsubprog Hsubprog' with nat (start+3) (start'+3).
  (* jmp alloc *)
  instr_jumpcst start start' 2 Hsubprog Hsubprog'.
  (* reorganized to apply the allocator*)
  focus 3;focus 5;focus 6;focus 9.
  do 3 rewrite stateRelTensor_assoc.
  match goal with
  | |- judgment (?P##?Q) ?p ?p' ?l ?l' =>  set (Conj_Block := P);
      assert (Block : stateRelEq Conj_Block Conj_Block); try ring
  end;
  unfold Conj_Block at 1 in Block.
  clear Block; rename Conj_Block into Rc_alloc.
  focus 1;eapply continuation_enrichment  with (Post := SPost_alloc Ra Rc_alloc 2 2 stack_ptr stack_ptr' fb fb').
  Focus 2.
  specialize (Halloc Rc_alloc 2 2 stack_ptr stack_ptr' fb fb').
  unfold SPre_alloc in Halloc.
  eapply judgment_leq.
  Focus 2.
  eapply total_to_partial.
  eassumption.
  Focus 2.
  unfold unchange_stack, natPconstant.
  ineq_ring.

  unfold SPost_alloc, unchange_stack.
  rewrite <- existsJudgment; intro fb2.
  rewrite <- existsJudgment; intro fb2' ...
  normal_form.
  omega_replace (fb2+0) with fb2; omega_replace (fb2'+0) with fb2'.
  focus 1; focus 2.
  focus 9.
  (* [sp] <- [sp]+1 shift the stack pointer *)
  instr_opimmleft plus start start' 3 Hsubprog Hsubprog' stack_ptr stack_ptr' 1 1.
  assert (Hstack1 : S (stack_free - 1) = stack_free); try omega.
  assert (Hstack1' : S (stack_free' - 1) = stack_free'); try omega.
  unfold Rc_alloc ;clear Rc_alloc...
  focus 8; rewrite <- Hstack1; rewrite <- Hstack1'.
  rewrite block_unfold_start...
  focus 14; do 6 rewrite stateRelTensor_assoc.
  hide_tail...
  focus 6;focus 2;focus 4.
  (* [ret] <- [[sp]] shift the stack pointer *)
  instr_moveind start start' 4 Hsubprog Hsubprog' with nat fb2 fb2'.
  
  focus 5;focus 3.
  (* [work] <- [[sp]] set code pointer *)
  instr_moveind start start' 5 Hsubprog Hsubprog' with nat (snd code1_gen) (snd code1_gen0).
  focus 4; focus 6;focus 2.
  omega_replace (S fb2) with (fb2 +1); omega_replace (S fb2') with (fb2'+1).
  (* [[ret]+1] <- [env] set pointer to rest of environment *)

  instr_moveindo start start' 6 Hsubprog Hsubprog' with nat fb fb' on values retreg retreg fb2 fb2'.
  (* applying hyp*)
  hreplace hyp with (hyp fb2 fb2'); rewrite hcode in hyp;rewrite hcode' in hyp; simpl in hyp.
  set (fcode := snd code1_gen) in *.
  set (fcode' := snd code1_gen0) in *.
  rewrite ptsto_blank_leq.
  replace (S(stack_ptr +1)) with (stack_ptr+2) by omega.
  replace (S(stack_ptr' +1)) with (stack_ptr'+2) by omega.
  unfold tail; clear tail.
  eapply judgment_leq; try apply hyp.
  ineq_ring.
Qed.

(* This lemma is here to make the proof below shorter, because setoid is making Coq crazy*)
Lemma correction_epilog : forall p p' stack_size stack_size' dealloc dealloc' ptr_a ptr_a' start start' Ra stack_ptr stack_ptr' n n' b Rc Rc_cloud a env ptr ptr' new_env new_env',
    (forall Rc_alloc size size' base base' n5 n5' n6 n6', total_judgment (SPre_dealloc Ra Rc_alloc base base' size size' n5 n5' n6 n6')
            (SPost_dealloc Ra Rc_alloc n5 n5' n6 n6') p p' dealloc dealloc') ->
    stack_size >=1 ->
    stack_size' >=1 ->
    PfunctionExtendsFunction p
    	(fragfromprog (progfromlist (closure_epilog start stack_size dealloc)) start) ->
    PfunctionExtendsFunction p'
    	(fragfromprog (progfromlist (closure_epilog start' stack_size' dealloc')) start') ->
    judgment (memory_specification Ra (({stack_ptr,stack_ptr'} |-> {ptr_a,ptr_a'}) ## ({stack_ptr+1,stack_ptr'+1}|->-) ##
    cloudToState (({n+4,n'+4} ||=> Later (Perp(Post_arrow ([|b|]) Ra Rc Rc_cloud n n' stack_ptr stack_ptr'))) //\\ stateToCloud Rc))
    Rc_cloud (a :: env) ((([|b|]) Ra, (ptr,ptr')) :: nil) (stack_size-1) (stack_size'-1) (new_env+5) (new_env'+5) new_env n stack_ptr new_env' n' stack_ptr') p p' start start'.
Proof with normal_form.
  intros until new_env'; intros Hdealloc Hsize Hsize' Hprog Hprog'; unfold memory_specification.
  unfold closure_epilog in *;  simpl.
  focus 4. focus 2. focus 6.
  (* [wk] <- [[sp]]         save return value *)
  instr_moveimmind start start' 0 Hprog Hprog' with nat ptr ptr' on values spreg spreg (new_env+5) (new_env'+5).
  unfold storing_space.
  do 7 rewrite stateRelTensor_assoc; hide_tail...
  focus 7;focus 6. focus 3.
  (* [sp] <- [[env]+3]      restore old sp *)
  instr_moveimmindo start start' 1 Hprog Hprog' with nat stack_ptr stack_ptr' on values envreg envreg new_env new_env'.
  unfold tail; clear tail; focus 12.
  do 8 rewrite stateRelTensor_assoc; hide_tail...
  (* [[sp]] <- [wk]         push return value *)
  focus 4;focus 1;focus 2.
  instr_moveind start start' 2 Hprog Hprog' with nat ptr ptr'.
  (* [wk] <- [env] +2        save current frame *)
  focus 3;focus 3.
  instr_opimmleft plus start start' 3 Hprog Hprog' new_env new_env' 2 2.
  (* [env] <- [[env]+2]     restore old frame *)
  focus 8;focus 2.
  instr_moveimmindo start start' 4 Hprog Hprog' with nat n n' on values envreg envreg new_env new_env'.
  (* [arg] <- stack_size+3  deallocate frame *)
  unfold tail; clear tail; focus 11;focus 11.
  rewrite stateRelTensor_assoc; hide_tail...
  instr_moveimm start start' 5 Hprog Hprog' with nat (stack_size+3) (stack_size'+3).
  (* [ret] <- label + 10*)
  focus 1.
  instr_moveimm start start' 6 Hprog Hprog' with nat (start+8) (start'+8).
  (* jmp dealloc *)
  instr_jumpcst start start' 7 Hprog Hprog'.
  unfold tail; clear tail.
  (* reorganized to apply the deallocator*)
  focus 10;focus 9; rewrite ptsto_blank_leq.
  omega_replace (new_env+5+1) with (S(new_env+5)).
  omega_replace (new_env'+5+1) with (S(new_env'+5)).
  rewrite stateRelTensor_assoc; rewrite <- block_unfold_start.
  omega_replace (S(stack_size -1)) with (stack_size).
  omega_replace (S(stack_size' -1)) with (stack_size').
  focus 10; rewrite ptsto_blank_leq.
  omega_replace (new_env+5) with (S(new_env+4));omega_replace (new_env'+5) with (S(new_env'+4)).
  rewrite stateRelTensor_assoc; rewrite <- block_unfold_start.
  omega_replace (S stack_size) with (stack_size+1).
  omega_replace (S stack_size') with (stack_size'+1).
  focus 8; rewrite ptsto_blank_leq.
  omega_replace (new_env+4) with (S(new_env+3));omega_replace (new_env'+4) with (S(new_env'+3)).
  rewrite stateRelTensor_assoc; rewrite <- block_unfold_start.
  omega_replace (S (stack_size+1)) with (stack_size+2).
  omega_replace (S (stack_size'+1)) with (stack_size'+2).
  focus 4; rewrite ptsto_blank_leq.
  omega_replace (new_env+3) with (S(new_env+2));omega_replace (new_env'+3) with (S(new_env'+2)).  
  rewrite stateRelTensor_assoc; rewrite <- block_unfold_start.
  omega_replace (S (stack_size+2)) with (stack_size+3).
  omega_replace (S (stack_size'+2)) with (stack_size'+3).
  focus 6;focus 7;focus 9;focus 10.
  do 3 rewrite stateRelTensor_assoc; isolate_conj; rename Conj_Block into Rc_alloc; clear Block.
  unfold SPre_dealloc, unchange_stack in Hdealloc.
  hreplace Hdealloc with (Hdealloc Rc_alloc (stack_size+3)(stack_size'+3) (new_env+2) (new_env'+2) stack_ptr stack_ptr' n n').
  assert (Hdealloc2 := total_to_partial _ _ _ _ _ _ _ Hdealloc).
  focus 2;eapply continuation_enrichment  with (Post := SPost_dealloc Ra Rc_alloc stack_ptr stack_ptr' n n');
   [idtac | eapply judgment_leq; try apply Hdealloc2; ineq_ring].

  (* jmp [[env]+4]          return to application code *)
  unfold SPost_dealloc, Rc_alloc. clear Hdealloc2 Hdealloc Rc_alloc.
  focus 6;  unfold unchange_stack;  focus 11.
  
  instr_jumpindo start start' 8 Hprog Hprog'.
  apply continuation_simplification_cloud.
  focus 1;  unfold Post_arrow.
  rewrite cloudRelConj_comm. rewrite cloudRelConj_weakening.
  rewrite <- bijection_state_pred1.
  focus 3;  isolate_conj.
  normal_form_conj in Block;  focus_conj_hyp 1.
  rewrite cloudRelConj_comm in Block.
  rebuilt_conj Conj_Block Block.
  rewrite cloudRelConj_weakening.
  apply stateExistsLeq; exists ptr.
  apply stateExistsLeq; exists ptr'...
  focus' 2; eapply ppredtensorcompat2; try ineq_ring.
Qed.

Lemma correction_of_lambda_aux : forall base base' Ra alloc dealloc alloc' dealloc' env e a b (t:a::env|--e:::b), 
     forall aux_code aux_code' code_of_e code_of_e' p p' global_code global_code'
     (hprog : global_code = compile (LAMBDA e) aux_code (base+length aux_code) alloc dealloc)
     (hprog' : global_code' = compile (LAMBDA e) aux_code' (base'+length aux_code') alloc' dealloc')
     (hsubcode : code_of_e = compile e aux_code (base+length aux_code) alloc dealloc)
     (hsubcode' : code_of_e' = compile e aux_code' (base'+length aux_code') alloc' dealloc'),
     PfunctionExtendsFunction p (fragfromprog (progfromlist (snd(fst global_code))) base)  ->
     PfunctionExtendsFunction p' (fragfromprog (progfromlist (snd(fst global_code'))) base')  ->
     (forall Rc_alloc size size' n5 n5' n6 n6',  total_judgment (SPre_alloc Ra Rc_alloc size size' n5 n5' n6 n6') 
            (SPost_alloc Ra Rc_alloc size size' n5 n5' n6 n6') p p' alloc alloc') ->
     (forall Rc_alloc size size' base base' n5 n5' n6 n6', total_judgment (SPre_dealloc Ra Rc_alloc base base' size size' n5 n5' n6 n6')
            (SPost_dealloc Ra Rc_alloc n5 n5' n6 n6') p p' dealloc dealloc') ->
     (forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3',
        statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' (a::env) b e t Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
     forall ptr ptr' fb fb', judgment (Pre_arrow ({ptr+1,ptr'+1}||->{fb,fb'} //\\ ([||env||]) Ra fb fb') ptr ptr' Ra ([|a|]) ([|b|])) p p' (snd code_of_e) (snd code_of_e').
Proof with normal_form.
  intros until global_code'.
  intros hprog hprog' hsubcode hsubcode' Hsubprog_aux Hsubprog_aux' Halloc Hdealloc HypRec ptr ptr' fb fb'.
  simpl in hprog; simpl in hprog'.
  (* recover the actual piece of code *)
  surjective_pairing_function e aux_code base alloc dealloc in hprog.
  surjective_pairing_function e code1_gen in hprog.
  surjective_pairing_function e aux_code' base' alloc' dealloc' in hprog'.
  surjective_pairing_function e code1_gen0 in hprog'.
  rewrite hsubcode; fold code1_gen; clear hsubcode code_of_e.
  rewrite hprog in Hsubprog_aux;  simpl in Hsubprog_aux.
  rewrite hsubcode'; fold code1_gen0; clear hsubcode' code_of_e'.
  rewrite hprog' in Hsubprog_aux';  simpl in Hsubprog_aux'.
  match goal with | h:global_code = ?code |- _ =>
    set (code_next := snd code) end.
  simpl in code_next; clear hprog.
  match goal with | h:global_code' = ?code |- _ =>
    set (code_next' := snd code) end.
  simpl in code_next'; clear hprog'.
  unfold memory_specification.
  (* split the auxiliary code into aux_code, prolog, code of e, epilog*)
  set (H := dismember_program _ _ Hsubprog_aux).
  destruct H as (Hsubprog_aux1,Hsubprog_aux2); clear Hsubprog_aux.
  set (H := dismember_program _ _ Hsubprog_aux').
  destruct H as (Hsubprog_aux1',Hsubprog_aux2'); clear Hsubprog_aux'.
  set (new_start := base + length (snd (fst code1_gen))) in *.
  set (new_start' := base' + length (snd (fst code1_gen0))) in *.
  match goal with | h: PfunctionExtendsFunction ?p (fragfromprog (progfromlist ?l) new_start) |- _ =>  set (ugly_list := l) in * end.
  assert (ugly_list = (fuse (takedrop 11 nil ugly_list))); try (simpl;trivial).
  Opaque fuse. simpl in H. Transparent fuse.
  unfold fuse in H; rewrite H in Hsubprog_aux2; clear H ugly_list.
  match goal with | h: PfunctionExtendsFunction ?p (fragfromprog (progfromlist ?l) new_start') |- _ =>  set (ugly_list := l) in * end.
  assert (ugly_list = (fuse (takedrop 11 nil ugly_list))); try (simpl;trivial).
  Opaque fuse. simpl in H. Transparent fuse.
  unfold fuse in H; rewrite H in Hsubprog_aux2'; clear H ugly_list.
  set (H := dismember_program _ _ Hsubprog_aux2).
  destruct H as (foo,Hsubprog_aux); clear Hsubprog_aux2; rename foo into Hsubprog_aux2.
  simpl in Hsubprog_aux.
  set (H := dismember_program _ _ Hsubprog_aux2').
  destruct H as (foo,Hsubprog_aux'); clear Hsubprog_aux2'; rename foo into Hsubprog_aux2'.
  simpl in Hsubprog_aux'.
  set (new_start2 := new_start +11) in *.
  set (new_start2' := new_start' +11) in *.
  set (H := dismember_program _ _ Hsubprog_aux).
  destruct H as (Hsubprog_aux3,Hsubprog_aux4); clear Hsubprog_aux.
  set (H := dismember_program _ _ Hsubprog_aux').
  destruct H as (Hsubprog_aux3',Hsubprog_aux4'); clear Hsubprog_aux'.
  unfold code1,code1_gen in Hsubprog_aux4; rewrite compile_length_main in Hsubprog_aux4.
  fold code1_gen code1 in Hsubprog_aux4; unfold code1_gen in Hsubprog_aux4;  rewrite compile_length_aux in Hsubprog_aux4.
  unfold code0,code1_gen0 in Hsubprog_aux4'; rewrite compile_length_main in Hsubprog_aux4'.
  fold code1_gen0 code0 in Hsubprog_aux4'; unfold code1_gen0 in Hsubprog_aux4';  rewrite compile_length_aux in Hsubprog_aux4'.
  unfold closure_prolog_size in Hsubprog_aux4; fold code1_gen code1 new_start new_start2 in Hsubprog_aux4.
  unfold closure_prolog_size in Hsubprog_aux4'; fold code1_gen0 code0 new_start' new_start2' in Hsubprog_aux4'.
  assert (HstackE : snd(fst code1) >= 1); [unfold code1,code1_gen; apply stack_size_positive | idtac].
  unfold code1_gen; rewrite compile_length_aux; fold code1_gen; fold new_start.
  assert (HstackE' : snd(fst code0) >= 1); [unfold code0,code1_gen0; apply stack_size_positive | idtac].
  unfold code1_gen0; rewrite compile_length_aux; fold code1_gen0; fold new_start'.
  unfold Pre_arrow.

  rewrite  <- existsJudgment; intro Rc.
  rewrite  <- existsJudgment; intro Rc_cloud.
  rewrite  <- existsJudgment; intro n.
  rewrite  <- existsJudgment; intro n'.
  rewrite  <- existsJudgment; intro ptr_a.
  rewrite  <- existsJudgment; intro ptr_a'.
  rewrite  <- existsJudgment; intro stack_ptr.
  rewrite  <- existsJudgment; intro stack_ptr'...

  replace (snd(fst code1)+0+5) with (snd(fst code1)+5) in Hsubprog_aux2 by omega.
  replace (snd(fst code0)+0+5) with (snd(fst code0)+5) in Hsubprog_aux2' by omega.

(** Prolog  **)
  
  (* [arg] <- stack_size + 5  allocate stack frame*)
  focus 5.  
  instr_moveimm new_start new_start' 0 Hsubprog_aux2 Hsubprog_aux2' 
    with nat (snd(fst code1) + 5) (snd(fst code0) + 5).
  (* [ret] <- label + 3 *)
  focus 6.
  instr_moveimm new_start new_start' 1 Hsubprog_aux2 Hsubprog_aux2' 
    with nat (snd code1_gen +3) (snd code1_gen0 +3).
  (* jmp alloc *)
  instr_jumpcst new_start new_start' 2 Hsubprog_aux2 Hsubprog_aux2'.
  (* reorganized to apply the allocator*)
  focus 2;focus 3;focus 4;focus 5.
  do 3 rewrite stateRelTensor_assoc; isolate_conj; clear Block; rename Conj_Block into Rc_alloc.
  unfold SPre_alloc, unchange_stack in Halloc.
  focus 1.
  assert (Halloc' := Halloc Rc_alloc  (snd(fst code1)+5) (snd(fst code0)+5) (stack_ptr+1) (stack_ptr'+1) n n').
  assert (Halloc2 := total_to_partial _ _ _ _ _ _ _ Halloc'); clear Halloc'.
  eapply continuation_enrichment  with (Post := SPost_alloc Ra Rc_alloc (snd(fst code1)+5) (snd(fst code0)+5) (stack_ptr+1) (stack_ptr'+1) n n');
  [idtac | eapply judgment_leq; try apply Halloc2 ;ineq_ring ].
  unfold SPost_alloc, unchange_stack; clear Halloc2.
  rewrite <- existsJudgment; intro new_env.
  rewrite <- existsJudgment; intro new_env'...
  unfold code1_gen; rewrite compile_length_aux; fold code1_gen new_start...
  unfold code1_gen0; rewrite compile_length_aux; fold code1_gen0 new_start...
  focus 1.
  omega_replace (snd (fst code1)+5) with (S(snd (fst code1)+4)).
  omega_replace (snd (fst code0)+5) with (S(snd (fst code0)+4)).
  rewrite block_unfold_start ;normal_form;focus 1.
  omega_replace (snd (fst code1)+4) with (S(snd (fst code1)+3)).
  omega_replace (snd (fst code0)+4) with (S(snd (fst code0)+3)).
  rewrite block_unfold_start ;normal_form;focus 1.
  omega_replace (snd (fst code1)+3) with (S(snd (fst code1)+2)).
  omega_replace (snd (fst code0)+3) with (S(snd (fst code0)+2)).
  rewrite block_unfold_start ;normal_form;focus 1.
  omega_replace (snd (fst code1)+2) with (S(snd (fst code1)+1)).
  omega_replace (snd (fst code0)+2) with (S(snd (fst code0)+1)).
  rewrite block_unfold_start ;normal_form;focus 1.
  omega_replace (snd (fst code1)+1) with (S(snd (fst code1))).
  omega_replace (snd (fst code0)+1) with (S(snd (fst code0))).
  rewrite block_unfold_start ;normal_form;focus 1.
  do 5 rewrite stateRelTensor_assoc.
  match goal with | |- judgment (?P##?Q) ?p ?p' ?l ?l' => set (block_env := P)  end.
  unfold Rc_alloc ...
  (* [wk] <- [[sp]]           take function pointer *)
  focus 5;focus 11;focus 9.
  fold new_start'.
  instr_moveimmind new_start new_start' 3 Hsubprog_aux2 Hsubprog_aux2' with nat ptr ptr' on values spreg spreg (stack_ptr +1) (stack_ptr' +1).  
  unfold block_env...
  focus 8;focus 8;focus 8;focus 9;focus 13; focus 13.
  do 8 rewrite stateRelTensor_assoc; hide_tail...
  normal_form_conj.
  (* [[ret]+1] <- [[wk]+1]    store environment pointer *)
  focus 6;focus 5;focus 4.
  omega_replace (S new_env) with (new_env +1).
  omega_replace (S new_env') with (new_env' +1).

  instr_moveindo new_start new_start' 4 Hsubprog_aux2 Hsubprog_aux2' with nat fb fb' on values workreg workreg ptr ptr'.
  (* [sp] <- [sp]-1           focus on argument *)
  focus 7.
  instr_opimmleft minus new_start new_start' 5 Hsubprog_aux2 Hsubprog_aux2' (stack_ptr+1) (stack_ptr'+1) 1 1.
  omega_replace (stack_ptr +1-1) with stack_ptr.
  omega_replace (stack_ptr' +1-1) with stack_ptr'.
  (* [[ret]] <- [[sp]]        store argument's value in envrionment *)
  focus 5;focus 1;focus 7;focus 3.
  instr_moveind new_start new_start' 6 Hsubprog_aux2 Hsubprog_aux2' with nat ptr_a ptr_a'.
  unfold tail; clear tail.
  focus 7;focus 11;focus 17. 
  do 5 rewrite stateRelTensor_assoc; hide_tail...
  (* [[ret]+2] <- [env]       store calling env *)
  focus 2;focus 3.
  omega_replace (S(new_env+1)) with (new_env+2).
  omega_replace (S(new_env'+1)) with (new_env'+2).
  instr_moveindo new_start new_start' 7 Hsubprog_aux2 Hsubprog_aux2' with nat n n' on values 0 0 0 0 .
  (* [[ret]+3] <- [sp]        store calling sp *)
  focus 5;focus 4;focus 2.
  omega_replace (S(S(S new_env))) with (new_env+3). 
  omega_replace (S(S(S new_env'))) with (new_env'+3). 
  instr_moveindo new_start new_start' 8 Hsubprog_aux2 Hsubprog_aux2' with nat stack_ptr stack_ptr' on values 0 0 0 0.
  (* [env] <- [ret]           switch to new environment *)
  focus 4.
  instr_moveimm new_start new_start' 9 Hsubprog_aux2 Hsubprog_aux2' with nat new_env new_env'.
  (* [sp] <- [ret] + 5        switch to new stack *)
  focus 1;focus 3.
  instr_opimmleft plus new_start new_start' 10 Hsubprog_aux2 Hsubprog_aux2' new_env new_env' 4 4.
  
  set (closure_entry := snd code1_gen) in *.
  set (closure_entry' := snd code1_gen0) in *.
  unfold tail; clear tail block_env; focus 7; focus 9;focus 12 ;do 8 rewrite stateRelTensor_assoc; hide_tail...
  omega_replace (S(S(S(S new_env)))) with (new_env+4).
  omega_replace (S(S(S(S new_env')))) with (new_env'+4).

(** applying HypRec **)
  
  focus 1;focus 2;  rewrite stateRelTensor_assoc; rewrite TensorConj_strength; rewrite stateCloudPtsot.
  focus 7; rewrite stateRelTensor_assoc; rewrite TensorConj_strength ; rewrite stateCloudPtsot.
  focus 1;  focus 7;  focus 6.
  do 2 rewrite stateRelTensor_assoc; isolate_conj.
  normal_form in Block; fold (storing_space new_env n stack_ptr new_env' n' stack_ptr') in *.
  rebuilt_conj Conj_Block Block.
  focus 1; isolate_conj.
  focus_conj_hyp 3;focus_conj_hyp 4;  focus_conj_hyp 3; focus_conj_hyp 3.
  do 3 rewrite cloudRelConj_assoc in Block.
  focus_conj_hyp 1; rewrite cloudRelConj_comm  in Block.
  rebuilt_conj Conj_Block Block.
  rewrite cloudRelConj_weakening.
  unfold statement_of_compiler_correctness in HypRec.
  unfold closure_entry, closure_prolog_size in code1.
  unfold closure_entry', closure_prolog_size in code0.
  (* instantiating Pprivate and stuff like that *)   
  hreplace HypRec with (HypRec (({stack_ptr,stack_ptr'}|->{ptr_a,ptr_a'})##({stack_ptr+1,stack_ptr'+1}|->-)##
  cloudToState (({n+4,n'+4} ||=> Later(Perp (Post_arrow ([|b|]) Ra Rc Rc_cloud n n'
  stack_ptr stack_ptr'))) //\\ stateToCloud Rc))
  Rc_cloud (snd code1_gen + 11) (snd code1_gen0 + 11) (snd (fst code1)) (snd (fst code0)) (new_env+4) (new_env'+4) new_env n stack_ptr new_env' n' stack_ptr' aux_code aux_code' code1 code0 p p' nil code1_gen code1_gen0).
    unfold prog_extends_code, prog_extends_auxcode, extract_code_from_globalcode, extract_stacksize_from_code in HypRec.
  unfold closure_prolog_size, closure_entry, closure_entry' in *.
  fold code1 code1_gen code0 code1_gen0 in HypRec; unfold code1_gen,code1_gen0 in HypRec.
  repeat rewrite compile_length_aux in HypRec;  fold code1_gen code1_gen0 in HypRec.
  hreplace HypRec with (HypRec (refl_equal code1_gen) (refl_equal code1) (le_refl (snd(fst code1))) (refl_equal code1_gen0) (refl_equal code0) (le_refl (snd(fst code0))) Hsubprog_aux1 Hsubprog_aux3 Hsubprog_aux1' Hsubprog_aux3' Halloc Hdealloc).
  destruct HypRec as (HypRec,foo); clear foo.
  eapply judgment_leq; try apply HypRec...
  clear HypRec; unfold memory_specification.
  simpl; unfold tail; clear tail...
  focus' 4.
  rewrite <- cloudExP_conj; rewrite <- cloudExP_extract; apply stateExistsLeq; exists ptr_a.  rewrite <- cloudExP_conj; rewrite <- cloudExP_extract; apply stateExistsLeq; exists ptr_a'.
  rewrite <- cloudExP_conj; rewrite <- cloudExP_extract; apply stateExistsLeq; exists fb.
  rewrite <- cloudExP_conj; rewrite <- cloudExP_extract; apply stateExistsLeq; exists fb'.
  eapply ppredtensorcompat2.
  apply cloudtostatecompat2; ineq_ring_conj.
  focus 2; rewrite ptsto_blank_leq.
  focus 3; setoid_rewrite ptsto_blank_leq at 1.
  omega_replace (new_env+4+1) with (S(S(S(S(S(new_env)))))).
  omega_replace (new_env'+4+1) with (S(S(S(S(S(new_env')))))).
  focus 13;rewrite ptsto_blank_leq.
  unfold storing_space;  ineq_ring.

(** Epilog **)

  clear HypRec Hsubprog_aux1 Hsubprog_aux2 Hsubprog_aux3  Hsubprog_aux1' Hsubprog_aux2' Hsubprog_aux3' Halloc.
  intros ptr0 ptr0'; unfold new_start2,new_start2' in *.
  fold new_start new_start'.
  clear new_start2 new_start2'.
  set (new_start3 := new_start + 11 + length (fst(fst code1))) in *.
  set (new_start3' := new_start' + 11 + length (fst(fst code0))) in *.
  clear tail.
  omega_replace (new_env +4 +1) with (new_env+5).
  omega_replace (new_env' +4 +1) with (new_env'+5).
  replace (snd(fst code1)+0) with (snd(fst code1)) in Hsubprog_aux4 ; try omega.
  replace (snd(fst code0)+0) with (snd(fst code0)) in Hsubprog_aux4' ; try omega.
  apply (correction_epilog p p' (snd(fst code1)) (snd(fst code0)) dealloc dealloc'); assumption.
Qed.