(* correction_gt.v

   This is the partial correctness proof for comparison for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import correction_compiler_definition.

Typeclasses eauto := dfs 10000.

Lemma correction_of_gt : forall base base' Ra alloc alloc' dealloc dealloc' env P1 P2 e1 e2 (t1:env|--e1:::Int P1) (t2:env|--e2:::Int P2),
(forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env (Int P1) e1 t1 Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
(forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env (Int P2) e2 t2 Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env (Bool(AbstractGt P1 P2)) (GT e1 e2) (TGT t1 t2) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3'.
Proof with normal_form.
  intros until t2; intros HypRecE1 HypRecE2; intros. 
  unfold statement_of_compiler_correctness in *.
  unfold prog_extends_code, prog_extends_auxcode, extract_code_from_globalcode, extract_stacksize_from_code in *;
  intros until global_code'; intros hprog hcode Hstack hprog' hcode' Hstack' Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc.
  split ; [idtac | intros e' Hlambda; inversion Hlambda]; intro hyp.

  (* recover the actual piece of code *)

  simpl in hprog; surjective_pairing_two_arg e1 e2 aux_code base alloc dealloc in hprog;
  rewrite <- hcode in Hsubprog; rewrite <- hcode' in Hsubprog';        
  rewrite hprog in hcode; simpl in hcode;
  surjective_pairing_two_arg e1 e2 code1_gen code2_gen start in hcode.
  simpl in hprog'; surjective_pairing_two_arg e1 e2 aux_code' base' alloc' dealloc' in hprog';
  rewrite hprog' in hcode'; simpl in hcode';
  rename code2_gen0 into code2_gen'; rename code1_gen0 into code1_gen';
  surjective_pairing_two_arg e1 e2 code1_gen' code2_gen' start' in hcode'.

  unfold pair_code in hcode; simpl in hcode.
  rewrite hprog in Hsubprog_aux; simpl in Hsubprog_aux; clear hprog.
  rewrite hcode in Hstack; simpl in Hstack.
  unfold pair_code in hcode'; simpl in hcode'.
  rewrite hprog' in Hsubprog_aux'; simpl in Hsubprog_aux'; clear hprog'.
  rewrite hcode' in Hstack'; simpl in Hstack'.
  rename code0 into code2'; rename code3 into code1'.
  replace (snd(fst code2)+0) with (snd(fst code2)) in *; try omega.
  replace (snd(fst code2')+0) with (snd(fst code2')) in *; try omega.
  assert (Hmin : min (snd(fst code2)) 1 = 1); try (apply min_r;  unfold code2,code2_gen; apply stack_size_positive).
  rewrite Hmin in Hstack.
  assert (max (snd (fst code2)) (snd(fst code1) +1) <= stack_free); [eapply le_trans ; try apply Hstack; apply le_max_l | idtac].
  clear Hmin Hstack; rename H into Hstack.
  assert (Hmin' : min (snd(fst code2')) 1 = 1); try (apply min_r;  unfold code2',code2_gen'; apply stack_size_positive).
  rewrite Hmin' in Hstack'.
  assert (max (snd (fst code2')) (snd(fst code1') +1) <= stack_free'); [eapply le_trans ; try apply Hstack'; apply le_max_l | idtac].
  clear Hmin' Hstack'; rename H into Hstack'.

  (* apply each argument *)

  apply_arg_e2 code2_gen code1_gen code2 code2_gen' code1_gen' code2' hcode hcode' stack_free stack_free' p p' start base start' base' Rc Rc_cloud stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' stack_list  Halloc Hdealloc Hsubprog Hsubprog_aux Hsubprog' Hsubprog_aux' HypRecE2 Hstack Hstack'.

  intros ptr_b ptr_b'.

  apply_arg_e1 code2_gen code1_gen code2 code1 code2_gen' code1_gen'
  code2' code1' hcode hcode' stack_free stack_free' p p' start base
  start' base' Ra Rc Rc_cloud stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code
  aux_code' stack_list Halloc Hdealloc Hsubprog
  Hsubprog_aux Hsubprog' Hsubprog_aux' (Int P2) ptr_b ptr_b' HypRecE1.

  match goal with | h : code = (?l1++?l,?stack,?next) |- _ => set (code_pair := l) end; rewrite hcode in Hsubprog.
  match goal with | h : code' = (?l1++?l,?stack,?next) |- _ => set (code_pair' := l) end; rewrite hcode' in Hsubprog'.
  assert (HsubprogGt : PfunctionExtendsFunction p (fragfromprog (progfromlist (code_pair)) (start+length (fst (fst code2))+length (fst (fst code1))) )).
  eapply subPfunction_extends; try apply Hsubprog.
  rewrite <- sub_is_extends; simpl;rewrite <- plus_assoc; rewrite <- app_length;  apply suffix_code.
  unfold code_pair in HsubprogGt ; clear Hsubprog code_pair.
  assert (HsubprogGt' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (code_pair')) (start'+length (fst (fst code2'))+length (fst (fst code1'))) )).
  eapply subPfunction_extends; try apply Hsubprog'.
  rewrite <- sub_is_extends; simpl;rewrite <- plus_assoc; rewrite <- app_length;  apply suffix_code.
  unfold code_pair' in HsubprogGt' ; clear Hsubprog' code_pair'.

  unfold memory_specification; simpl.
  intros ptr_a ptr_a'.
  unfold code2, code2_gen; rewrite compile_length_main; fold code2_gen code2.
  unfold code2', code2_gen'; rewrite compile_length_main; fold code2_gen' code2'.
  set (new_start := start+length (fst (fst code2))+length (fst (fst code1))) in *.
  set (new_start' := start'+length (fst (fst code2'))+length (fst (fst code1'))) in *.
  unfold gt_code in HsubprogGt'.
  (* [sp] <- [sp] - 1            pop *)
  instr_opimmleft minus new_start new_start' 0 HsubprogGt HsubprogGt' (stack_ptr+1+1) (stack_ptr'+1+1) 1 1.
  replace (stack_ptr +1 + 1 -1) with (stack_ptr+1) ; try omega...
  replace (stack_ptr' +1 + 1 -1) with (stack_ptr'+1) ; try omega...

  focus 6.
  do 2 (apply judgment_lift_stack;intro).
  destruct H as (HP2,(HP2',Hptr_b')).
  destruct H0 as (HP1,(HP1',Hptr_a')).
  subst.
  focus 5;focus 6; focus 3;focus 7.
  do 3 rewrite stateRelTensor_assoc; hide_tail...
  (* [wrk] <- [[sp]] - [[sp]+1] *)
  eapply (opimm_partial minus) with (w:=ptr_a') (u:=ptr_b') (w':=ptr_a') (u':=ptr_b'); 
    try apply HsubprogGt ; try apply HsubprogGt' ;
      [instrextract new_start 1 | instrextract new_start' 1 | 
        idtac |
	  rewrite_start new_start new_start' 1 ].
  destruct_opimm op.
  eapply (ptsto_statevalOp minus).
  assert (H:s spreg = stack_ptr+1 /\ s' spreg = stack_ptr'+1);
  [ eapply ptsto_stateval; focus_hyp 1;apply Hpost1 | 
    destruct H as (H,H');  rewrite H ,H';focus_hyp 2;focus_hyp 3;apply Hpost1].
  (* we have to apply hyp twice (case 0 and 1) so we have defined a tactic to write it only once *)

  instr_brzimm new_start new_start' 2 HsubprogGt HsubprogGt'.

  intro Hineq; compute_code_size code1_gen code2_gen code1 code2; fold new_start;
  compute_code_size code1_gen' code2_gen' code1' code2'; fold new_start';
    focus 2;focus 2; rewrite Hineq .
  (* [[sp]] <- [wrk]*)
  instr_moveind new_start new_start' 4 HsubprogGt HsubprogGt' with nat 0 0.

  assert (H5:start+length ((fst (fst code2) ++ fst (fst code1)) ++
             gt_code (snd code1) alloc (snd code1_gen)) = new_start+5);
  [unfold new_start;  simpl;  repeat rewrite app_length;  simpl;  omega | idtac];
  assert (H5':start'+length ((fst (fst code2') ++ fst (fst code1')) ++
             gt_code (snd code1') alloc' (snd code1_gen')) = new_start'+5);
  [unfold new_start';  simpl;  repeat rewrite app_length;  simpl;  omega | idtac];
  simpl in hyp;
  rewrite <- plus_assoc; rewrite H5 in hyp;
  rewrite <- plus_assoc; rewrite H5' in hyp.
  unfold memory_specification in hyp;  simpl in hyp;
  hreplace hyp with (hyp 0 0); unfold tail ; clear tail; normal_form.
  focus 5;focus 4; rewrite ptsto_blank_leq.
  omega_replace (stack_ptr+1+1+1) with (S(stack_ptr+1+1)); omega_replace (stack_ptr+2) with (stack_ptr+1+1);
  omega_replace (stack_ptr'+1+1+1) with (S(stack_ptr'+1+1)); omega_replace (stack_ptr'+2) with (stack_ptr'+1+1);
  rewrite stateRelTensor_assoc; rewrite <- block_unfold_start.
  eapply judgment_leq; try apply hyp .
  
  assert (snd(fst code1) >= 1); [ unfold code1, code1_gen, code2_gen ;  rewrite compile_length_aux;
  apply stack_size_positive | idtac];
  assert (snd(fst code1') >= 1); [ unfold code1', code1_gen', code2_gen' ;  rewrite compile_length_aux;
  apply stack_size_positive | idtac].
  unfold ge in Hstack; assert (2<=stack_free);
  [eapply le_trans ; try apply Hstack;  apply le_trans with (m := snd(fst code1)+1) ; try apply le_max_r ; omega| idtac];
  unfold ge in Hstack'; assert (2<=stack_free');
  [eapply le_trans ; try apply Hstack';  apply le_trans with (m := snd(fst code1')+1) ; try apply le_max_r ; omega| idtac];
  omega_replace (S(stack_free-1-1)) with (stack_free-1);
  omega_replace (stack_ptr+1-1) with stack_ptr;
  omega_replace (S(stack_free'-1-1)) with (stack_free'-1);
  omega_replace (stack_ptr'+1-1) with stack_ptr'...

  focus 3;rewrite ptsto_blank_leq.
  focus 4;focus' 5.
  eapply ppredtensorcompat2.
    
  setoid_replace (AbstractGt P1 P2 (bool_of_nat 0)) with True using relation iff ;[idtac | unfold AbstractGt; simpl; intuition; exists ptr_a'; exists ptr_b'; intuition];
  setoid_replace (false = false) with True using relation iff by intuition;
  setoid_replace (True /\ True /\ True) with True using relation iff; [idtac | intuition];
  fold cloudPtrue; setoid_rewrite cloudRelConj_unit.
  reflexivity.
  
  ineq_ring.
  
  intro Hineq; 
    (* [wrk] <- 1 *) 
    instr_moveimm new_start new_start' 3 HsubprogGt HsubprogGt' with nat 1 1.
    focus 2;focus 2.
    (* [[sp]] <- [wrk] *)
    instr_moveind new_start new_start' 4 HsubprogGt HsubprogGt' with nat 1 1.

  assert (H5:start+length ((fst (fst code2) ++ fst (fst code1)) ++
             gt_code (snd code1) alloc (snd code1_gen)) = new_start+5);
  [unfold new_start;  simpl;  repeat rewrite app_length;  simpl;  omega | idtac];
  assert (H5':start'+length ((fst (fst code2') ++ fst (fst code1')) ++
             gt_code (snd code1') alloc' (snd code1_gen')) = new_start'+5);
  [unfold new_start';  simpl;  repeat rewrite app_length;  simpl;  omega | idtac];
  simpl in hyp;
  rewrite <- plus_assoc; rewrite H5 in hyp;
  rewrite <- plus_assoc; rewrite H5' in hyp.
  unfold memory_specification in hyp;  simpl in hyp;
  hreplace hyp with (hyp 1 1); unfold tail ; clear tail; normal_form.
  focus 5;focus 4; rewrite ptsto_blank_leq.
  omega_replace (stack_ptr+1+1+1) with (S(stack_ptr+1+1)); omega_replace (stack_ptr+2) with (stack_ptr+1+1);
  omega_replace (stack_ptr'+1+1+1) with (S(stack_ptr'+1+1)); omega_replace (stack_ptr'+2) with (stack_ptr'+1+1);
  rewrite stateRelTensor_assoc; rewrite <- block_unfold_start.
  eapply judgment_leq; try apply hyp.
  assert (snd(fst code1) >= 1); [ unfold code1, code1_gen, code2_gen ;  rewrite compile_length_aux;
  apply stack_size_positive | idtac];
  assert (snd(fst code1') >= 1); [ unfold code1', code1_gen', code2_gen' ;  rewrite compile_length_aux;
  apply stack_size_positive | idtac].
  unfold ge in Hstack; assert (2<=stack_free);
  [eapply le_trans ; try apply Hstack;  apply le_trans with (m := snd(fst code1)+1) ; try apply le_max_r ; omega| idtac];
  unfold ge in Hstack'; assert (2<=stack_free');
  [eapply le_trans ; try apply Hstack';  apply le_trans with (m := snd(fst code1')+1) ; try apply le_max_r ; omega| idtac];
  omega_replace (S(stack_free-1-1)) with (stack_free-1);
  omega_replace (stack_ptr+1-1) with stack_ptr;
  omega_replace (S(stack_free'-1-1)) with (stack_free'-1);
  omega_replace (stack_ptr'+1-1) with stack_ptr'...
  focus 3;rewrite ptsto_blank_leq.
  focus 4;focus' 5.
  eapply ppredtensorcompat2;[
  setoid_replace (AbstractGt P1 P2 (bool_of_nat 1)) with True using relation iff ;[idtac | unfold AbstractGt; simpl; intuition; exists ptr_a'; exists ptr_b'; intuition];
  setoid_replace (true = true) with True using relation iff by intuition;
  setoid_replace (True /\ True /\ True) with True using relation iff; [idtac | intuition];
  fold cloudPtrue; repeat rewrite cloudRelConj_unit; reflexivity |
  ineq_ring].
Qed.




