(* correction_ifThenElse.v

   This is the partial correctness proof for the if-then-else for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)


Require Import correction_compiler_definition.

Typeclasses eauto := dfs 10000.

Lemma correction_of_ifThenElse : forall base base' Ra alloc alloc' dealloc dealloc' env P a e1 e2 e3 (t1:env|--e1:::Bool P) (t2:P true -> env|--e2:::a) (t3:P false -> env|--e3:::a), 
(forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env (Bool P) e1 t1 Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
(forall (p:P true) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env a e2 (t2 p) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
(forall (p:P false) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env a e3 (t3 p) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env a (IF_THEN_ELSE e1 e2 e3) (TIF_THEN_ELSE t1 t2 t3) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3'.
Proof with normal_form.
  intros until t3; intros HypRecE1 HypRecE2 HypRecE3; intros. 
  unfold statement_of_compiler_correctness in *.
  unfold prog_extends_code, prog_extends_auxcode, extract_code_from_globalcode, extract_stacksize_from_code in *;
  intros until global_code'; intros hprog hcode Hstack_free hprog' hcode' Hstack_free' Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc.
  split ; [idtac | intros e' Hlambda; inversion Hlambda]; intro hyp.

    
  (* recover the actual piece of code *)

  simpl in hprog; surjective_pairing_three_arg e1 e2 e3 aux_code base alloc dealloc in hprog;
  rewrite <- hcode in Hsubprog; rewrite <- hcode' in Hsubprog';        
  rewrite hprog in hcode; simpl in hcode;
  surjective_pairing_three_arg2 e1 e2 e3 code1_gen code2_gen code3_gen start in hcode.
  simpl in hcode; rewrite hprog in Hsubprog_aux; simpl in Hsubprog_aux; clear hprog.
  rewrite hcode in Hstack_free; simpl in Hstack_free.
  simpl in hprog'; surjective_pairing_three_arg e1 e2 e3 aux_code' base' alloc' dealloc' in hprog'.
  rename code1_gen0 into code1_gen'; rename code2_gen0 into code2_gen'; rename code3_gen0 into code3_gen'.
  rewrite hprog' in hcode'; simpl in hcode'.
  surjective_pairing_three_arg2 e1 e2 e3 code1_gen' code2_gen' code3_gen' start' in hcode'.
  rename code0 into code1'; rename code4 into code2'; rename code5 into code3'.
  simpl in hcode'; rewrite hprog' in Hsubprog_aux'; simpl in Hsubprog_aux'; clear hprog'.
  rewrite hcode' in Hstack_free'; simpl in Hstack_free'.

  (* proving some assertions to be able to apply the three induction hypotheses *)

  assert (hcode2 : fst (fst code) = (fst(fst code1)) ++ (if_1_code (snd code2) 0 0) ++ (fst(fst code2)) ++ (if_2_code (snd code3) 0 0) ++ (fst(fst code3))); try (rewrite hcode;simpl;reflexivity).
  assert (Hstack_free1 : snd(fst code1)  <= stack_free); try omega.
  assert (Hstack_free2' : snd(fst code2)  <= stack_free-snd(fst code1)); try (eapply le_trans; try apply (le_max_l (snd(fst code2)) (snd(fst code3))); omega).
  assert (Hstack_free3' : snd(fst code3)  <= stack_free-snd(fst code1)); try (eapply le_trans; try apply (le_max_l (snd(fst code3)) (snd(fst code2))); rewrite max_comm; omega).
  assert (foo : snd(fst code1)  >= 1); try (unfold code1, code1_gen; apply stack_size_positive).
  assert (Hstack_free2 : snd(fst code2)  <= stack_free); try omega.
  assert (Hstack_free3 : snd(fst code3)  <= stack_free); try omega.
  assert (Hstack_free_base: 1  <= stack_free); try omega.
  clear foo Hstack_free2' Hstack_free3'.
  assert (hcode2' : fst (fst code') = (fst(fst code1')) ++ (if_1_code (snd code2') 0 0) ++ (fst(fst code2')) ++ (if_2_code (snd code3') 0 0) ++ (fst(fst code3'))); try (rewrite hcode';simpl;reflexivity).
  assert (Hstack_free1' : snd(fst code1')  <= stack_free'); try omega.
  assert (Hstack_free2'' : snd(fst code2')  <= stack_free'-snd(fst code1')); try (eapply le_trans; try apply (le_max_l (snd(fst code2')) (snd(fst code3'))); omega).
  assert (Hstack_free3'' : snd(fst code3')  <= stack_free'-snd(fst code1')); try (eapply le_trans; try apply (le_max_l (snd(fst code3')) (snd(fst code2'))); rewrite max_comm; omega).
  assert (foo : snd(fst code1')  >= 1); try (unfold code1', code1_gen'; apply stack_size_positive).
  assert (Hstack_free2' : snd(fst code2')  <= stack_free'); try omega.
  assert (Hstack_free3' : snd(fst code3')  <= stack_free'); try omega.
  assert (Hstack_free_base': 1  <= stack_free'); try omega.
  clear foo Hstack_free2'' Hstack_free3''.

  assert (Hsubprog1 : PfunctionExtendsFunction p (fragfromprog (progfromlist (fst(fst code1))) start));
  [eapply subPfunction_extends; try apply Hsubprog;  rewrite <- sub_is_extends;  rewrite hcode; simpl; apply prefix_code | idtac].
  assert (Hsubprog2 : PfunctionExtendsFunction p (fragfromprog (progfromlist (fst(fst code2))) (start+length ((fst (fst code1))++(if_1_code (snd code2) 0 0)))));
  [eapply subPfunction_extends; try apply Hsubprog;  rewrite <- sub_is_extends;  rewrite hcode2; rewrite <- app_ass; apply sub_code| idtac].
  assert (Hsubprog3 : PfunctionExtendsFunction p (fragfromprog (progfromlist (fst(fst code3))) (start+
  length ((fst (fst code1))++ (if_1_code (snd code2) 0 0) ++ (fst(fst code2)) ++(if_2_code (snd code3) 0 0)  ))));
  [eapply subPfunction_extends; try apply Hsubprog;  rewrite <- sub_is_extends;  rewrite hcode2;
  repeat rewrite <- app_ass; apply suffix_code  | idtac].
  assert (Hsubprog2_aux : PfunctionExtendsFunction p (fragfromprog (progfromlist (snd(fst code2_gen))) base));
  [eapply subPfunction_extends; try apply Hsubprog_aux;  rewrite <- sub_is_extends;
  unfold code3_gen, code2_gen, code1_gen; rewrite compile_length_aux; rewrite compile_length_aux;
  rewrite <- (compile_length_aux e1);  fold code1_gen code2_gen; apply auxcode_monotony | idtac].
  assert (Hsubprog1_aux : PfunctionExtendsFunction p (fragfromprog (progfromlist (snd(fst code1_gen))) base));
  [ eapply subPfunction_extends; try apply Hsubprog2_aux;  rewrite <- sub_is_extends;
  unfold code2_gen, code1_gen; rewrite compile_length_aux; fold code1_gen code2_gen; apply auxcode_monotony | idtac].
  assert (Hlength:start + length ((fst (fst code1))++(if_1_code (snd code2) 0 0)++ (fst(fst code2)) ++ (if_2_code (snd code3) 0 0)) = (snd code2)+1).
  do 3 rewrite app_length; simpl; unfold code1, code1_gen;  rewrite plus_assoc;
  rewrite <- compile_length_main;  fold code1_gen code1;  omega_replace ((snd code1)+S(S(length (fst (fst code2)) +1))) with ((snd code1)+2+length(fst(fst code2))+1);
  unfold code2, code2_gen, code1_gen; rewrite compile_length_aux;  rewrite <- compile_length_main ; rewrite <- compile_length_aux;  fold code1_gen code1 code2_gen code2; reflexivity.

  assert (Hsubprog1' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (fst(fst code1'))) start'));
  [eapply subPfunction_extends; try apply Hsubprog';  rewrite <- sub_is_extends;  rewrite hcode'; simpl; apply prefix_code | idtac].
  assert (Hsubprog2' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (fst(fst code2'))) (start'+length ((fst (fst code1'))++(if_1_code (snd code2') 0 0)))));
  [eapply subPfunction_extends; try apply Hsubprog';  rewrite <- sub_is_extends;  rewrite hcode2'; rewrite <- app_ass; apply sub_code| idtac].
  assert (Hsubprog3' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (fst(fst code3'))) (start'+ length ((fst (fst code1'))++ (if_1_code (snd code2') 0 0) ++ (fst(fst code2')) ++(if_2_code (snd code3') 0 0)  ))));
  [eapply subPfunction_extends; try apply Hsubprog';  rewrite <- sub_is_extends;  rewrite hcode2';
  repeat rewrite <- app_ass; apply suffix_code  | idtac].
  assert (Hsubprog2_aux' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (snd(fst code2_gen'))) base'));
  [eapply subPfunction_extends; try apply Hsubprog_aux';  rewrite <- sub_is_extends;
  unfold code3_gen', code2_gen', code1_gen'; rewrite compile_length_aux; rewrite compile_length_aux;
  rewrite <- (compile_length_aux e1);  fold code1_gen' code2_gen'; apply auxcode_monotony | idtac].
  assert (Hsubprog1_aux' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (snd(fst code1_gen'))) base'));
  [ eapply subPfunction_extends; try apply Hsubprog2_aux';  rewrite <- sub_is_extends;
  unfold code2_gen', code1_gen'; rewrite compile_length_aux; fold code1_gen' code2_gen'; apply auxcode_monotony | idtac].
  assert (Hlength':start' + length ((fst (fst code1'))++(if_1_code (snd code2') 0 0)++ (fst(fst code2')) ++ (if_2_code (snd code3') 0 0)) = (snd code2')+1).
  do 3 rewrite app_length; simpl; unfold code1', code1_gen';  rewrite plus_assoc;
  rewrite <- compile_length_main;  fold code1_gen' code1';  omega_replace ((snd code1')+S(S(length (fst (fst code2')) +1))) with ((snd code1')+2+length(fst(fst code2'))+1);
  unfold code2', code2_gen', code1_gen'; rewrite compile_length_aux;  rewrite <- compile_length_main ; rewrite <- compile_length_aux;  fold code1_gen' code1' code2_gen' code2'; reflexivity.

  rewrite hcode2 in hyp; repeat rewrite <- app_ass in hyp;  rewrite app_length in hyp;  repeat rewrite app_ass in hyp.
  rewrite plus_assoc in hyp; rewrite Hlength in hyp.
  rewrite hcode2' in hyp; repeat rewrite <- app_ass in hyp;  rewrite app_length in hyp;  repeat rewrite app_ass in hyp.
  rewrite plus_assoc in hyp; rewrite Hlength' in hyp.

  (* apply each argument *)

  hreplace HypRecE1 with (HypRecE1 Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' code1 code1' p p' stack_list code1_gen code1_gen');
  fold code1 code1_gen code1' code1_gen' in HypRecE1.
  hreplace HypRecE1 with (HypRecE1 (refl_equal code1_gen) (refl_equal code1) Hstack_free1 (refl_equal code1_gen') (refl_equal code1') Hstack_free1' Hsubprog1_aux Hsubprog1 Hsubprog1_aux' Hsubprog1' Halloc Hdealloc).
  destruct HypRecE1 as (HypRecE1,Huseless); clear Huseless.
  apply HypRecE1;  clear HypRecE1 Hstack_free1  Hstack_free1' Hsubprog1 Hsubprog1' Hsubprog1_aux Hsubprog1_aux'.

  intros ptr_e1 ptr_e1'.
  
  assert (HsubprogIf1 : PfunctionExtendsFunction p (fragfromprog (progfromlist ((if_1_code (snd code2) 0 0))) (start+length(fst(fst code1)))));
  [eapply subPfunction_extends; try apply Hsubprog;  rewrite <- sub_is_extends;  rewrite hcode2; apply sub_code | idtac].
  assert (HsubprogIf1' : PfunctionExtendsFunction p' (fragfromprog (progfromlist ((if_1_code (snd code2') 0 0))) (start'+length(fst(fst code1')))));
  [eapply subPfunction_extends; try apply Hsubprog';  rewrite <- sub_is_extends;  rewrite hcode2'; apply sub_code | idtac].
  set (new_start:= start + length(fst(fst code1))) in *.
  set (new_start':= start' + length(fst(fst code1'))) in *.
  unfold memory_specification; simpl.
  (* [sp] <- [sp] - 1         pop *)
  instr_opimmleft minus new_start new_start' 0 HsubprogIf1 HsubprogIf1' (stack_ptr+1) (stack_ptr'+1) 1 1.
  omega_replace (stack_ptr+1-1) with stack_ptr.
  omega_replace (stack_ptr'+1-1) with stack_ptr'...
  focus 5.
  apply judgment_lift_stack.
  intro H; destruct H as (HPe1,(HPe2,He1e2)).
  focus 5;focus 2.
  (* branch on the value of the boolean [[sp]+1]  *)

  instr_brzindo new_start new_start' 1 HsubprogIf1 HsubprogIf1' He1e2.
  
  (* we have to apply Hyprec3 *)
  intros bool0.
  focus 3;focus 2 ; rewrite ptsto_blank_leq.
  omega_replace (stack_ptr+1+1) with (S(stack_ptr+1)); 
  omega_replace (stack_ptr'+1+1) with (S(stack_ptr'+1)).
  rewrite stateRelTensor_assoc; rewrite <- block_unfold_start.
  omega_replace (S(stack_free-1)) with stack_free.
  omega_replace (S(stack_free'-1)) with stack_free'.
  rewrite bool0 in *; simpl in *.
  hreplace HypRecE3 with (HypRecE3 HPe1 Rc Rc_cloud (snd code2+1) (snd code2'+1) stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3' (snd(fst code2_gen)) (snd(fst code2_gen')) code3 code3' p p' stack_list code3_gen code3_gen').
  fold code3  in HypRecE3;  unfold code2_gen in HypRecE3 at 1; unfold code1_gen in HypRecE3 at 1;
  rewrite compile_length_aux in HypRecE3;  fold code1_gen code1 in HypRecE3; rewrite <- compile_length_aux in HypRecE3;
  unfold code1_gen in HypRecE3; rewrite <- compile_length_aux in HypRecE3; fold code1_gen code1 code2_gen code3_gen in HypRecE3.
  fold code3' in HypRecE3;  unfold code2_gen' in HypRecE3 at 1; unfold code1_gen' in HypRecE3 at 1;
  rewrite compile_length_aux in HypRecE3;  fold code1_gen' code1' in HypRecE3; rewrite <- compile_length_aux in HypRecE3;
  unfold code1_gen' in HypRecE3; rewrite <- compile_length_aux in HypRecE3; fold code1_gen' code1' code2_gen' code3_gen' in HypRecE3.
  rewrite Hlength in Hsubprog3; rewrite Hlength' in Hsubprog3'.
  hreplace HypRecE3 with (HypRecE3 (refl_equal code3_gen) (refl_equal code3) Hstack_free3 (refl_equal code3_gen') (refl_equal code3') Hstack_free3' Hsubprog_aux Hsubprog3 Hsubprog_aux' Hsubprog3' Halloc Hdealloc).
  destruct HypRecE3 as (HypRecE3,Huseless); clear Huseless;
  eapply judgment_leq; try apply HypRecE3.
  unfold memory_specification; simpl; ineq_ring.

  intros ptr ptr'; unfold memory_specification; simpl.
  set (new_start2:= snd code2 + 1 + length (fst(fst code3))) in *.
  set (new_start2':= snd code2' + 1 + length (fst(fst code3'))) in *.
  focus 1.
  hreplace hyp with (hyp ptr ptr'); eapply judgment_leq ; try apply hyp.
  unfold memory_specification; simpl;  ineq_ring.
  
  (* we have to apply Hyprec2 *)
  intro bool1.
  focus 3;focus 2; rewrite ptsto_blank_leq.
  omega_replace (stack_ptr+1+1) with (S(stack_ptr+1)); 
  omega_replace (stack_ptr'+1+1) with (S(stack_ptr'+1)); 
  rewrite stateRelTensor_assoc; rewrite <- block_unfold_start.
  omega_replace (S(stack_free-1)) with stack_free.
  omega_replace (S(stack_free'-1)) with stack_free'.
  rewrite bool1 in *; simpl in *.
  specialize (HypRecE2 HPe1 Rc Rc_cloud (new_start+2) (new_start'+2) stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3' (snd(fst code1_gen)) (snd(fst code1_gen')) code2 code2' p p' stack_list code2_gen code2_gen');
  unfold new_start, new_start' in HypRecE2.
  
  unfold code1_gen in HypRecE2; rewrite <- compile_length_aux in HypRecE2;  fold code1_gen code1 code2_gen in HypRecE2;
  do 2 (unfold code1 at 1 in HypRecE2); unfold code1_gen in HypRecE2; repeat (rewrite <- compile_length_main in HypRecE2);  fold code1_gen code1 code2_gen code2  in HypRecE2.
  rewrite app_length in Hsubprog2;  simpl in Hsubprog2;  rewrite plus_assoc in Hsubprog2.
  unfold code1_gen' in HypRecE2; rewrite <- compile_length_aux in HypRecE2;  fold code1_gen' code1' code2_gen' in HypRecE2;
  do 2 (unfold code1' at 1 in HypRecE2);  unfold code1_gen' in HypRecE2; repeat (rewrite <- compile_length_main in HypRecE2);  fold code1_gen' code1' code2_gen' code2'  in HypRecE2.
  rewrite app_length in Hsubprog2';  simpl in Hsubprog2';  rewrite plus_assoc in Hsubprog2'.  
  specialize (HypRecE2 (refl_equal code2_gen) (refl_equal code2) Hstack_free2 (refl_equal code2_gen') (refl_equal code2') Hstack_free2' Hsubprog2_aux Hsubprog2 Hsubprog2_aux' Hsubprog2' Halloc Hdealloc);
  destruct HypRecE2 as (HypRecE2,Huseless); clear Huseless.
  eapply judgment_leq; try apply HypRecE2.
  unfold memory_specification; simpl;ineq_ring.
  
  (* now applying if2 *)

  intros ptr ptr';   unfold memory_specification; simpl...
  set (new_start3:= start + length(fst(fst code1)) + 2 + length(fst(fst code2))) in *.
  set (new_start3':= start' + length(fst(fst code1')) + 2 + length(fst(fst code2'))) in *.

  assert (HsubprogIf2 : PfunctionExtendsFunction p (fragfromprog (progfromlist ((if_2_code (snd code3) 0 0))) (start + length(fst(fst code1)++(if_1_code (snd code2) 0 0)++ fst(fst code2)))));
  [eapply subPfunction_extends; try apply Hsubprog;  rewrite <- sub_is_extends;  rewrite hcode2;
  unfold new_start3 | idtac].
  replace (fst (fst code1) ++
            i_op minus (d_imm spreg) (s_imm spreg) (s_cst 1)
            :: i_brz (s_indo 1 spreg) (s_cst (snd code2 + 1))
               :: fst (fst code2) ++
                  i_branch (s_cst (snd code3)) :: fst (fst code3)) with
  ((fst (fst code1) ++
            i_op minus (d_imm spreg) (s_imm spreg) (s_cst 1)
            :: i_brz (s_indo 1 spreg) (s_cst (snd code2 + 1))
               :: fst (fst code2)) ++
                  ([i_branch (s_cst (snd code3))]) ++ fst (fst code3)).
  apply sub_code.
  simpl; repeat rewrite app_ass.
  reflexivity.

  assert (HsubprogIf2' : PfunctionExtendsFunction p' (fragfromprog (progfromlist ((if_2_code (snd code3') 0 0))) (start' + length(fst(fst code1')++(if_1_code (snd code2') 0 0)++ fst(fst code2')))));
  [eapply subPfunction_extends; try apply Hsubprog';  rewrite <- sub_is_extends;  rewrite hcode2';
  unfold new_start3 | idtac].
  replace (fst (fst code1') ++
            i_op minus (d_imm spreg) (s_imm spreg) (s_cst 1)
            :: i_brz (s_indo 1 spreg) (s_cst (snd code2' + 1))
               :: fst (fst code2') ++
                  i_branch (s_cst (snd code3')) :: fst (fst code3')) with
  ((fst (fst code1') ++
            i_op minus (d_imm spreg) (s_imm spreg) (s_cst 1)
            :: i_brz (s_indo 1 spreg) (s_cst (snd code2' + 1))
               :: fst (fst code2')) ++
                  ([i_branch (s_cst (snd code3'))]) ++ fst (fst code3')).
  apply sub_code.
  simpl; repeat rewrite app_ass.
  reflexivity.
  (* branch to the end of the code of e3 *)
  do 2 rewrite app_length in HsubprogIf2; simpl in HsubprogIf2.
  do 2 rewrite app_length in HsubprogIf2'; simpl in HsubprogIf2'.
  replace (start +(length (fst (fst code1)) + S(S(length(fst (fst code2)))))) with new_start3 in HsubprogIf2; try(unfold new_start3; omega).
  replace (start' +(length (fst (fst code1')) + S(S(length(fst (fst code2')))))) with new_start3' in HsubprogIf2'; try(unfold new_start3'; omega).
  unfold if_2_code in HsubprogIf2.
  unfold if_2_code in HsubprogIf2'.
  instr_jumpcst new_start3 new_start3' 0 HsubprogIf2 HsubprogIf2'.
  hreplace hyp with (hyp ptr ptr').
  unfold code3, code3_gen, code2_gen, code2_gen, code1_gen , code1;
  unfold code3', code3_gen', code2_gen', code2_gen', code1_gen' , code1'.
  rewrite compile_length_aux;  rewrite compile_length_aux;
  rewrite compile_length_main;  fold code1_gen code1;
  rewrite <- compile_length_aux;  unfold code1_gen;
  rewrite <- compile_length_aux;  fold code1_gen code2_gen code2 code3_gen code3.
  rewrite compile_length_aux;  rewrite compile_length_aux;
  rewrite compile_length_main;  fold code1_gen' code1';
  rewrite <- compile_length_aux;  unfold code1_gen';
  rewrite <- compile_length_aux;  fold code1_gen' code2_gen' code2' code3_gen' code3'.
  
  set (end_if:=snd code2+1+length(fst (fst code3))) in *.
  set (end_if':=snd code2'+1+length(fst (fst code3'))) in *.
  eapply judgment_leq; try apply hyp.
  unfold memory_specification; simpl;  ineq_ring.
Qed.




