(* correction_lambda.v

   This is the partial correctness proof for functions for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import correction_compiler_definition.
Require Import correction_functions.

Typeclasses eauto := dfs 10000.

Lemma correction_of_lambda : forall base base' Ra alloc alloc' dealloc dealloc' env a b e (t : a::env|-- e:::b), 
(forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' (a::env) b e t Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
  forall Rc Rc_cloud start  start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env (a--->b) (LAMBDA e) (TLAMBDA t) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3'.
Proof with normal_form.
  intros until e; intros t HypRec; intros; unfold statement_of_compiler_correctness.
  unfold prog_extends_code, prog_extends_auxcode, extract_code_from_globalcode, extract_stacksize_from_code;
  intros until global_code'; intros hprog hcode Hstack hprog' hcode' Hstack' Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc.
  set (code_of_e := compile e aux_code (base+length aux_code) alloc dealloc).
  set (code_of_e' := compile e aux_code' (base'+length aux_code') alloc' dealloc').
  split.
  
  (* apply the lemma giving the shape of the heap after closure_make  *)
  
  assert (Hmain := correction_of_lambda_main base base' Ra alloc alloc' dealloc dealloc' env e Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n n2 n3 n' n' n2' n3' aux_code aux_code' code code' p p' stack_list code_of_e code_of_e' global_code global_code' hprog hcode).
  fold code_of_e in Hmain.
  rename Hstack into Hstack_free; rename Hstack' into Hstack_free'.
  hreplace Hmain with (Hmain (refl_equal code_of_e) Hstack_free hprog' hcode').
  fold code_of_e' in Hmain.
  rewrite <- hcode in Hsubprog; rewrite <- hcode' in Hsubprog';
  hreplace Hmain with (Hmain (refl_equal code_of_e') Hstack_free' Hsubprog Hsubprog' Halloc).
  intro hyp; eapply Hmain.
  clear Hmain; intros fb fb'.

  (* apply hyp *)

  unfold memory_specification in hyp; set (new_start := start + length (fst(fst code))) in *; simpl in hyp.
  set (new_start' := start' + length (fst(fst code'))) in *; simpl in hyp.
  hreplace hyp with (hyp fb fb'); normal_form in hyp.
  omega_replace (S(stack_ptr +1)) with (stack_ptr+2); replace (stack_ptr +1+1) with (stack_ptr+2) in hyp; try omega.
  omega_replace (S(stack_ptr' +1)) with (stack_ptr'+2); replace (stack_ptr' +1+1) with (stack_ptr'+2) in hyp; try omega.
  focus_hyp 5; rewrite <- cloudExP_conj in hyp;  rewrite <- stateExP_stack in hyp;  rewrite <- existsJudgment  in hyp.
  set (fcode:= snd code_of_e) in *.
  set (fcode':= snd code_of_e') in *.
  focus 7;focus 1; rewrite stateRelTensor_assoc; rewrite stack_description_strength; rewrite stateCloudPtsot.
  focus 1; rewrite stateRelTensor_assoc; rewrite stack_description_strength; setoid_rewrite stateCloudPtsot.
  hreplace hyp with (hyp (({fb+1,fb'+1}||->{n,n'}) //\\ [||env||] Ra n n')).
  rewrite (cloudConjContraction ([||env||] Ra n n') ([||env||] Ra n n')) by reflexivity.
  normal_form_conj in hyp.
  replace (stack_ptr+1-1) with stack_ptr in hyp; try omega.
  replace (stack_ptr'+1-1) with stack_ptr' in hyp; try omega.
  eapply continuation_enrichment_stack; [idtac | rewrite ptsto_later_c; 
    eapply judgment_leq; try apply hyp ; eapply ppredtensorcompat2; 
    [normal_form_conj;setoid_reflexivity | focus' 4; reflexivity]].
  
  (* apply the lemma that certifies that the code the stack is pointing to satisfies the good property *)

  assert (Haux := correction_of_lambda_aux base base' Ra alloc dealloc alloc' dealloc' env e a b t aux_code aux_code' code_of_e code_of_e' p p' global_code global_code' hprog hprog').
  fold code_of_e in Haux; fold code_of_e' in Haux.
  hreplace Haux with (Haux (refl_equal code_of_e) (refl_equal code_of_e') Hsubprog_aux Hsubprog_aux' 
    Halloc Hdealloc).
  unfold fcode,fcode'.
  apply Haux.
  intros; apply HypRec.
  
  (* apply again the former lemma to prove the extra property on the behaviour of a lambda *)

  intros e' He'; inversion He'.
  rewrite <- H0; clear H0 He' e'; simpl.
  assert (Haux := correction_of_lambda_aux base base' Ra alloc dealloc alloc' dealloc' env e a b t aux_code aux_code' code_of_e code_of_e' p p' global_code global_code' hprog hprog').
  fold code_of_e in Haux.
  fold code_of_e' in Haux.
  specialize (Haux (refl_equal code_of_e) (refl_equal code_of_e') Hsubprog_aux Hsubprog_aux' Halloc Hdealloc).
  fold code_of_e; fold code_of_e'.
  apply Haux; apply HypRec.
Qed.


