(* correction_pair.v

   This is the partial correctness proof for pairs for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import correction_compiler_definition.

Typeclasses eauto := dfs 10000.

Lemma correction_of_pair : forall base base' Ra alloc alloc' dealloc dealloc' env a b e1 e2 (t1:env|--e1:::a) (t2:env|--e2:::b), 
(forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env a e1 t1 Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
(forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env b e2 t2 Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env (a**b) (P e1 e2) (TP t1 t2) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3'.
Proof with normal_form.
  intros until t2; intros HypRecE1 HypRecE2; intros. 
  unfold statement_of_compiler_correctness in *.
  unfold prog_extends_code, prog_extends_auxcode, extract_code_from_globalcode, extract_stacksize_from_code in *;
  intros until global_code'; intros hprog hcode Hstack hprog' hcode' Hstack' Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc.
  split ; [idtac | intros e' Hlambda; inversion Hlambda]; intro hyp.

  (* recover the actual piece of code *)

  simpl in hprog; surjective_pairing_two_arg e1 e2 aux_code base alloc dealloc in hprog;
  rewrite <- hcode in Hsubprog; rewrite <- hcode' in Hsubprog';        
  rewrite hprog in hcode; simpl in hcode;
  surjective_pairing_two_arg e1 e2 code1_gen code2_gen start in hcode.
  simpl in hprog'; surjective_pairing_two_arg e1 e2 aux_code' base' alloc' dealloc' in hprog';
  rewrite hprog' in hcode'; simpl in hcode';
  rename code2_gen0 into code2_gen'; rename code1_gen0 into code1_gen';
  surjective_pairing_two_arg e1 e2 code1_gen' code2_gen' start' in hcode'.

  unfold pair_code in hcode; simpl in hcode.
  rewrite hprog in Hsubprog_aux; simpl in Hsubprog_aux; clear hprog.
  rewrite hcode in Hstack; simpl in Hstack.
  unfold pair_code in hcode'; simpl in hcode'.
  rewrite hprog' in Hsubprog_aux'; simpl in Hsubprog_aux'; clear hprog'.
  rewrite hcode' in Hstack'; simpl in Hstack'.
  rename code0 into code2'; rename code3 into code1'.
  replace (snd(fst code2)+0) with (snd(fst code2)) in *; try omega.
  replace (snd(fst code2')+0) with (snd(fst code2')) in *; try omega.
  assert (Hmin : min (snd(fst code2)) 1 = 1); try (apply min_r;  unfold code2,code2_gen; apply stack_size_positive).
  rewrite Hmin in Hstack.
  assert (max (snd (fst code2)) (snd(fst code1) +1) <= stack_free); [eapply le_trans ; try apply Hstack; apply le_max_l | idtac].
  clear Hmin Hstack; rename H into Hstack.
  assert (Hmin' : min (snd(fst code2')) 1 = 1); try (apply min_r;  unfold code2',code2_gen'; apply stack_size_positive).
  rewrite Hmin' in Hstack'.
  assert (max (snd (fst code2')) (snd(fst code1') +1) <= stack_free'); [eapply le_trans ; try apply Hstack'; apply le_max_l | idtac].
  clear Hmin' Hstack'; rename H into Hstack'.

  (* apply each argument *)

  apply_arg_e2 code2_gen code1_gen code2 code2_gen' code1_gen' code2' hcode hcode' stack_free stack_free' p p' start base start' base' Rc Rc_cloud stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' stack_list  Halloc Hdealloc Hsubprog Hsubprog_aux Hsubprog' Hsubprog_aux' HypRecE2 Hstack Hstack'.

  intros ptr_b ptr_b'.

  apply_arg_e1 code2_gen code1_gen code2 code1 code2_gen' code1_gen'
  code2' code1' hcode hcode' stack_free stack_free' p p' start base
  start' base' Ra Rc Rc_cloud stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code
  aux_code' stack_list Halloc Hdealloc Hsubprog
  Hsubprog_aux Hsubprog' Hsubprog_aux' b ptr_b ptr_b' HypRecE1.

  match goal with | h : code = (?l1++?l,?stack,?next) |- _ => set (code_pair := l) end; rewrite hcode in Hsubprog.
  match goal with | h : code' = (?l1++?l,?stack,?next) |- _ => set (code_pair' := l) end; rewrite hcode' in Hsubprog'.
  assert (HsubprogPair : PfunctionExtendsFunction p (fragfromprog (progfromlist (code_pair)) (start+length (fst (fst code2))+length (fst (fst code1))) )).
  eapply subPfunction_extends; try apply Hsubprog.
  rewrite <- sub_is_extends; simpl;rewrite <- plus_assoc; rewrite <- app_length;  apply suffix_code.
  unfold code_pair in HsubprogPair ; clear Hsubprog code_pair.
  assert (HsubprogPair' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (code_pair')) (start'+length (fst (fst code2'))+length (fst (fst code1'))) )).
  eapply subPfunction_extends; try apply Hsubprog'.
  rewrite <- sub_is_extends; simpl;rewrite <- plus_assoc; rewrite <- app_length;  apply suffix_code.
  unfold code_pair' in HsubprogPair' ; clear Hsubprog' code_pair'.

  unfold memory_specification; simpl.
  intros ptr_a ptr_a'; unfold code2, code2_gen, code2', code2_gen'; repeat rewrite compile_length_main; fold code2_gen code2 code2_gen' code2'.
  set (new_start := start+length (fst (fst code2))+length (fst (fst code1))) in *.
  set (new_start' := start'+length (fst (fst code2'))+length (fst (fst code1'))) in *.
  (* [sp] <- [sp] - 1         pop *)
  focus 7. focus 8. focus 9.
  do 5 setoid_rewrite stateRelTensor_assoc at 1;  hide_tail...
  focus 3.
  instr_opimmleft minus new_start new_start' 0 HsubprogPair HsubprogPair' (stack_ptr +1+1) (stack_ptr' +1+1) 1 1.
  (* [arg] <- 2              allocate space *)
  replace (stack_ptr +1 + 1 -1) with (stack_ptr+1) ; try omega.  
  focus 1;  instr_moveimm new_start new_start' 1 HsubprogPair HsubprogPair' with nat 2 2.
  (* [ret] <- label + 4 *)
  focus 3;  instr_moveimm new_start new_start' 2 HsubprogPair HsubprogPair' with nat (snd code1+4) (snd code1'+4).
  (* jmp alloc *)
  instr_jumpcst new_start new_start' 3 HsubprogPair HsubprogPair'.
  (* reorganized to apply the allocator*)
  unfold tail; clear tail.
  focus 4; focus 7;focus 8. focus 9. focus 9. focus 11.
  do 5 setoid_rewrite stateRelTensor_assoc at 1; isolate_conj; clear Block; rename Conj_Block into Rc_alloc.
  unfold SPre_alloc, unchange_stack in Halloc.
  hreplace Halloc with (Halloc Rc_alloc 2 2 (stack_ptr+1) (stack_ptr'+1) n n').
  assert (Halloc2 := total_to_partial _ _ _ _ _ _ _ Halloc).
  omega_replace (stack_ptr' + 1 + 1 - 1) with (stack_ptr' + 1).
  focus 1;eapply continuation_enrichment  with (Post := SPost_alloc Ra Rc_alloc 2 2 (stack_ptr+1) (stack_ptr'+1) n n');
  [idtac | eapply judgment_leq; try apply Halloc2; ineq_ring].
  
  unfold SPost_alloc, unchange_stack.
  rewrite <- existsJudgment; intros fb.
  rewrite <- existsJudgment; intros fb'...
  normal_form; focus 1; focus 2.
  unfold Rc_alloc;clear Halloc Halloc2 Rc_alloc...
  compute_code_size code1_gen code2_gen code1 code2; fold new_start.
  compute_code_size code1_gen' code2_gen' code1' code2'; fold new_start'.
  omega_replace (stack_ptr+1-1) with stack_ptr.
  omega_replace (stack_ptr'+1-1) with stack_ptr'.
  omega_replace (stack_ptr+1+1-1-1) with stack_ptr.
  omega_replace (stack_ptr'+1+1-1-1) with stack_ptr'.
  omega_replace (stack_ptr+1+1-1) with (stack_ptr+1).
  omega_replace (stack_ptr'+1+1-1) with (stack_ptr'+1).

  focus 7;focus 8; focus 14.
  do 6 setoid_rewrite stateRelTensor_assoc at 1; hide_tail...
  (*  [[ret]] <- [[sp]]  store expressions *)
  focus 2; focus 1; focus 3;focus 5.
  replace (fb+0) with fb by omega; replace (fb'+0) with fb' by omega.

  instr_moveind new_start new_start' 4 HsubprogPair HsubprogPair' with nat ptr_b ptr_b'.
  (* [[ret]+1] <- [[sp]+1] store expressions *)
  focus 4;focus 3;focus 5;focus 3.
  instr_moveindo new_start new_start' 5 HsubprogPair HsubprogPair' with nat ptr_a ptr_a' on values spreg spreg (stack_ptr +1) (stack_ptr' +1).

  (*  [[sp]] <- [ret]  but the ptr in the stack*)
  focus 5;focus 3.
  instr_moveind new_start new_start' 6 HsubprogPair HsubprogPair' with nat fb fb'.
  (* apply hyp  *)
  assert (start+length (fst (fst code)) = new_start+7).
  unfold new_start;  rewrite hcode; simpl;  do 2 rewrite app_length;  simpl; omega.
  assert (start'+length (fst (fst code')) = new_start'+7).
  unfold new_start';  rewrite hcode'; simpl;  do 2 rewrite app_length;  simpl; omega.
  rewrite <- plus_assoc; rewrite H in hyp; rewrite H0 in hyp; 
  unfold memory_specification in hyp;  simpl in hyp.
  hreplace hyp with (hyp fb fb') ; simpl in hyp.
  eapply judgment_leq; try apply hyp...
  focus' 4... 
  focus' 5.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack.
  apply stateExistsLeq; exists ptr_b.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack.
  apply stateExistsLeq; exists ptr_a.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack.
  apply stateExistsLeq; exists ptr_b'.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack.
  apply stateExistsLeq; exists ptr_a'.
  unfold tail; clear tail...
  focus 9; focus 4.
  setoid_rewrite stateRelTensor_assoc at 1; rewrite stack_description_strength ; rewrite stateCloudPtsot.
  focus 5; setoid_rewrite stateRelTensor_assoc at 1; rewrite stack_description_strength ; rewrite stateCloudPtsot.  
  eapply ppredtensorcompat2.
  omega_replace (stack_ptr+1+1-1-1) with (stack_ptr+1-1);
  omega_replace (stack_ptr'+1+1-1-1) with (stack_ptr'+1-1);
  apply stackdescriptioncompat2; try trivial; try omega.
  ineq_ring_conj.
  focus 2; rewrite ptsto_blank_leq.
  focus 7;  focus 4; rewrite ptsto_blank_leq.
  omega_replace (stack_ptr+1+1+1) with (S(stack_ptr+1+1));  omega_replace (stack_ptr+2) with (stack_ptr+1+1).
  omega_replace (stack_ptr'+1+1+1) with (S(stack_ptr'+1+1)).
  setoid_rewrite stateRelTensor_assoc at 1; rewrite <- block_unfold_start.
  unfold ge in Hstack; assert (2<=stack_free).
  eapply le_trans ; try apply Hstack; apply le_trans with (m := snd(fst code1)+1) ; try apply le_max_r.
  assert (snd(fst code1) >= 1); [ unfold code1, code1_gen, code2_gen ;  rewrite compile_length_aux;
  apply stack_size_positive | omega].
  unfold ge in Hstack'; assert (2<=stack_free').
  eapply le_trans ; try apply Hstack'; apply le_trans with (m := snd(fst code1')+1) ; try apply le_max_r.
  assert (snd(fst code1') >= 1); [ unfold code1', code1_gen', code2_gen' ;  rewrite compile_length_aux;
  apply stack_size_positive | omega].
  omega_replace (S(stack_free-1-1)) with (stack_free-1).
  omega_replace (S(stack_free'-1-1)) with (stack_free'-1).
  ineq_ring.
Qed.




