(* correction_second.v

   This is the partial correctness proof for the second projection for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import correction_compiler_definition.

Typeclasses eauto := dfs 10000.

Lemma correction_of_second : forall base base' Ra alloc alloc' dealloc dealloc' env a b e (t : env|-- e:::a**b), 
(forall Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env (a**b) e t Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3') ->
  forall Rc Rc_cloud start  start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3', 
    statement_of_compiler_correctness base base' Ra alloc alloc' dealloc dealloc' env b (SND e) (TSND t) Rc Rc_cloud start start' stack_free stack_free' stack_ptr stack_ptr' n n2 n3 n' n2' n3'.
Proof with normal_form.
  intros until e; intros t HypRecE1; intros; unfold statement_of_compiler_correctness.
  unfold prog_extends_code, prog_extends_auxcode, extract_code_from_globalcode, extract_stacksize_from_code;
  intros until global_code'; intros hprog hcode Hstack_free hprog' hcode' Hstack_free' Hsubprog_aux Hsubprog Hsubprog_aux' Hsubprog' Halloc Hdealloc.
  split ; [idtac | intros e' Hlambda; inversion Hlambda]; intro hyp.
    
  (* recover the actual piece of code *)
  
  simpl in hprog; surjective_pairing_one_arg e aux_code base alloc dealloc in hprog.
  rewrite <- hcode in Hsubprog; rewrite <- hcode' in Hsubprog';        
  rewrite hprog in hcode; simpl in hcode;
  surjective_pairing_one_arg e code1_gen start in hcode.
  unfold first_code in hcode; simpl in hcode.
  rewrite hprog in Hsubprog_aux; simpl in Hsubprog_aux; clear hprog.
  rewrite hcode in Hstack_free; simpl in Hstack_free.

  simpl in hprog'; surjective_pairing_one_arg e aux_code' base' alloc' dealloc' in hprog'.
  rename code1_gen0 into code1_gen'.
  rewrite hprog' in hcode'; simpl in hcode';
  surjective_pairing_one_arg e code1_gen' start' in hcode'.
  unfold first_code in hcode'; simpl in hcode'.
  rewrite hprog' in Hsubprog_aux'; simpl in Hsubprog_aux'; clear hprog'.
  rewrite hcode' in Hstack_free'; simpl in Hstack_free'.
  rename code0 into code1'.

  (* apply each argument *)

  apply_arg code1_gen code1_gen' code1 code1' hcode hcode' stack_free stack_free' p p' start start' base base' Rc Rc_cloud stack_ptr stack_ptr' n n2 n3 n' n2' n3' aux_code aux_code' stack_list Halloc Hdealloc Hsubprog Hsubprog' Hsubprog_aux Hsubprog_aux' type ptr ptr' Ra HypRecE1.
  
  match goal with | h : code = (?l1++?l,?stack,?next) |- _ => set (code_first := l) end; rewrite hcode in Hsubprog.
  rewrite hcode' in Hsubprog'.
  assert (HsubprogFirst : PfunctionExtendsFunction p (fragfromprog (progfromlist (code_first)) (start+length (fst (fst code1))) )).
  eapply subPfunction_extends; try apply Hsubprog.
  rewrite <- sub_is_extends; simpl; apply suffix_code.
  unfold code_first in HsubprogFirst ; clear Hsubprog; unfold first_code in HsubprogFirst.
  assert (HsubprogFirst' : PfunctionExtendsFunction p' (fragfromprog (progfromlist (code_first)) (start'+length (fst (fst code1'))) )).
  eapply subPfunction_extends; try apply Hsubprog'.
  rewrite <- sub_is_extends; simpl; apply suffix_code.
  unfold code_first in HsubprogFirst' ; clear Hsubprog' code_first; unfold first_code in HsubprogFirst'.
  
  (* [wk] <- [[sp]]        grap the pointer of the pair *)
  intros ptr_pair ptr_pair';  unfold memory_specification; simpl.
  set (new_start := start+length (fst (fst code1))) in *.
  set (new_start' := start'+length (fst (fst code1'))) in *.
  omega_replace (stack_ptr+1-1) with stack_ptr...
  omega_replace (stack_ptr'+1-1) with stack_ptr'...
  focus 5; rewrite <- cloudExP_conj; rewrite <- stateExP_stack;  rewrite <- existsJudgment; intro value2.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack;  rewrite <- existsJudgment; intro value.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack;  rewrite <- existsJudgment; intro value2'.
  rewrite <- cloudExP_conj; rewrite <- stateExP_stack;  rewrite <- existsJudgment; intro value'.
  focus 5;focus 2;focus 6.
  instr_moveimmind new_start new_start' 0 HsubprogFirst HsubprogFirst' with nat ptr_pair ptr_pair' on values spreg spreg (stack_ptr+1) (stack_ptr'+1).
  (* [[sp]] <- [[wk]+1]    push first element (which is actually the second :-) *)
  focus 3; normal_form_conj.
  unfold second_code in *.
  focus 1;focus 3;focus 3.
  instr_moveindindo new_start new_start' 1 HsubprogFirst HsubprogFirst' with nat value2 value2' on values workreg workreg ptr_pair ptr_pair'. 
  (* apply hyp  *)
  clear -hyp hcode hcode'.
  assert (H2 :start+length (fst (fst code)) = new_start+2);
  [unfold new_start;  rewrite hcode; simpl; rewrite app_length;  simpl;  omega | idtac].
  assert (H2':start'+length (fst (fst code')) = new_start'+2);
  [unfold new_start';  rewrite hcode'; simpl; rewrite app_length;  simpl;  omega | idtac].
  rewrite <- plus_assoc; rewrite H2, H2' in hyp.
  unfold memory_specification in hyp;  simpl in hyp.
  hreplace hyp with (hyp value2 value2').
  focus 2; rewrite ptsto_blank_leq.
  eapply judgment_leq; try apply hyp...
  omega_replace (stack_ptr+1-1) with stack_ptr; omega_replace (stack_ptr+2) with (stack_ptr+1+1).
  omega_replace (stack_ptr'+1-1) with stack_ptr'...
  focus 3;focus' 5.
  eapply ppredtensorcompat2 ; try ineq_ring.
  apply stackdescriptioncompat2; trivial;focus_conj 3; do 2 rewrite cloudRelConj_assoc.
  rewrite cloudRelConj_comm; rewrite cloudRelConj_weakening; ineq_ring_conj.
Qed.