Require Import Setoid.

Add LoadPath "~/code_coq/compiler_with_predicative_logic/".

Record stateRel : Type := mkStateRel {
  stateof :> nat -> Prop;
  stateRel_cond : forall k1 k2, 
    stateof k1 -> 
    (k2 <= k1) -> 
    stateof k2}.

Implicit Arguments stateRel_cond [].

Definition stateRelLeq (P Q: stateRel) := forall k , P k -> Q k.

Lemma stateRelLeq_refl : forall (P : stateRel), stateRelLeq P P.
Proof.
  firstorder.
Qed.

Lemma stateRelLeq_trans : forall (P Q R : stateRel), stateRelLeq P Q -> stateRelLeq Q R -> stateRelLeq P R.
Proof.
  firstorder.
Qed.

Hint Resolve stateRelLeq_refl stateRelLeq_trans.

Add Relation stateRel stateRelLeq
   reflexivity proved by stateRelLeq_refl
   transitivity proved by stateRelLeq_trans
 as stateRelLeqRel.

Definition stateRelTensor (P1 P2 : stateRel) : stateRel.
Proof. 
  intros P1 P2.
  refine (mkStateRel (fun k => P1 k /\ P2 k)
                _).
  firstorder.
  apply (stateRel_cond P1 k1 k2); assumption.
  apply (stateRel_cond P2 k1 k2); assumption.
Defined.

Add Morphism stateRelTensor with signature stateRelLeq ==> stateRelLeq ==> stateRelLeq
as ppredtensorcompat.
Proof.
  unfold stateRelLeq, stateRelTensor; simpl; firstorder.
Qed.

Notation "P1 ## P2" := (stateRelTensor P1 P2) (at level 75, right associativity).

Lemma foo : forall p1 p2 p3 p4 p5 p6, stateRelLeq p1 p2 -> 
  stateRelLeq 
  (p6 ## p5 ## p4 ## p3 ## p1)
  (p6 ## p5 ## p4 ## p3 ## p2).
Proof.
  intros.
  rewrite <- H.
  reflexivity.
Qed.

Lemma bar : forall p1 p2 p3 p4 p5 p6 p7, stateRelLeq p1 p2 -> 
  stateRelLeq 
  (p7 ## p6 ## p5 ## p4 ## p3 ## p1)
  (p7 ## p6 ## p5 ## p4 ## p3 ## p2).
Proof.
  intros.
  rewrite <- H.
  reflexivity.
Qed.
  

