(* listP.v

   Define the structure of linked list that describes the environment
   
   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)


Require Import evaluation.
Require Import tools.

(* lsitP : data structure for ntallocator*)


Fixpoint listP  (xs : list nat) (l : nat) {struct xs} : stateRel :=
  match xs with
    | nil => (l |-> 0) ## Topfrom (l+1) (l+1)
    | h::t => Ex y, (stateRel_lift (h > 1) ##(l |-> h) ## (S l |-> y) ## (listP t y) 
      ## (Block (l+2) (l+2) (h-2) (h-2)))
  end.

Fixpoint listseg (xs : list nat) (m n : nat) {struct xs} : stateRel :=
  match xs with
    | nil => stateRel_lift (m=n)
    | h::t => Ex y, (stateRel_lift (h > 1) ## (m |-> y) ## (y|->h) ## (listseg t (S y) n) ## 
      Block (y + 2) (y + 2) (h - 2) (h - 2))
  end.

Lemma listseg_fold : forall xs m n k a,  
                 stateRelLeq (((((listseg xs m n) ## stateRel_lift (a>1)) ## 
                                         (n |->k )) ## ( k |-> a)) ## (Block (k+2) (k+2) (a-2) (a-2)))
                 (listseg (xs ++ (a::nil)) m (S k)).
Proof.
  induction xs.
  unfold listseg; fold listseg; simpl; intuition.
  apply stateExistsLeq_single; exists k.
  normal_form; focus 2; focus 1.
  setoid_rewrite stateRelTensor_assoc.
  setoid_rewrite <- stateRellift_equality.
  normal_form;  elim_head.
  setoid_replace (S k=S k) with True using relation iff; try (intuition; omega).
  fold statePtrue;  ineq_ring.
  
  intuition; simpl; normal_form.
  setoid_rewrite <- stateExP_tensor; trivial.
  apply stateExP_mono; intro x.
  hreplace IHxs with (IHxs (S x) n k a0);  normal_form.
  rewrite <- IHxs.
  ineq_ring.
Qed.


Lemma listseg_to_listP_null : forall xs a b c n ,  
                 stateRelLeq ((listseg xs (S a) b) ## stateRel_lift(n>1) ## (a |-> n ) ## 
                                        ( b |-> c) ## (c |-> 0) ## Topfrom (S c) (S c) 
                                        ## (Block (a+2) (a+2) (n-2) (n-2) ) )
                 (listP (n::xs) a).
Proof.
  induction xs.
  unfold listseg; simpl; intuition.
  apply stateExistsLeq_single; exists c.
  focus 3;focus 1.
  setoid_rewrite stateRelTensor_assoc.
  setoid_rewrite <- stateRellift_equality.
  normal_form; elim_head.
  replace (c+1) with (S c) ; try omega.
  ineq_ring.
  
  intuition.
  unfold listseg; fold listseg.
  set (xsAux := a ::xs).
  unfold listP at 1; fold listP; unfold xsAux.
  setoid_rewrite <- stateExP_tensor; trivial.
  apply stateExP_mono;  intro x; normal_form.
  hreplace IHxs with (IHxs x b c a).
  rewrite <- IHxs.
  ineq_ring.
Qed.

Lemma listseg_to_listP : forall xsHead xs a b c n ,  
                 stateRelLeq ((listseg xsHead (S a) b) ## stateRel_lift(n>1) ## (a |-> n ) ## 
                                        ( b |-> c) ## listP xs c ## (Block (a+2) (a+2) (n-2) (n-2)))
                 (listP ((n::xsHead)++xs) a).
Proof.
  intros xsHead xs.
  dcase xs.
  intros; replace ((n::xsHead)++nil) with (n::xsHead); [idtac | apply app_nil_end].
  unfold listP at 1; normal_form.
  replace (c+1) with (S c) ; try omega.
  apply listseg_to_listP_null.
  
  induction xsHead.
  intros; simpl.
  apply stateExistsLeq_single;  exists c.
  focus 4;  setoid_rewrite <- stateExP_tensor; trivial.
  focus' 4; setoid_rewrite <- stateExP_tensor; trivial.
  apply stateExP_mono;  intro x.
  normal_form.
  focus 8; focus 6.
  setoid_rewrite stateRelTensor_assoc; setoid_rewrite <- stateRellift_equality.
  normal_form; elim_head; ineq_ring.
  
  intros;  unfold listseg; fold listseg.
  rewrite <- app_comm_cons.
  unfold listP at 2; fold listP.
  setoid_rewrite <- stateExP_tensor; trivial.
  apply stateExP_mono; intros x.
  hreplace IHxsHead with (IHxsHead n l H x b c a).  
  rewrite <- IHxsHead.
  ineq_ring.
Qed.

