(* nt_allocator_alloc.v

   This is the total correctness proof for the non-trivial allocator with a free list for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import allocator_spec.
Require Import program_semantics.
Require Import listP.

Typeclasses eauto := dfs 10000.

(* Dealloc spec *)

Lemma ntalloc_judgment : forall p p' alloc alloc' n n5 n5' n6 n6' Rc,
  PfunctionExtendsFunction p (fragfromprog (progfromlist (ntallocinstructions alloc)) alloc) ->
  PfunctionExtendsFunction p' (fragfromprog (progfromlist (ntallocinstructions alloc')) alloc') ->
  total_judgment (SPre_alloc Ra_nt Rc n n n5 n5' n6 n6') 
                     (SPost_alloc Ra_nt Rc n n n5 n5' n6 n6') p p' alloc alloc'.
Proof with normal_form.

  unfold SPre_alloc, Ra_nt; intros until Rc ; intros Hprog Hprog'.
  unfold total_judgment; intros l l'.

  focus 1; rewrite <- existsJudgment_total;  intro xs.
  rewrite <- existsJudgment_total; intro m...
  
  
(* [2] := [0] *)
   
  focus 2;focus 5.
  
  instr_moveimm alloc alloc' 0 Hprog Hprog' with nat l l'.

(* 1: i_brz (s_imm 1) (s_cst (ALLOC + 14)) *)

  focus 5.
  instr_brzimm alloc alloc' 1 Hprog Hprog'.
  
(* 14: [1] == 0  branch *)
  
  intro Hn; subst.
  focus 1.
  instr_jumpimm alloc alloc' 14 Hprog Hprog'.
  

  unfold SPost_alloc, Block.
  rewrite (ptsto_blank_leq workreg); rewrite (ptsto_blank_leq argreg).
  apply total_refl; simpl.
  apply stateExistsLeq; exists l.
  apply stateExistsLeq; exists l'.
  focus' 1.
  apply stateExistsLeq; exists xs.
  apply stateExistsLeq; exists m.
  ineq_ring.

(* [1] =/= 0  branch *)

  intro Hneq.
  
(* 2: i_move (d_imm 4) (s_cst 10) *)

  focus 7.
  instr_moveimm alloc alloc' 2 Hprog Hprog' with nat 10 10.
   
(* 3: i_move (d_imm 0) (s_imm 10) *)

  focus 4;focus 4.
  instr_moveimm alloc alloc' 3 Hprog Hprog' with nat m m.
  
(* induction hypothesis on xs, we have to generalize using listseg*)

  rename m into m'; rename xs into xs'.
  assert (IndHyp : forall xs xsHead m n0,
    explicit_judgment (
      (retreg|-> m) ## (4|-> n0) ## (n0|->m) ## (argreg|->n) ## 
      listseg xsHead 10 n0 ## 
      (workreg|-> {l,l'}(* Later(Perp(SPost_alloc Ra_nt Rc n n n5 n5' n6 n6')) *))
      ## listP xs m ## Rc ## (3|->-) ## unchange_stack n5 n5' n6 n6') 
    (SPost_alloc (Ex xs, Ex m, ((10 |-> m) ## listP xs m)) Rc n n n5 n5' n6 n6')
    p p' (alloc +4) (alloc'+4) l l').
  induction xs.

  intros; unfold listP at 1...
  focus 6;focus 1.

(* 4: i_brz (s_ind 0) (s_cst (ALLOC + 9)) *)

  instr_brzind alloc alloc' 4 Hprog Hprog'.

(* list empty and m = 0 - this branch is always true *)

(* 9: i_add (d_ind 4) (s_imm 1) (s_imm 0) *)    

  intro e; focus 4;focus 4; focus 4.
  instr_opind plus alloc alloc' 9 Hprog Hprog' n n m m.

(* 10: i_move (d_imm 3) (s_ind 4) *)  

  focus 9.

  instr_moveimmind alloc alloc' 10 Hprog Hprog' with nat (n+m) (n+m) on values 4 4 n0 n0.
 
(* 11: i_move (d_ind 3) (s_cst 0) *)

  focus 8.
  rewrite (topfromsplit (S (n-1)) (m+1)); unfold Block;fold Block.
  omega_replace (m+1+(n-1)) with (n+m)...
  focus 3.
  instr_moveind alloc alloc' 11 Hprog Hprog' with nat 0 0.

(* 12: i_branch (s_cst (ALLOC + 14)) *)  

  instr_jumpcst alloc alloc' 12 Hprog Hprog'.
  
(* 14: i_branch (s_imm 2) *)

  focus 10.

  instr_jumpimm alloc alloc' 14 Hprog Hprog'.
  
  rewrite (ptsto_blank_leq workreg); rewrite (ptsto_blank_leq argreg); unfold SPost_alloc; unfold Ra_nt.
  apply total_refl.
  apply stateExistsLeq; exists m.
  apply stateExistsLeq; exists m.
  rewrite (ptsto_blank_leq 3); rewrite (ptsto_blank_leq 4).
  focus 3; focus 9; rewrite (ptsto_blank_leq m).
  setoid_rewrite stateRelTensor_assoc at 1; omega_replace (m+1) with (S m).
  rewrite <- block_unfold_start; omega_replace (S(n-1)) with n.
  focus' 1.
  
(*  case on the shape of XsHead *)
  dcase xsHead.

  intro; unfold listseg.
  focus 6;focus 9.
  setoid_rewrite stateRelTensor_assoc at 1; rewrite <- stateRellift_equality.
  apply stateExistsLeq; exists (@nil nat).
  apply stateExistsLeq; exists (n+m).
  unfold listP...
  elim_head; omega_replace (n+m+1) with (S m +n).
  focus' 10;  ineq_ring.

  intros;  unfold listseg;fold listseg.  
  apply stateExistsLeq; exists xsHead.
  focus 9.
  repeat (setoid_rewrite <- stateExP_tensor at 1). 
  apply stateExP_mono; intro...
  subst. 
  rewrite <- (listseg_to_listP_null l0 x n0 (n+m) n1).
  omega_replace (S (m + n)) with (S (n+m)).
  ineq_ring.

(* list empty and m0 <>  0 - this branch is never true *)

  intro Habsurd; inversion Habsurd.

(* back to the beginning - first iteration through the loop  when list is non-empty *)
 
  intros;  unfold listP; fold listP.
  focus 6; rewrite <- existsJudgment_total;  intro ...
  focus 1;focus 5.  
  
(* 4: i_brz (s_ind 0) (s_cst (ALLOC + 9)) *)  

  instr_brzind alloc alloc' 4 Hprog Hprog'.
    
(* list non-empty and n0 = 0 - this branch is never true *)
  
  intro e; subst.
  focus 2; setoid_replace (0>1) with False using relation iff; try omega.
  rewrite stateRelTensor_zero.
  apply explicitjudgment_false.
  
(* list non-empty and n0 <> 0 - this branch is always true *)

  compare a 1.
  intros; rewrite e; focus 2.
  setoid_replace (1>1) with False using relation iff; try omega.
  rewrite stateRelTensor_zero; apply explicitjudgment_false.

(* we now check if the current element of the free list has required size *)
(* 5: i_eqtst (s_ind 0) (s_imm 1) (s_cst (ALLOC + 13))*)

  intros; focus 8;focus 2;focus 2.
  instr_eqtst alloc alloc' 5 Hprog Hprog'.

(* the current element matches the required size *)
(* 13: i_move (d_ind 4) (s_indo 1 0)*) 
  
  intro;  subst.
  focus 4;focus 1;focus 8;focus 8; omega_replace (S m) with (m +1).
  instr_moveindindo alloc alloc' 13 Hprog Hprog' with nat x x on values retreg retreg m m.
  
(* 14: i_branch (s_imm 2) *)
  
  focus 10.
  instr_jumpimm alloc alloc' 14 Hprog Hprog'.
  unfold SPost_alloc, Block; fold Block.
  rewrite (ptsto_blank_leq workreg); rewrite (ptsto_blank_leq argreg).
  apply total_refl; simpl.

  do 2 (apply stateExistsLeq; exists m)...
  focus 7; setoid_replace (n>1) with True using relation iff; [ idtac | intuition; omega].
  fold statePtrue; setoid_rewrite stateRelTensor_unit at 1.
  focus 8; focus 5.
  rewrite (ptsto_blank_leq (m+1)).
  omega_replace (m+2) with (S(m+1)).
  setoid_rewrite stateRelTensor_assoc at 1; rewrite <- block_unfold_start.
  omega_replace (S(n-2)) with (n-1).
  focus 5; rewrite (ptsto_blank_leq m).
  setoid_rewrite stateRelTensor_assoc at 1; omega_replace (m+1) with (S m); rewrite <- block_unfold_start.
  omega_replace (S(n-1)) with n.

(*  case on the shape of XsHead *)
  dcase xsHead.
  
  intro; unfold listseg; focus' 2.
  apply stateExistsLeq; exists xs.
  apply stateExistsLeq; exists x.
  focus 3;focus 7.
  setoid_rewrite stateRelTensor_assoc at 1; rewrite <- stateRellift_equality...
  elim_head.
  rewrite (ptsto_blank_leq 4);  ineq_ring.
 
  intros; unfold listseg;fold listseg; focus' 2.
  apply stateExistsLeq; exists (xsHead++xs).
  focus 7.
  repeat (setoid_rewrite <- stateExP_tensor at 1). 
  apply stateExP_mono; intro...
  subst. 

  rewrite <- (listseg_to_listP l0 xs x0 n0 x n2).
  rewrite (ptsto_blank_leq 4);  ineq_ring.
 
(* the current element size is not matching *)
(* 6: i_add (d_imm 4) (s_imm 0) (s_cst 1) *) 
 
  intros;  focus 7.
  instr_opimmleft plus alloc alloc' 6 Hprog Hprog' m m 1 1.
    
(* 7: i_move (d_imm 0) (s_ind 4) *)

  focus 5;focus 1;focus 2.
  omega_replace (m+1) with (S m).
  instr_moveimmind alloc alloc' 7 Hprog Hprog' with nat x x on values 4 4 (S m) (S m).

(* 8: i_branch (s_cst (ALLOC + 4))*)

  instr_jumpcst alloc alloc' 8 Hprog Hprog'.

(* finding out the induction hypothesis*)

  specialize (IHxs (xsHead++(a::nil)) x (S m)).
  eapply explicitjudgmentcompat_leq; try apply IHxs; trivial; unfold flip. 
  rewrite <- (listseg_fold xsHead 10 n0 m a).
  ineq_ring.

(* And now we can conclude ! *)

  rename m' into m; rename xs' into xs.
  specialize (IndHyp xs nil m 10).
  unfold listseg in IndHyp; focus_hyp 4.
  setoid_replace (10=10) with True using relation iff in IndHyp ; [idtac|intuition;tauto].
  fold statePtrue in IndHyp.
  eapply explicitjudgmentcompat_leq; try apply IndHyp; trivial; unfold flip.
  ineq_ring.
Qed.

