(* nt_allocator_dealloc.v

   This is the total correctness proof for the non-trivial allocator with a free list for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import allocator_spec.
Require Import program_semantics.
Require Import listP.

(* Dealloc spec *)

Lemma ntdealloc_judgment : forall p p' dealloc dealloc' base size n5 n5' n6 n6' Rc,
  PfunctionExtendsFunction p (fragfromprog (progfromlist (ntdeallocinstructions dealloc)) dealloc) ->
  PfunctionExtendsFunction p' (fragfromprog (progfromlist (ntdeallocinstructions dealloc')) dealloc') ->
  total_judgment (SPre_dealloc Ra_nt Rc base base size size n5 n5' n6 n6') 
                     (SPost_dealloc Ra_nt Rc n5 n5' n6 n6') p p' dealloc dealloc'.
Proof with normal_form.

  unfold SPre_dealloc; intros until Rc ; intros Hprog Hprog'.
  unfold explicit_judgment; intros l l'.
  focus 1.
  
  instr_brzimm dealloc dealloc' 0 Hprog Hprog'.

(******************************************)
(*  [1] == 0   jump (dealloc + 8)  branch *)
(******************************************)

  intro Hsize; subst.
  focus 1.
  instr_jumpimm dealloc dealloc' 8 Hprog Hprog'.

  unfold SPost_dealloc, Block.
  rewrite (ptsto_blank_leq workreg); rewrite (ptsto_blank_leq argreg); rewrite (ptsto_blank_leq retreg).
  apply total_refl; simpl.
  ineq_ring.

(******************************************)
(*  [1] == 0   jump (dealloc + 8)  branch *)
(******************************************)

(* [3] <- [1]*)

  intro Hsize; focus 4.
  instr_moveimm dealloc dealloc' 1 Hprog Hprog' with nat size size.
  
(*   [3] <-  [3] - 1 *)

  instr_opimmleft minus dealloc dealloc' 2 Hprog Hprog' size size 1 1.
  
(* jump [3]  dealloc + 8 *)

  assert (foo := strictPosSucc size Hsize).
  rename size into size'; clear Hsize.
  destruct foo as (size & Hsize); subst.
  omega_replace (S size -1) with size.
  instr_brzimm dealloc dealloc' 3 Hprog Hprog'.
  
(*****************************************)
(*  [3] == 0  jump (dealloc + 8)) branch *)
(*****************************************)

  intro Hsize; subst.
  focus 2.
  instr_jumpimm dealloc dealloc' 8 Hprog Hprog'.
  unfold SPost_dealloc, Block.
  rewrite (ptsto_blank_leq workreg); rewrite (ptsto_blank_leq argreg); rewrite (ptsto_blank_leq retreg); rewrite (ptsto_blank_leq 3); simpl...
  focus 4.
  apply total_refl.
  elim_head.
  ineq_ring.

(*********************)
(* [3] =/= 0  branch *)
(*********************)
  (* [3] <-  [10] *)

  unfold Ra_nt at 1; intro Hsize.
  focus 6.
  do 2 (rewrite <-  existsJudgment_total ; intro)...
  normal_form;  focus 2.
  instr_moveimm dealloc dealloc' 4 Hprog Hprog' with nat x0.
  
(* [10] <-  [2] *)

  focus 5;focus 2.
  instr_moveimm dealloc dealloc' 5 Hprog Hprog' with nat base base.

(* [[10]] <- [1] *)

  focus 7; focus 7.
  rewrite stateRelTensor_assoc at 1.
  rewrite <- block_unfold_end.
  assert (foo := strictPosSucc size Hsize).
  rename size into size'; clear Hsize.
  destruct foo as (size & Hsize); subst.
  rewrite block_unfold_start.
  replace (S base) with (base+1) by omega.
  replace (S size) with (size+1) by omega...
  focus 1.
  replace (size+1) with (S size) by omega.
  setoid_rewrite block_unfold_start...
  focus 7;focus 3;focus 4.
  instr_moveind dealloc dealloc' 6 Hprog Hprog' with nat (S (S size)) (S (S size)).
  
(* [[10]+1] <- [3] *)

  focus 6;focus 4;focus 2.
  instr_moveindo dealloc dealloc' 7 Hprog Hprog' with nat x0 x0 on values 0 0 0 0.

(* branch 0 *) 

  focus 8.
  unfold SPost_dealloc,Ra_nt.
  focus' 5.
  eapply explicitjudgment_leq; [idtac | instr_jumpimm dealloc dealloc' 8 Hprog Hprog'].
  
  (* Dealing with the continuation *)

  apply stateExistsLeq; exists ((S (S size))::x) ;  apply stateExistsLeq; exists base.
  reflexivity.
  rewrite (ptsto_blank_leq retreg);rewrite (ptsto_blank_leq argreg);rewrite (ptsto_blank_leq 3).
  rewrite (ptsto_blank_leq workreg).
  omega_replace (S(base+1)) with (base + 2 ).
  unfold listP;fold listP...
  focus' 1.
  apply total_refl.
  apply stateExistsLeq; exists x0...
  setoid_replace (S (S size) >1) with True using relation iff by (intuition;omega).
  fold statePtrue.
  omega_replace (base+1) with (S base).
  omega_replace (size - 0) with size.
  ineq_ring.
Qed.
