(* nt_allocator_init.v

   This is the total correctness proof for the non-trivial allocator with a free list for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Import allocator_spec.
Require Import program_semantics.
Require Import listP.

(* Init spec *)

Lemma ntinit_judgment : forall p p' init init' n5 n5' n6 n6',
  PfunctionExtendsFunction p (fragfromprog (progfromlist (ntinitinstructions init)) init) ->
  PfunctionExtendsFunction p' (fragfromprog (progfromlist (ntinitinstructions init')) init') ->
  total_judgment (SPre_init Ra_nt n5 n5' n6 n6') 
                     (SPost_init Ra_nt n5 n5' n6 n6') p p' init init'.
Proof with normal_form.

  unfold SPre_init; intros until n6'; intros Hprog Hprog'.
  unfold total_judgment; intros l l'.

  setoid_rewrite (topfromsplit 2 10).
  unfold Block; simpl; omega_replace (10+2) with 12; omega_replace (10+1) with 11...
  focus 7.

  (* [10] := 11 *)

  instr_moveimm init init' 0 Hprog Hprog' with nat 11 11.

  (* [[10]] := [0] *)
  
  focus 7; focus 1.
  instr_moveind init init' 1 Hprog Hprog' with nat 0 0.
  unfold SPost_init, Ra_nt.
  (* jump [0] *)

  focus 2.

  eapply explicitjudgment_leq; [idtac | instr_jumpimm init init' 2 Hprog Hprog'].
  
  (* Dealing with the continuation *)

  apply stateExistsLeq; exists (@nil nat);  apply stateExistsLeq; exists 11.
  reflexivity.
  rewrite ptsto_blank_leq; simpl.
  apply total_refl.
  ineq_ring.
Qed.
