(* nt_allocator_proof.v

   This is the total correctness proof for the non-trivial allocator with a free list for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)


Require Import allocator_spec.
Require Import program_semantics.
Require Import nt_allocator_init.
Require Import nt_allocator_alloc.
Require Import nt_allocator_dealloc.

Definition allocator_spec_diag Ra init init' alloc alloc' dealloc dealloc':=
 fun (p p' : program) => 
   (forall n5 n5' n6 n6',  total_judgment (SPre_init Ra n5 n5' n6 n6') 
     (SPost_init Ra n5 n5' n6 n6') p p' init init') /\
   (forall Rc_alloc size n5 n5' n6 n6',  total_judgment (SPre_alloc Ra Rc_alloc size size n5 n5' n6 n6') 
     (SPost_alloc Ra Rc_alloc size size n5 n5' n6 n6') p p' alloc alloc') /\
   (forall Rc_alloc size base n5 n5' n6 n6', total_judgment (SPre_dealloc Ra Rc_alloc base base size size n5 n5' n6 n6')
     (SPost_dealloc Ra Rc_alloc n5 n5' n6 n6') p p' dealloc dealloc').
 
Definition nt_allocator_program init alloc dealloc : program_fragment := 
                  (fragfromprog (progfromlist (ntinitinstructions init)) init) ;;
                  (fragfromprog (progfromlist (ntallocinstructions alloc)) alloc) ;;
                  (fragfromprog (progfromlist (ntdeallocinstructions dealloc)) dealloc).

Theorem na_allocator_spec : forall init init' alloc alloc' dealloc dealloc' p p',
  let nt_init := fragfromprog (progfromlist (ntinitinstructions init)) init in 
  let nt_alloc := fragfromprog (progfromlist (ntallocinstructions alloc)) alloc in
  let nt_dealloc := fragfromprog (progfromlist (ntdeallocinstructions dealloc)) dealloc in
  let nt_program := nt_allocator_program init alloc dealloc in
  let nt_init' := fragfromprog (progfromlist (ntinitinstructions init')) init' in 
  let nt_alloc' := fragfromprog (progfromlist (ntallocinstructions alloc')) alloc' in
  let nt_dealloc' := fragfromprog (progfromlist (ntdeallocinstructions dealloc')) dealloc' in
  let nt_program' := nt_allocator_program init' alloc' dealloc' in
  nt_init <#> nt_alloc ->
  nt_alloc <#> nt_dealloc ->
  nt_init <#> nt_dealloc ->
  nt_init' <#> nt_alloc' ->
  nt_alloc' <#> nt_dealloc' ->
  nt_init' <#> nt_dealloc' ->
  PfunctionExtendsFunction p nt_program-> 
  PfunctionExtendsFunction p' nt_program'-> 
  exists Ra, allocator_spec_diag Ra init init' alloc alloc' dealloc dealloc' p p'.
Proof with resolve_frag.
  intros.
  unfold allocator_spec_diag.
  exists Ra_nt; intros.
  split.

(* initialisation  *)

  assert (Hinit : nt_init << nt_program).
  unfold nt_program, nt_allocator_program in *.
  apply extends_join...
  eapply subPfunction_extends in H5;[idtac | eapply sub_is_extends; try apply Hinit].
  assert (Hinit' : nt_init' << nt_program').
  unfold nt_program', nt_allocator_program in *.
  apply extends_join...
  eapply subPfunction_extends in H6;[idtac | eapply sub_is_extends; try apply Hinit'].
  intros; apply ntinit_judgment; try assumption.
  
  split.

(* allocation *)

  assert (Halloc : nt_alloc << nt_program).
  unfold nt_program, nt_allocator_program, nt_alloc...
  rewrite join_disj_comm...
  rewrite <- disj_join_assoc...
  apply extends_join...
  eapply subPfunction_extends in H5;[idtac | eapply sub_is_extends; try apply Halloc].
  assert (Halloc' : nt_alloc' << nt_program').
  unfold nt_program', nt_allocator_program, nt_alloc'...
  rewrite join_disj_comm...
  rewrite <- disj_join_assoc...
  apply extends_join...
  eapply subPfunction_extends in H6;[idtac | eapply sub_is_extends; try apply Halloc'].

  intros; apply ntalloc_judgment; try assumption.

(* deallocation *)

  assert (Hdealloc : nt_dealloc << nt_program).
  unfold nt_program, nt_allocator_program...
  rewrite disj_join_assoc ...
  rewrite join_disj_comm ...
  apply extends_join...
  eapply subPfunction_extends in H5;[idtac | eapply sub_is_extends; try apply Hdealloc].
  assert (Hdealloc' : nt_dealloc' << nt_program').
  unfold nt_program', nt_allocator_program...
  rewrite disj_join_assoc ...
  rewrite join_disj_comm ...
  apply extends_join...
  eapply subPfunction_extends in H6;[idtac | eapply sub_is_extends; try apply Hdealloc'].

  intros; apply ntdealloc_judgment; try assumption.
Qed.

  


