(* program_semantics.v

   Lemmas describing the trnasformation of the relational description
   of the stack after each basic instruction of assembly code.
   (Both for partial and total judgments)

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export evaluation.

Ltac unfold_kstepterm := unfold kstepterm; simpl; fold kstepterm.

Ltac rewrite_start start start' offset := 
   replace (start + offset + 1) with (start + (offset + 1));[simpl | omega];
   replace (start' + offset + 1) with (start' + (offset + 1));[simpl | omega].

Tactic Notation "unfold_kstepterm" "in" ident(H) := 
  unfold kstepterm in H; simpl in H; fold kstepterm in H.

Ltac interp_cond :=
  unfold Interp in *; eapply stateRel_cond; 
  [eassumption | apply subPfunction_refl | apply subPfunction_refl | omega].

Ltac interp_light :=  unfold Interp in *; eapply stateRel_cond;
  [idtac | apply subPfunction_refl | apply subPfunction_refl | idtac].


Ltac try_step j k H := 
  induction j as [ |j IHj]; [split; intro Habsurd; inversion Habsurd | clear IHj];
  let id := fresh in assert (id : j <= k); [omega | idtac] ; clear H; rename id into H.

Ltac rewrite_start_succ start start':=  
  replace (start+1) with (S start) by omega;
  replace (start'+1) with (S start') by omega.

Ltac unfold_judgment start start' Hinst Hinst' Hsubprog Hsubprog' :=
  unfold judgment, derivation, Perp; simpl;
  unfold perpOf; intro k; intros s s' j Hineq Hpre;
  assert (Hsubprog_start := (Hsubprog start));
  rewrite Hinst in Hsubprog_start;
  assert (Hsubprog_start' := (Hsubprog' start'));
  rewrite Hinst' in Hsubprog_start';
  try_step j k Hineq; unfold_kstepterm;
  rewrite Hsubprog_start,Hsubprog_start'; simpl.

Ltac unfold_sem start start' Hinst Hinst' Hsubprog Hsubprog' :=
  assert (Hsubprog_start := (Hsubprog start));
  rewrite Hinst in Hsubprog_start;
  assert (Hsubprog_start' := (Hsubprog' start'));
  rewrite Hinst' in Hsubprog_start'.

Tactic Notation "conclude" "using" "pretac" tactic(Hbefore) tactic(Hbefore2) "posttac" tactic(Hafter) tactic(Hafter2) 
  "with" ident(Hm) ident(Hm') ident(k0) ident(Hnext) ident(Hnext') ident(Hsubprog_start) ident(Hsubprog_start') :=
  unfold terminates in *; split ; intro Hterm;
    [ Hbefore ; hreplace Hnext with (Hnext Hterm);
      destruct Hnext as (k0,Hnext);
      exists (S k0); 
      unfold_kstepterm; rewrite Hsubprog_start'; simpl;
      Hafter; assumption |
      Hbefore2; hreplace Hnext' with (Hnext' Hterm);
      destruct Hnext' as (k0,Hnext');
      exists (S k0); 
      unfold_kstepterm; rewrite Hsubprog_start; simpl;
      Hafter2 ; assumption].

(*  instruction MOVE *)

    
Lemma sem_moveimm : forall pf pf' p p' start start' m m' Rpre P src src' Q,
  pf start = Some(i_move (d_imm m) src) ->
  pf' start' = Some(i_move (d_imm m') src') ->
  (forall s s' p p' k,  Interp ({m,m'} |=>P ## Rpre) s s' p p'  k ->  
    (natrel Q p p' (sem_src src s) (sem_src src' s') k)) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (forall s1 s1' j, Interp ({m,m'}|=>P ## Rpre) s1 s1' p p' j ->
    Interp ({m,m'}|=>Q ## Rpre) (update s1 m (sem_src src s1)) (update s1' m' (sem_src src' s1')) p p' j /\
    sem_instr (p start) s1 start = Some (update s1 m (sem_src src s1) , start+1) /\
    sem_instr (p' start') s1' start' = Some (update s1' m' (sem_src src' s1') , start'+1)).
Proof.
  intros until Q;  rewrite_start_succ start start'.
  intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  intros until j; intros Hinterp.
  split;
  set (next_state := update s1 m (sem_src src s1)) in *;
  set (next_state' := update s1' m' (sem_src src' s1')) in *;
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.


  hreplace Hinstr2 with (Hinstr2 s1 s1' p p' j Hinterp).
  interp_light. 
  unfold next_state, next_state'; apply indepupdate ; try eassumption.
  rewrite <- (ptsto_blank_leq m m' P); assumption.
  omega.

  simpl; rewrite Hsubprog_start, Hsubprog_start'; simpl.
  intuition.
Qed.

Lemma moveimm_total : forall pf pf' p p' start start' m m' Rpre Rpost lpost lpost' P source source' Q,
  pf start = Some(i_move (d_imm m) source) ->
  pf' start' = Some(i_move (d_imm m') source') ->
  (forall s s' p p' k,  Interp ({m,m'} |=>P ## Rpre) s s' p p'  k ->  
    (natrel Q p p' (sem_src source s) (sem_src source' s') k)) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  explicit_judgment ({m,m'}|=> Q ## Rpre) Rpost p p' (start+1) (start'+1) lpost lpost' -> 
  explicit_judgment ({m,m'}|=> P ## Rpre) Rpost p p' start start' lpost lpost'.
Proof.
  intros until Q; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  eapply sem_total_src with (m:=m) (m':=m').
  eapply sem_moveimm; try eassumption.
Qed.

Lemma moveimm_partial : forall pf pf' p p' start start' m m' Rpre P source source' Q,
  pf start = Some(i_move (d_imm m) source) ->
  pf' start' = Some(i_move (d_imm m') source') ->
  (forall s s' p p' k,  Interp ({m,m'} |=>P ## Rpre) s s' p p'  k ->  
    (natrel Q p p' (sem_src source s) (sem_src source' s') k)) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment ({m,m'} |=> Q ##  Rpre) p p' (start+1) (start'+1) ->
  judgment ({m,m'} |=>P ##  Rpre) p p' start start'.
Proof.
  intros until Q; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  apply sem_partial_src with (m:=m) (m':=m') (source:=source) (source':=source').
  simple eapply sem_moveimm; try eassumption.
Qed.

Ltac destruct_moveimm := 
  intros until k; intro Hpost1; simpl sem_src; clear - Hpost1; 
  first [(simpl; split ; split ; reflexivity) | 
         (simple eapply ptstointerp; focus_hyp 1; simple apply Hpost1)].

Tactic Notation (at level 0) "instr_moveimm" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "nat" constr(n) := 
	first [eapply moveimm_partial with (Q:=(natPconstant n)) | 
	       eapply moveimm_total with (Q:=(natPconstant n))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveimm) | 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveimm" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "nat" constr(n) constr(n') := 
        first [eapply moveimm_partial with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' )) | 
	       eapply moveimm_total with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' ))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveimm) | 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveimm" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "Rel" constr(Rel) := 
	first [eapply moveimm_partial with (Q:=Rel) | 
	       eapply moveimm_total with (Q:=Rel)]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveimm) | 
	 rewrite_start start start' offset ].

Ltac destruct_moveimmind a a' b b' := 
  intros s s'; intros until k; intro Hpost1; simpl sem_src; clear - Hpost1;
  first [(simpl; split ; split; reflexivity) |
         simple eapply ptstointerp; let foo := fresh in assert (foo : s a = b /\ s' a' = b');
	 [focus_hyp 1;simple eapply ptsto_stateval; simple apply Hpost1 | idtac];
         destruct foo as (foo,foo'); rewrite foo, foo';
         focus_hyp 2; simple apply Hpost1;(focus_hyp 1; simple apply Hpost1)]. 

Tactic Notation (at level 0) "instr_moveimmind" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "nat" constr(n) "on" "values" constr(a) constr(a') constr(b) constr(b'):= 
	first [eapply moveimm_partial with (Q:=(natPconstant n)) | 
	       eapply moveimm_total with (Q:=(natPconstant n))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveimmind a a' b b')| 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveimmind" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "nat" constr(n) constr(n') "on" "values" constr(a) constr(a') constr(b) constr(b'):= 
        first [eapply moveimm_partial with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' )) | 
	       eapply moveimm_total with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' ))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveimmind a a' b b')| 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveimmind" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "rel" constr(Rel) "on" "values" constr(a) constr(a') constr(b) constr(b'):= 
	first [eapply moveimm_partial with (Q:=Rel) | 
	       eapply moveimm_total with (Q:=Rel)]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveimmind a a' b b') | 
	 rewrite_start start start' offset ].

Ltac destruct_moveimmindo a a' b b' := 
  intros s s'; intros until k; intro Hpost1; simpl sem_src; clear - Hpost1;
  first [
  (simpl; split ; split ; reflexivity) | 
  let foo := fresh in assert (foo : s a = b /\ s' a' = b');
    [ simple eapply ptsto_stateval; 
      first[simple apply Hpost1 | focus_hyp 1;simple apply Hpost1]  | idtac];
    destruct foo as (foo,foo'); rewrite foo, foo';
  try first [
    (simple eapply ptstointerp; focus_hyp 1; simple apply Hpost1) |
    (simple eapply ptstointerp; focus_hyp 2; simple apply Hpost1) |
    (simple eapply ptstointerp_cloud; focus_hyp 1; simple apply Hpost1) |
    (simple eapply ptstointerp_cloud; focus_hyp 2; simple apply Hpost1) |
    (simple eapply ptstointerp_stack; focus_hyp 1; simple apply Hpost1) |
    (simple eapply ptstointerp_stack; focus_hyp 2; simple apply Hpost1)]].

Tactic Notation (at level 0) "instr_moveimmindo" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "nat" constr(n) "on" "values" constr(a) constr(a') constr(b) constr(b'):= 
	first [eapply moveimm_partial with (Q:=(natPconstant n)) | 
	       eapply moveimm_total with (Q:=(natPconstant n))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveimmindo a a' b b')| 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveimmindo" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "nat" constr(n) constr(n') "on" "values" constr(a) constr(a') constr(b) constr(b'):= 
        first [eapply moveimm_partial with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' )) | 
	       eapply moveimm_total with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' ))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveimmindo a a' b b')| 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveimmindo" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "rel" constr(Rel) "on" "values" constr(a) constr(a') constr(b) constr(b'):= 
	first [eapply moveimm_partial with (Q:=Rel) | 
	       eapply moveimm_total with (Q:=Rel)]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveimmindo a a' b b')| 
	 rewrite_start start start' offset ].

Lemma sem_moveind : forall pf pf' p p' start start' m n n' P source source' Rpre Q,
  pf start = Some(i_move (d_ind m) source) ->
  pf' start' = Some(i_move (d_ind m) source') ->
  (forall s s' p p' k,  Interp ((m |->{n,n'}) ## ({n,n'}|=>P) ## Rpre) s s' p p'  k ->  
    (natrel Q p p' (sem_src source s) (sem_src source' s') k)) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (forall s1 s1' j, Interp ((m |->{n,n'}) ## ({n,n'}|=>P) ## Rpre) s1 s1' p p' j ->
    Interp ((m |->{n,n'}) ## ({n,n'}|=>Q) ## Rpre) (update s1 n (sem_src source s1)) (update s1' n' (sem_src source' s1')) p p' j /\
    sem_instr (p start) s1 start = Some (update s1 n (sem_src source s1) , start+1) /\
    sem_instr (p' start') s1' start' = Some (update s1' n' (sem_src source' s1') , start'+1)).
Proof.
  intros until Q;  rewrite_start_succ start start'.
  intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  intros until j; intros Hinterp.
  assert (Hstate :s1 m = n /\ s1' m = n').
  eapply ptsto_stateval; apply Hinterp.
  split;
  set (next_state := update s1 m (sem_src source s1)) in *;
  set (next_state' := update s1' m (sem_src source' s1')) in *;
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.


 setoid_rewrite stateRelTensor_assoc; rewrite (stateRelTensor_comm (m |-> {n,n'})); normal_form.
  hreplace Hinstr2 with (Hinstr2 s1 s1' p p' j Hinterp).
  interp_light. 
  unfold next_state, next_state'; apply indepupdate ; try eassumption.
  rewrite <- (ptsto_blank_leq n n' P).
  setoid_rewrite stateRelTensor_assoc; rewrite (stateRelTensor_comm ({n,n'} |=> P)); normal_form.
  assumption.
  trivial.

  simpl; rewrite Hsubprog_start, Hsubprog_start'; simpl.
  destruct Hstate as (Hm,Hm'); rewrite Hm,Hm'.
  intuition.
Qed.

Lemma moveind_total : forall pf pf' p p' start start' l l' m n n' P source source' Rpre Rpost Q,
  pf start = Some(i_move (d_ind m) source) ->
  pf' start' = Some(i_move (d_ind m) source') ->
  (forall s s' p p' k,  Interp ((m |->{n,n'}) ## ({n,n'}|=>P) ## Rpre) s s' p p'  k ->  
    (natrel Q p p' (sem_src source s) (sem_src source' s') k)) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  explicit_judgment ((m |-> {n,n'}) ## ({n,n'} |=> Q) ##  Rpre) Rpost p p' (start+1) (start'+1) l l' ->
  explicit_judgment ((m |-> {n,n'}) ## ({n,n'} |=> P) ##  Rpre) Rpost p p' start start' l l'.
Proof.
  intros until Q; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  apply sem_total_src with (m:=n) (m':=n') (source:=source) (source':=source').
  eapply sem_moveind; try eassumption.
Qed.


Lemma moveind_partial : forall pf pf' p p' start start' m n n' P source source' Pinv Q,
  pf start = Some(i_move (d_ind m) source) ->
  pf' start' = Some(i_move (d_ind m) source') ->
  (forall s s' p p' k,  Interp ((m |->{n,n'}) ## ({n,n'}|=>P) ## Pinv) s s' p p'  k ->  
    (natrel Q p p' (sem_src source s) (sem_src source' s') k)) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment ((m |-> {n,n'}) ## ({n,n'} |=> Q) ##  Pinv) p p' (start+1) (start'+1) ->
  judgment ((m |-> {n,n'}) ## ({n,n'} |=> P) ##  Pinv) p p' start start'.
Proof.
  intros until Q; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  apply sem_partial_src with (m:=n) (m':=n') (source:=source) (source':=source').
  eapply sem_moveind; try eassumption.
Qed.

Ltac destruct_moveind := 
  intros until k; intro Hpost1; simpl sem_src; clear - Hpost1;
  try first [simpl; split; split ; reflexivity | 
             simple eapply ptstointerp; focus_hyp 2; simple apply Hpost1 |
             focus_hyp 3; focus_hyp 3;
             match goal with 
             | H:Interp (?P ## ({?ptr,?ptr'} |-> {?a,?a'}) ## ?R)  ?s ?s' ?p ?p' ?k   |- _ => 
               simple eapply ptstointerp_ind      
             | H:Interp (?P ## cloudToState ?Q ## ?R)  ?s ?s' ?p ?p' ?k   |- _ => 
               simple eapply ptstointerp_ind_cloud
             | H:Interp (?P ## stack_description ?l ?ptr ?ptr' ?Q ## ?R)  ?s ?s' ?p ?p' ?k   |- _ => 
               simple eapply ptstointerp_ind_stack
             end;
             simple apply Hpost1].

Tactic Notation (at level 0) "instr_moveind" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "nat" constr(n) := 
	first [eapply moveind_partial with (Q:=(natPconstant n)) | 
	       eapply moveind_total with (Q:=(natPconstant n))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveind) | 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveind" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "nat" constr(n) constr(n') := 
        first [eapply moveind_partial with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' )) | 
	       eapply moveind_total with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' ))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveind) | 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveind" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
 "with" "rel" constr(Rel) := 
	first [eapply moveind_partial with (Q:=Rel) | 
	       eapply moveind_total with (Q:=Rel)]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         (destruct_moveind)| 
	 rewrite_start start start' offset ].

Ltac destruct_moveindindo a a' b b':= 
  intros s s'; intros until k; intro Hpost1; simpl sem_src;
  first [simpl; split ; split ; reflexivity | 
    clear - Hpost1; let foo := fresh in assert (foo : s a = b /\ s' a' = b');
    [simple eapply ptsto_stateval;first [ apply Hpost1 | focus_hyp 2 ; simple apply Hpost1] | idtac];
    destruct foo as (foo,foo'); rewrite foo, foo';
    try first [
         simple eapply ptstointerp; focus_hyp 3; simple apply Hpost1 |
         simple eapply ptstointerp_stack; focus_hyp 3; simple apply Hpost1 ]].

Tactic Notation (at level 0) "instr_moveindindo" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
    "with" "nat" constr(n) "on" "values" constr(a) constr(a') constr(b) constr(b'):= 
	first [eapply moveind_partial with (Q:=(natPconstant n)) | 
	       eapply moveind_total with (Q:=(natPconstant n))]; 
        try apply Hsubprog ;try apply Hsubprog' ;
	[instrextract start offset |instrextract start' offset | (destruct_moveindindo a a' b b')| 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveindindo" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
    "with" "nat" constr(n) constr(n') "on" "values" constr(a) constr(a') constr(b) constr(b'):= 
        first [eapply moveind_partial with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' )) | 
	       eapply moveind_total with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' ))]; 
        try apply Hsubprog ;try apply Hsubprog' ;
	[instrextract start offset |instrextract start' offset | 
          (destruct_moveindindo a a' b b')| 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveind" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog')
"with" "rel" constr(Rel) "on" "values" constr(a) constr(a') constr(b) constr(b'):= 
	first [eapply moveind_partial with (Q:=Rel) | 
	       eapply moveind_total with (Q:=Rel)]; 
	try apply Hsubprog ;try apply Hsubprog' ;
	[instrextract start offset |instrextract start' offset | 
          (destruct_moveindindo a a' b b') | 
	 rewrite_start start start' offset ].

Lemma sem_moveindo : forall pf pf' p p' start start' m n n' cst P source source' Rpre Q,
  pf start = Some(i_move (d_indo cst m) source) ->
  pf' start' = Some(i_move (d_indo cst m) source') ->
  (forall s s' p p' k,  Interp ((m |->{n,n'}) ## ({n+cst,n'+cst}|=>P) ## Rpre) s s' p p'  k ->  
    (natrel Q p p' (sem_src source s) (sem_src source' s') k)) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (forall s1 s1' j, Interp ((m |->{n,n'}) ## ({n+cst,n'+cst}|=>P) ## Rpre) s1 s1' p p' j ->
    Interp ((m |->{n,n'}) ## ({n+cst,n'+cst}|=>Q) ## Rpre) (update s1 (n+cst) (sem_src source s1)) (update s1' (n'+cst) (sem_src source' s1')) p p' j /\
    sem_instr (p start) s1 start = Some (update s1 (n+cst) (sem_src source s1) , start+1) /\
    sem_instr (p' start') s1' start' = Some (update s1' (n'+cst) (sem_src source' s1') , start'+1)).
Proof.
  intros until Q;  rewrite_start_succ start start'.
  intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  intros until j; intros Hinterp.
  assert (Hstate :s1 m = n /\ s1' m = n').
  eapply ptsto_stateval; apply Hinterp.
  split;
  set (next_state := update s1 m (sem_src source s1)) in *;
  set (next_state' := update s1' m (sem_src source' s1')) in *;
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.


 setoid_rewrite stateRelTensor_assoc; rewrite (stateRelTensor_comm (m |-> {n,n'})); normal_form.
  hreplace Hinstr2 with (Hinstr2 s1 s1' p p' j Hinterp).
  interp_light. 
  unfold next_state, next_state'; apply indepupdate ; try eassumption.
  rewrite <- (ptsto_blank_leq (n+cst) (n'+cst) P).
  setoid_rewrite stateRelTensor_assoc; rewrite (stateRelTensor_comm ({n+cst,n'+cst} |=> P)); normal_form.
  assumption.
  trivial.

  simpl; rewrite Hsubprog_start, Hsubprog_start'; simpl.
  destruct Hstate as (Hm,Hm'); rewrite Hm,Hm'.
  intuition.
Qed.

Lemma moveindo_total : forall pf pf' p p' start start' l l' m n n' cst P source source' Rpre Rpost Q,
  pf start = Some(i_move (d_indo cst m) source) ->
  pf' start' = Some(i_move (d_indo cst m) source') ->
  (forall s s' p p' k,  Interp ((m |->{n,n'}) ## ({n+cst,n'+cst}|=>P) ## Rpre) s s' p p'  k ->  
    (natrel Q p p' (sem_src source s) (sem_src source' s') k)) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  explicit_judgment ((m |-> {n,n'}) ## ({n+cst,n'+cst} |=> Q) ##  Rpre) Rpost p p' (start+1) (start'+1) l l' ->
  explicit_judgment ((m |-> {n,n'}) ## ({n+cst,n'+cst} |=> P) ##  Rpre) Rpost p p' start start' l l'.
Proof.
  intros until Q; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  apply sem_total_src with (m:=n+cst) (m':=n'+cst) (source:=source) (source':=source').
  eapply sem_moveindo; try eassumption.
Qed.

Lemma moveindo_partial : forall pf pf' p p' start start' m n n' cst P source source' Rpre Q,
  pf start = Some(i_move (d_indo cst m) source) ->
  pf' start' = Some(i_move (d_indo cst m) source') ->
  (forall s s' p p' k,  Interp ((m |->{n,n'}) ## ({n+cst,n'+cst}|=>P) ## Rpre) s s' p p'  k ->  
    (natrel Q p p' (sem_src source s) (sem_src source' s') k)) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment ((m |-> {n,n'}) ## ({n+cst,n'+cst} |=> Q) ##  Rpre) p p' (start+1) (start'+1) ->
  judgment ((m |-> {n,n'}) ## ({n+cst,n'+cst} |=> P) ##  Rpre) p p' start start'.
Proof.
  intros until Q; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  apply sem_partial_src with (m:=n+cst) (m':=n'+cst) (source:=source) (source':=source').
  eapply sem_moveindo; try eassumption.
Qed.

Ltac destruct_moveindo a a' b b' := 
  intros s s'; intros until k; intro Hpost1; simpl sem_src; clear - Hpost1;
  first [(simpl; split ; split ; reflexivity)  | 
         (simpl; focus_hyp 2; simple eapply ptsto_stateval; simple apply Hpost1) |
    let foo := fresh in assert (foo : s a = b /\ s' a' = b');
    [simple eapply ptsto_stateval; focus_hyp 2; simple apply Hpost1 | idtac];
    destruct foo as (foo,foo'); rewrite foo; rewrite  foo';
    try first [
    simple eapply ptstointerp; focus_hyp 2; simple apply Hpost1 | 
    simple eapply ptstointerp; focus_hyp 3; simple apply Hpost1 |
    simple eapply ptstointerp_cloud; focus_hyp 3;simple apply Hpost1 |
    simple eapply ptstointerp_stack; focus_hyp 3;simple apply Hpost1]].


Tactic Notation (at level 0) "instr_moveindo" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog') 
    "with" "nat" constr(n) "on" "values" constr(a) constr(a') constr(b') constr(b) :=  
	first [eapply moveindo_partial with (Q:=(natPconstant n)) | 
	       eapply moveindo_total with (Q:=(natPconstant n))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset |instrextract start' offset |
         (destruct_moveindo a a' b b') | 
	 rewrite_start start start' offset ].

Tactic Notation (at level 0) "instr_moveindo" 
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog') 
    "with" "nat" constr(n) constr(n') "on" "values" constr(a) constr(a') constr(b) constr(b'):=  
        first [eapply moveindo_partial with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' )) | 
	       eapply moveindo_total with 
                (Q:=natPconstantRel  (fun l => fun l' => l=n /\ l'=n' ))]; 
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset |instrextract start' offset |
         (destruct_moveindo a a' b b') | 
	 rewrite_start start start' offset ].

Lemma sem_jumpimm : forall pf pf' p p' start start' m m' Rpre,
  pf start =  Some(i_branch (s_imm m))  ->
  pf' start' = Some(i_branch (s_imm m')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (forall n n' s1 s1' j, Interp ({m,m'} |->{n,n'} ## Rpre) s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, n) /\
    sem_instr (p' start') s1' start' = Some (s1', n')).
Proof.
  intros until Rpre.
  intros Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  intros until j; intros Hinterp.
  assert (Hstate :s1 m = n /\ s1' m' = n').
  eapply ptsto_stateval; apply Hinterp.
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  rewrite Hsubprog_start, Hsubprog_start'; simpl.
  destruct Hstate as (Hstate & Hstate').
  rewrite Hstate, Hstate'; intuition.
Qed.

Lemma jumpimm_total : forall pf pf' p p' start start' lpost lpost' l l' m m' Rpost Rpre,
  pf start =  Some(i_branch (s_imm m))  ->
  pf' start' = Some(i_branch (s_imm m')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  explicit_judgment ({m,m'} |-> {l,l'} ## Rpre) Rpost p p' l l' lpost lpost' ->
  explicit_judgment ({m,m'} |-> {l,l'} ## Rpre) Rpost p p' start start' lpost lpost'.
Proof.
  intros until Rpre;  intros Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  eapply sem_total2.
  eapply sem_jumpimm; try eassumption.
Qed.

Lemma jumpimm_partial : forall pf pf' p p' start start' m Rpre,
  pf start =  Some(i_branch (s_imm m))  ->
  pf' start' = Some(i_branch (s_imm m)) ->
  stateRelLeq Rpre (m |=> Later(Perp Rpre)) -> 
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment Rpre p p' start start'.
Proof.
  intros until Rpre; intros Hinstr1 Hinstr1' Hnext Hsubprog Hsubprog'.
  unfold_judgment start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  hreplace Hnext with (Hnext (functionToPfunction s) (functionToPfunction s') p p' (S j) Hpre).
  simpl in Hnext; destruct Hnext as (l & l' & Hl & Hl' & Hnext).
  inversion Hl;inversion Hl'.
  assert (Hj : j < S j); try omega.
  assert (Hj2 : j <= j); try omega.
  hreplace Hnext with (Hnext j Hj s s' j Hj2).
  destruct Hnext as (Hnext,Hnext').
  interp_cond.
  subst.

  conclude using pretac (idtac) (idtac) posttac (idtac) (idtac) with Hm Hm' k0 Hnext Hnext' Hsubprog_start Hsubprog_start'. 
Qed.

Ltac instr_jumpimm start start' offset Hsubprog Hsubprog':=
 first [
    eapply jumpimm_partial; try apply Hsubprog;try apply Hsubprog';
     [instrextract start offset | instrextract start' offset | idtac] |
    eapply jumpimm_total; try apply Hsubprog;try apply Hsubprog';
	[instrextract start offset | instrextract start' offset | idtac]].


Lemma sem_jumpind : forall pf pf' p p' start start' m m' Rpre,
  pf start =  Some(i_branch (s_ind m))  ->
  pf' start' = Some(i_branch (s_ind m')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (forall n n' o o' s1 s1' j, Interp (({m,m'} |->{n,n'}) ## ({n,n'} |->{o,o'}) ## Rpre) s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, o) /\
    sem_instr (p' start') s1' start' = Some (s1', o')).
Proof.
  intros until Rpre.
  intros Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  intros until j; intros Hinterp.
  assert (Hstate :s1 m = n /\ s1' m' = n').
  eapply ptsto_stateval; apply Hinterp.
  focus_hyp 1.
  assert (Hstate2 :s1 n = o /\ s1' n' = o').
  eapply ptsto_stateval; apply Hinterp.
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  rewrite Hsubprog_start, Hsubprog_start'; simpl.
  destruct Hstate as (Hstate & Hstate').
  destruct Hstate2 as (Hstate2 & Hstate2').
  rewrite Hstate, Hstate',Hstate2, Hstate2'; intuition.
Qed.

Lemma jumpind_total : forall pf pf' p p' start start' lpost lpost' m m' n n' l l' Rpost Rpre,
  pf start =  Some(i_branch (s_ind m))  ->
  pf' start' = Some(i_branch (s_ind m')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  explicit_judgment ({m,m'} |-> {n,n'} ## {n,n'} |-> {l,l'} ## Rpre) 
                 Rpost p p' l l' lpost lpost' ->
  explicit_judgment ({m,m'} |-> {n,n'} ## {n,n'} |-> {l,l'} ## Rpre) 
                 Rpost p p' start start' lpost lpost'.
Proof.
  intros until Rpre;  intros Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  eapply sem_total2.
  eapply sem_jumpind; try eassumption.
Qed.

Lemma jumpind_partial : forall pf pf' p p' start start' m m' n n' Pinv,
  pf start =  Some(i_branch (s_ind m))  ->
  pf' start' = Some(i_branch (s_ind m')) ->
  stateRelLeq (({m,m'}|->{n,n'}) ## Pinv) ({n,n'} |=> Later(Perp (({m,m'}|->{n,n'}) ## Pinv)))-> 
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment (({m,m'}|->{n,n'}) ## Pinv) p p' start start'.
Proof.
  intros until Pinv; intros Hinstr1 Hinstr1' Hnext Hsubprog Hsubprog'.
  unfold_judgment start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  hreplace Hnext with (Hnext (functionToPfunction s) (functionToPfunction s') p p' (S j) Hpre).
  simpl in Hnext.
  destruct Hnext as (l & l' & Hl & Hl' & Hnext).
  inversion Hl;inversion Hl'.
  assert (Hj : j < S j); try omega.
  assert (Hj2 : j <= j); try omega.
  hreplace Hnext with (Hnext j Hj s s' j Hj2).
  destruct Hnext as (Hnext,Hnext').
  interp_cond.
  subst.
  
  assert (Hstate :s m = n /\ s' m' = n').
  eapply ptsto_stateval; apply Hpre.
  destruct Hstate as (Hm,Hm').

  conclude using pretac (rewrite Hm in * ) (rewrite Hm' in * ) posttac (rewrite Hm') (rewrite Hm) with Hm Hm' k0 Hnext Hnext' Hsubprog_start Hsubprog_start'. 
Qed.

Ltac instr_jumpind start start' offset Hsubprog Hsubprog':=
 first [
    eapply jumpind_partial; try apply Hsubprog;try apply Hsubprog';
     [instrextract start offset | instrextract start' offset | idtac] |
    eapply jumpind_total; try apply Hsubprog;try apply Hsubprog';
	[instrextract start offset | instrextract start' offset]].

Lemma jumpindo : forall pf pf' p p' start start' m m' n n' cst Pinv,
  pf start =  Some(i_branch (s_indo cst m))  ->
  pf' start' = Some(i_branch (s_indo cst m')) ->
  stateRelLeq (({m,m'}|->{n,n'}) ## Pinv) ({n +cst,n'+cst} |=> Later(Perp (({m,m'}|->{n,n'}) ## Pinv)))-> 
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment (({m,m'}|->{n,n'}) ## Pinv) p p' start start'.
Proof.
  intros until Pinv; intros Hinstr1 Hinstr1' Hnext Hsubprog Hsubprog'.
  unfold_judgment start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  hreplace Hnext with (Hnext (functionToPfunction s) (functionToPfunction s') p p' (S j) Hpre).
  simpl in Hnext.
  destruct Hnext as (l & l' & Hl & Hl' & Hnext).
  inversion Hl;inversion Hl'.
  assert (Hj : j < S j); try omega.
  assert (Hj2 : j <= j); try omega.
  hreplace Hnext with (Hnext j Hj s s' j Hj2).
  destruct Hnext as (Hnext,Hnext').
  interp_cond.
  subst.
  
  assert (Hstate :s m = n /\ s' m' = n').
  eapply ptsto_stateval; apply Hpre.
  destruct Hstate as (Hm,Hm').
  
  conclude using pretac (rewrite Hm in * ) (rewrite Hm' in * ) posttac (rewrite Hm') (rewrite Hm) with Hm Hm' k0 Hnext Hnext' Hsubprog_start Hsubprog_start'. 
Qed.

Ltac instr_jumpindo start start' offset Hsubprog Hsubprog':=
	eapply jumpindo; try apply Hsubprog; try apply Hsubprog';
	[instrextract start offset |instrextract start' offset |
         idtac ].

Lemma sem_jumpcst : forall pf pf' p p' start start' m m' Rpre,
  pf start =  Some(i_branch (s_cst m))  ->
  pf' start' = Some(i_branch (s_cst m')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (forall s1 s1' j, Interp Rpre s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, m) /\
    sem_instr (p' start') s1' start' = Some (s1', m')).
Proof.
  intros until Rpre.
  intros Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  intros until j; intros Hinterp.
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  rewrite Hsubprog_start, Hsubprog_start'; simpl; intuition.
Qed.

Lemma jumpcst_total : forall pf pf' p p' start start' lpost lpost' m m' Rpost Rpre,
  pf start =  Some(i_branch (s_cst m))  ->
  pf' start' = Some(i_branch (s_cst m')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  explicit_judgment Rpre Rpost p p' m m' lpost lpost' ->
  explicit_judgment Rpre Rpost p p' start start' lpost lpost'.
Proof.
  intros until Rpre;  intros Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  eapply sem_total2.
  eapply sem_jumpcst; try eassumption.
Qed.

Lemma jumpcst_partial : forall pf pf' p p' start start' m m' Pinv,
  pf start =  Some(i_branch (s_cst m))  ->
  pf' start' = Some(i_branch (s_cst m')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment Pinv p p' m m'  ->
  judgment Pinv p p' start start'.
Proof.
  intros until Pinv; intros Hinstr1 Hinstr1' Hsubprog Hsubprog' Hnext.
  unfold_judgment start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  unfold judgment, derivation, Perp in Hnext; simpl in Hnext;
  unfold perpOf in Hnext; hreplace Hnext with (Hnext k s s' j Hineq).
  destruct Hnext as (Hnext,Hnext').
  interp_cond.
  
  conclude using pretac (idtac) (idtac) posttac (idtac) (idtac) with Hm Hm' k0 Hnext Hnext' Hsubprog_start Hsubprog_start'. 
Qed.

Ltac instr_jumpcst start start' offset Hsubprog Hsubprog':=
  first [eapply jumpcst_partial | eapply jumpcst_total];
  try apply Hsubprog;try apply Hsubprog';
    [instrextract start offset | instrextract start' offset | idtac].

(*  instruction OP *)

Lemma sem_opimm : forall op pf pf' p p' start start' m m' P src1 src2 Rpre w u w' u',
  pf start = Some(i_op op (d_imm m) src1 src2) ->
  pf' start' = Some(i_op op (d_imm m') src1 src2) ->
  (forall s s' p p' k,  Interp (({m,m'}|=>P) ## Rpre) s s' p p'  k -> 
    op (sem_src src1 s) (sem_src src2 s) = op w u /\
    op (sem_src src1 s') (sem_src src2 s') = op w' u') ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (forall s1 s1' j, Interp ({m,m'}|=>P ## Rpre) s1 s1' p p' j ->
    Interp ({m,m'} |-> {op w u, op w' u'} ## Rpre) (update s1 m (op w u)) (update s1' m' (op w' u')) p p' j /\
    sem_instr (p start) s1 start = Some (update s1 m (op w u), start+1) /\
    sem_instr (p' start') s1' start' = Some (update s1' m' (op w' u'), start'+1)).
Proof.
  intros until u';  rewrite_start_succ start start'.
  intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  intros until j; intros Hinterp.
  split;
  set (next_state := update s1 m (op w u)) in *;
  set (next_state' := update s1' m' (op w' u')) in *;
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog';
  hreplace Hinstr2 with (Hinstr2 s1 s1' p p' j Hinterp);
  destruct Hinstr2 as (Hsem & Hsem').

  interp_light. 
  unfold next_state, next_state'; apply indepupdate ; try eassumption.
  subst; eapply natRel_cond; simpl; try eassumption; try omega; try trivial.
  rewrite <- (ptsto_blank_leq m m' P); unfold Interp; eassumption.
  omega.

  simpl; rewrite Hsubprog_start, Hsubprog_start';simpl; rewrite  Hsem, Hsem'; simpl.
  intuition.
Qed.

Lemma opimm_total : forall op pf pf' p p' start start' lpost lpost' m m' P src1 src2 Rpre Rpost w u w' u',
  pf start = Some(i_op op (d_imm m) src1 src2) ->
  pf' start' = Some(i_op op (d_imm m') src1 src2) ->
  (forall s s' p p' k,  Interp (({m,m'}|=>P) ## Rpre) s s' p p'  k -> 
    op (sem_src src1 s) (sem_src src2 s) = op w u /\
    op (sem_src src1 s') (sem_src src2 s') = op w' u') ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  explicit_judgment ({m,m'} |-> {op w u, op w' u'} ## Rpre) Rpost p p' (start+1) (start'+1) lpost lpost' -> 
  explicit_judgment ({m,m'}|=> P ## Rpre) Rpost p p' start start' lpost lpost'.
Proof.
  intros until u'; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  apply sem_total with (m:=m) (m':=m') (l:=op w u) (l':= op w' u').
  eapply sem_opimm; try eassumption.
Qed.

Lemma opimm_partial : forall op pf pf' p p' start start' m m' P src1 src2 Pinv w u w' u',
  pf start = Some(i_op op (d_imm m) src1 src2) ->
  pf' start' = Some(i_op op (d_imm m') src1 src2) ->
  (forall s s' p p' k,  Interp (({m,m'}|=>P) ## Pinv) s s' p p'  k -> 
    op (sem_src src1 s) (sem_src src2 s) = op w u /\
    op (sem_src src1 s') (sem_src src2 s') = op w' u') ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment (({m,m'} |-> {op w u, op w' u'}) ##  Pinv) p p' (start+1) (start'+1) ->
  judgment (({m,m'} |=>P) ##  Pinv) p p' start start'.
Proof.
  intros until u'; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  apply sem_partial with (m:=m) (m':=m') (l:=op w u) (l':=op w' u').
  eapply sem_opimm; try eassumption.
Qed.

Ltac destruct_opimm op :=
  intros until k; intro Hpost1; simpl;
  clear - Hpost1;
  try first [
    simple eapply (ptsto_statevalOp op); simple apply Hpost1 |
    simple eapply (ptsto_statevalOp op); focus_hyp 1; simple apply Hpost1 |
    simple eapply (ptsto_statevalOp op); focus_hyp 2; focus_hyp 2; simple apply Hpost1 ].

Ltac instr_opimm op start start' offset Hsubprog Hsubprog' m m' n n' :=
	first[eapply (opimm_partial op) with (w:=m) (u:=n) (w':=m') (u':=n')|
          eapply (opimm_total op) with (w:=m) (u:=n) (w':=m') (u':=n')]; 
          try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset | 
         destruct_opimm op |
	 rewrite_start start start' offset ].

Ltac destruct_opimmleft op :=
  intros until k; intro Hpost1; simpl;clear - Hpost1;
  try first [
    eapply (ptsto_statevalOpLeft op); apply Hpost1 | 
    eapply (ptsto_statevalOpLeft op); focus_hyp 1; apply Hpost1] .

Ltac instr_opimmleft op start start' offset Hsubprog Hsubprog' m m' n n' :=
	first[eapply (opimm_partial op) with (w:=m) (u:=n) (w':=m') (u':=n')|
          eapply (opimm_total op) with (w:=m) (u:=n) (w':=m') (u':=n')]; 
          try apply Hsubprog ;try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset |
         destruct_opimmleft op |
	 rewrite_start start start' offset ].

Ltac destruct_opimmind op :=
  intros until k; intro Hpost1; simpl;
    clear - Hpost1; try (simple eapply (ptsto_statevalOp op); focus_hyp 2; focus_hyp 2; simple apply Hpost1).

Ltac instr_opimmind op start start' offset Hsubprog Hsubprog' m m' n n':=
	first[eapply (opimm_partial op) with (w:=m) (u:=n) (w':=m') (u':=n')|
          eapply (opimm_total op) with (w:=m) (u:=n) (w':=m') (u':=n')]; 
          try apply Hsubprog ;try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset | 
         destruct_opimmind op |
	 rewrite_start start start' offset ].

Lemma sem_opind : forall op pf pf' p p' start start' a a' b b' P src1 src2 Rpre w u w' u',
  pf start =  Some(i_op op (d_ind a) src1 src2) ->
  pf' start' =  Some(i_op op (d_ind a') src1 src2) ->
  (forall s s' p p' k,  Interp (({a,a'}|->{b,b'}) ## {b,b'}|=>P ## Rpre) s s' p p'  k -> 
    op (sem_src src1 s) (sem_src src2 s) = op w u /\
    op (sem_src src1 s') (sem_src src2 s') = op w' u') ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (forall s1 s1' j, Interp ({a,a'}|->{b,b'} ## {b,b'}|=>P ## Rpre) s1 s1' p p' j ->
    Interp ({a,a'}|->{b,b'} ## {b,b'} |-> {op w u, op w' u'} ## Rpre) (update s1 b (op w u)) (update s1' b' (op w' u')) p p' j /\
    sem_instr (p start) s1 start = Some (update s1 b (op w u), start+1) /\
    sem_instr (p' start') s1' start' = Some (update s1' b' (op w' u'), start'+1)).
Proof.
  intros until u';  rewrite_start_succ start start'.
  intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  intros until j; intros Hinterp.
  assert (Hstate :s1 a = b /\ s1' a' = b').
  eapply ptsto_stateval; apply Hinterp.
  destruct Hstate as (Hstate & Hstate').
  set (next_state := update s1 b (op w u)) in *;
  set (next_state' := update s1' b' (op w' u')) in *;
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog';
  hreplace Hinstr2 with (Hinstr2 s1 s1' p p' j Hinterp);
  destruct Hinstr2 as (Hsem & Hsem').
  split.

 setoid_rewrite stateRelTensor_assoc; rewrite (stateRelTensor_comm ({a,a'} |-> {b,b'})); normal_form.
  unfold next_state, next_state'; apply indepupdate ; try eassumption.
  subst; eapply natRel_cond; simpl; try eassumption; try omega; try trivial.
  rewrite <- (ptsto_blank_leq b b' P).
  setoid_rewrite stateRelTensor_assoc; rewrite (stateRelTensor_comm ({b,b'} |=> P)); normal_form.
  assumption.

  simpl; rewrite Hsubprog_start, Hsubprog_start';simpl; rewrite  Hstate, Hstate'.
  simpl; rewrite  Hsem, Hsem'; simpl.
  intuition.
Qed.

Lemma opind_total : forall op pf pf' p p' start start' lpost lpost' Rpost a a' b b' P src1 src2 Rpre w u w' u',
  pf start =  Some(i_op op (d_ind a) src1 src2) ->
  pf' start' =  Some(i_op op (d_ind a') src1 src2) ->
  (forall s s' p p' k,  Interp (({a,a'}|->{b,b'}) ## {b,b'}|=>P ## Rpre) s s' p p'  k -> 
    op (sem_src src1 s) (sem_src src2 s) = op w u /\
    op (sem_src src1 s') (sem_src src2 s') = op w' u') ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  explicit_judgment (({a,a'}|->{b,b'}) ## {b,b'} |-> {op w u, op w' u'} ## Rpre) Rpost p p' (start+1) (start'+1) lpost lpost' -> 
  explicit_judgment (({a,a'}|->{b,b'}) ## {b,b'}|=>P ## Rpre) Rpost p p' start start' lpost lpost'.
Proof.
  intros until u'; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  apply sem_total with (m:=b) (m':=b') (l:=op w u) (l':= op w' u').
  eapply sem_opind; try eassumption.
Qed.


Lemma opind_partial : forall (op:nat->nat->nat) pf pf' p p' start start' a a' b b' P src1 src2 Pinv w u w' u',
  pf start =  Some(i_op op (d_ind a) src1 src2) ->
  pf' start' =  Some(i_op op (d_ind a') src1 src2) ->
  (forall s s' p p' k,  Interp (({a,a'}|->{b,b'}) ## ({b,b'}|=>P) ## Pinv) s s' p p'  k -> 
    op (sem_src src1 s) (sem_src src2 s) = op w u /\
    op (sem_src src1 s') (sem_src src2 s') = op w' u') ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment (({a,a'}|->{b,b'}) ## ({b,b'} |-> {op w u, op w' u'}) ##  Pinv) p p' (start+1) (start'+1) ->
  judgment (({a,a'}|->{b,b'}) ## ({b,b'} |=>P) ##  Pinv) p p' start start'.
Proof.
  intros until u'; intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog'.
  apply sem_partial with (m:=b) (m':=b') (l:=op w u) (l':=op w' u').
  eapply sem_opind; try eassumption.
Qed.

Ltac destruct_opind op :=
  intros until k; intro Hpost1; simpl;
  clear - Hpost1; try (simple eapply (ptsto_statevalOp op); focus_hyp 3; focus_hyp 3; simple apply Hpost1).

Ltac instr_opind op start start' offset Hsubprog Hsubprog' m m' n n':=
	first[ eapply (opind_partial op) with (w:=m) (u:=n) (w':=m') (u':=n')|
	eapply (opind_total op) with (w:=m) (u:=n) (w':=m') (u':=n')]; 
          try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset | instrextract start' offset | 
         destruct_opind op |
	 rewrite_start start start' offset ].

Lemma opindo : forall (op:nat->nat->nat) pf pf' p p' start start' cst a a' b b' P src1 src2 Pinv w u w' u',
  pf start =  Some(i_op op (d_indo cst a) src1 src2) ->
  pf' start' =  Some(i_op op (d_indo cst a') src1 src2) ->
  (forall s s' p p' k,  Interp (({a,a'}|->{b,b'}) ## ({b+cst,b'+cst}|=>P) ## Pinv) s s' p p'  k -> 
    op (sem_src src1 s) (sem_src src2 s) = op w u /\
    op (sem_src src1 s') (sem_src src2 s') = op w' u') ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  judgment (({a,a'}|->{b,b'}) ## ({b+cst,b'+cst} |-> {op w u, op w' u'}) ##  Pinv) p p' (start+1) (start'+1) ->
  judgment (({a,a'}|->{b,b'}) ## ({b+cst,b'+cst} |=>P) ##  Pinv) p p' start start'.
Proof.
  intros until u'; rewrite_start_succ start start'.
  intros Hinstr1 Hinstr1' Hinstr2 Hsubprog Hsubprog' Hnext.
  unfold_judgment start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  hreplace Hinstr2 with (Hinstr2 s s' p p' (S j) Hpre);
  destruct Hinstr2 as (Hs,Hs'); subst.  
  assert (Hstate :s a = b /\ s' a' = b').
  eapply ptsto_stateval; apply Hpre.
  destruct Hstate as (Hb,Hb'); rewrite Hb,Hb'.

  pose (next_state := update s (b+cst) (op w u)).
  pose (next_state' := update s' (b'+cst) (op w' u')).
  unfold judgment, derivation, Perp in Hnext; simpl in Hnext;
  unfold perpOf in Hnext; hreplace Hnext with (Hnext k next_state next_state' j Hineq).
  destruct Hnext as (Hnext, Hnext').
  setoid_rewrite stateRelTensor_assoc; rewrite (stateRelTensor_comm ({a,a'}|->{b,b'})); normal_form.
  unfold next_state, next_state'. 
  apply indepupdate ; try eassumption.
  eapply natRel_cond; [ simpl;intuition | eassumption].
  setoid_rewrite <- (ptsto_blank_leq (b+cst) (b'+cst) P).
  setoid_rewrite stateRelTensor_assoc; rewrite (stateRelTensor_comm ({b+cst,b'+cst} |=> P)); normal_form.
  interp_cond.
  
  conclude using pretac (rewrite Hs in Hterm) (rewrite Hs' in Hterm) posttac (rewrite Hs',Hb') (rewrite Hs,Hb) with Hm Hm' k0 Hnext Hnext' Hsubprog_start Hsubprog_start'.
Qed.

Ltac destruct_opindo op a b a' b':=
  intros s s'; intros until k; intro Hpost1; simpl sem_src; clear - Hpost1;
  first [
  simpl; split ; split ; reflexivity | 
  let foo := fresh in assert (foo : s a = b /\ s' a' = b');
    [simple eapply ptsto_stateval; simple apply Hpost1 | idtac];
    destruct foo as (foo,foo'); rewrite foo, foo';
  try first [
    simple eapply ptstointerp; focus_hyp 1; simple apply Hpost1 |
    simple eapply ptstointerp; focus_hyp 2; simple apply Hpost1 |
    simple eapply ptstointerp; focus_hyp 3; simple apply Hpost1 |
    simple eapply ptstointerp_cloud; focus_hyp 1; simple apply Hpost1 |
    simple eapply ptstointerp_cloud; focus_hyp 2; simple apply Hpost1 |
    simple eapply ptstointerp_cloud; focus_hyp 3; simple apply Hpost1 |
    simple eapply ptstointerp_stack; focus_hyp 1; simple apply Hpost1 |
    simple eapply ptstointerp_stack; focus_hyp 2; simple apply Hpost1 |
    simple eapply ptstointerp_stack; focus_hyp 3; simple apply Hpost1]].

Tactic Notation (at level 0) "instr_opindo" constr(op)
constr(start) constr(start') constr(offset) constr(Hsubprog) constr(Hsubprog') 
constr(m) constr(m') constr(n) constr(n') "on" "values" constr(a) constr(a') constr(b) constr(b'):=
	eapply (opindo op) with (w:=m) (u:=n) (w':=m') (u':=n'); 
          try apply Hsubprog ;try apply Hsubprog';
	    [instrextract start offset | instrextract start' offset | 
              (destruct_opindo op a a' b b')  |
	        rewrite_start start start' offset ].

Lemma sem_brzimm : forall pf pf' p p' start start' m m' l l' n Rpre,
  pf start =  Some( i_brz (s_imm m) (s_cst l)) ->
  pf' start' = Some( i_brz (s_imm m') (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  ((n=0 -> forall s1 s1' j, Interp ({m,m'} |->{n,n} ## Rpre) s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, l) /\
    sem_instr (p' start') s1' start' = Some (s1', l')) /\
   (n>0 -> forall s1 s1' j, Interp ({m,m'} |->{n,n} ## Rpre) s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, start+1) /\
    sem_instr (p' start') s1' start' = Some (s1', start'+1))).
Proof.
  intros until Rpre.
  intros Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  destruct n; split; intro Hn; try solve [inversion Hn].
  
  intros until j; intros Hinterp.
  assert (Hstate :s1 m = 0 /\ s1' m' = 0).
  eapply ptsto_stateval; apply Hinterp.
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  rewrite Hsubprog_start, Hsubprog_start'; simpl.
  destruct Hstate as (Hstate & Hstate').
  rewrite Hstate, Hstate'; intuition.

  intros until j; intros Hinterp.
  assert (Hstate :s1 m = S n /\ s1' m' = S n).
  eapply ptsto_stateval; apply Hinterp.
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  rewrite Hsubprog_start, Hsubprog_start'; simpl.
  destruct Hstate as (Hstate & Hstate').
  rewrite Hstate, Hstate'.
  omega_replace (start+1) with (S start) ;
  omega_replace (start'+1) with (S start') ; intuition.
Qed.

Lemma brzimm_total : forall pf pf' p p' start start' lpost lpost' m m' l l' n Rpost Rpre,
  pf start =  Some( i_brz (s_imm m) (s_cst l)) ->
  pf' start' = Some( i_brz (s_imm m') (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (n=0 -> explicit_judgment ({m,m'} |-> {n,n} ## Rpre) 
    Rpost p p' l l' lpost lpost') ->
  (n>0 -> explicit_judgment ({m,m'} |-> {n,n} ## Rpre) 
    Rpost p p' (start+1) (start'+1) lpost lpost') -> 
  explicit_judgment ({m,m'} |-> {n,n} ## Rpre) 
   Rpost p p' start start' lpost lpost'.
Proof.
  intros until Rpre;  intros Hinstr1 Hinstr1' Hsubprog Hsubprog' H0 Hsucc.
  assert (hyp := sem_brzimm pf pf' p p' start start' m m' l l' n Rpre Hinstr1 Hinstr1' Hsubprog Hsubprog').
  destruct n; intro Hn; try solve [inversion Hn].
  destruct hyp as (hyp & Habsurd).
  specialize (H0 (refl_equal 0)).
  specialize (hyp (refl_equal 0)).
  eapply sem_total2; eassumption.
  destruct hyp as (Habsurd & hyp).
  specialize (hyp (gt_Sn_O n)).
  specialize (Hsucc (gt_Sn_O n)).
  eapply sem_total2; eassumption.
Qed.

(*  instruction BRZ *)

Lemma brzimm_partial : forall pf pf' p p' start start' m m' l l' n Rpre,
  pf start =  Some( i_brz (s_imm m) (s_cst l)) ->
  pf' start' = Some( i_brz (s_imm m') (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (n=0 -> judgment (({m,m'} |-> {n,n}) ## Rpre) p p' l l') ->
  (n> 0 -> judgment (({m,m'}|-> {n,n}) ## Rpre) p p' (start+1) (start'+1)) ->
  judgment (({m,m'} |-> {n,n}) ##  Rpre) p p' start start'.
Proof.
  intros until Rpre;  intros Hinstr1 Hinstr1' Hsubprog Hsubprog' hyp0 hypsucc.
  assert (hyp := sem_brzimm pf pf' p p' start start' m m' l l' n Rpre Hinstr1 Hinstr1' Hsubprog Hsubprog').
  destruct n. 
  destruct hyp as (hyp & Habsurd).
  specialize (hyp (refl_equal 0)).
  specialize (hyp0 (refl_equal 0)).
  eapply sem_partial2; eassumption.

  destruct hyp as (Habsurd & hyp).
  specialize (hyp (gt_Sn_O n)).
  specialize (hypsucc (gt_Sn_O n)).
  eapply sem_partial2; eassumption.
Qed.

Tactic Notation (at level 0) "instr_brzimm" constr(start) constr(start') constr(offset) 
constr(Hsubprog) constr(Hsubprog') :=
	first [eapply brzimm_partial | eapply brzimm_total];
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset |instrextract start' offset | idtac |
	 rewrite_start start start' offset ].

Lemma sem_brzind : forall pf pf' p p' start start' a a' b b' l l' n Rpre,
  pf start =  Some( i_brz (s_ind a) (s_cst l)) ->
  pf' start' = Some( i_brz (s_ind a') (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  ((n=0 -> forall s1 s1' j, Interp ({a,a'}|-> {b,b'} ## {b,b'}|->{n,n} ## Rpre) s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, l) /\
    sem_instr (p' start') s1' start' = Some (s1', l')) /\
   (n>0 -> forall s1 s1' j, Interp ({a,a'}|-> {b,b'} ## {b,b'}|->{n,n} ## Rpre) s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, start+1) /\
    sem_instr (p' start') s1' start' = Some (s1', start'+1))).
Proof.
  intros until Rpre.
  intros Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  destruct n; split; intro Hn; try solve [inversion Hn];
  intros until j; intros Hinterp;
  assert (Hstate :s1 a = b /\ s1' a' = b');
  try (eapply ptsto_stateval; apply Hinterp);
  destruct Hstate as (Hstate & Hstate');
  focus_hyp 1.

  assert (HstateB :s1 b = 0 /\ s1' b' = 0).
  eapply ptsto_stateval; apply Hinterp.
  destruct HstateB as (HstateB & HstateB').
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  rewrite Hsubprog_start, Hsubprog_start'; simpl.
  rewrite Hstate, Hstate'; simpl.
  rewrite HstateB, HstateB'; intuition.

  assert (HstateB :s1 b = S n /\ s1' b' = S n).
  eapply ptsto_stateval; apply Hinterp.
  destruct HstateB as (HstateB & HstateB').
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  rewrite Hsubprog_start, Hsubprog_start'; simpl.
  rewrite Hstate, Hstate'; simpl.
  rewrite HstateB, HstateB'; omega_replace (start+1) with (S start) ; 
  omega_replace (start'+1) with (S start') ; intuition.
Qed.


Lemma brzind_total : forall pf pf' p p' start start' lpost lpost' a a' b b' l l' n Rpost Rpre,
  pf start =  Some( i_brz (s_ind a) (s_cst l)) ->
  pf' start' = Some( i_brz (s_ind a') (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (n=0 -> explicit_judgment ({a,a'}|-> {b,b'} ## {b,b'}|->{n,n} ## Rpre) 
       Rpost p p' l l' lpost lpost') ->
  (n>0 -> explicit_judgment ({a,a'}|-> {b,b'} ## {b,b'}|->{n,n} ## Rpre) 
       Rpost p p' (start+1) (start'+1) lpost lpost') ->
  explicit_judgment ({a,a'}|-> {b,b'} ## {b,b'}|->{n,n} ## Rpre) 
       Rpost p p' start start' lpost lpost'.
Proof.
  intros until Rpre;  intros Hinstr1 Hinstr1' Hsubprog Hsubprog' H0 Hsucc.
  assert (hyp := sem_brzind pf pf' p p' start start' a a' b b' l l' n Rpre Hinstr1 Hinstr1' Hsubprog Hsubprog').
  destruct n.
  destruct hyp as (hyp & Habsurd).
  specialize (hyp (refl_equal 0)).
  specialize (H0 (refl_equal 0)).
  eapply sem_total2; eassumption.

  destruct hyp as (Habsurd & hyp).
  specialize (hyp (gt_Sn_O n)).
  specialize (Hsucc (gt_Sn_O n)).
  eapply sem_total2; eassumption.
Qed.

Lemma brzind_partial : forall pf pf' p p' start start' a a' b b' l l' n Rpre,
  pf start =  Some( i_brz (s_ind a) (s_cst l)) ->
  pf' start' = Some( i_brz (s_ind a') (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (n=0  -> judgment (({a,a'}|-> {b,b'}) ## ({b,b'}|->{n,n}) ## Rpre) p p' l l') ->
  (n> 0 -> judgment (({a,a'}|-> {b,b'}) ## ({b,b'}|->{n,n}) ## Rpre) p p' (start+1) (start'+1)) ->
  judgment (({a,a'}|-> {b,b'}) ## ({b,b'}|->{n,n}) ##  Rpre) p p' start start'.
Proof.
  intros until Rpre;  intros Hinstr1 Hinstr1' Hsubprog Hsubprog' hyp0 hypsucc.
  assert (hyp := sem_brzind pf pf' p p' start start' a a' b b' l l' n Rpre Hinstr1 Hinstr1' Hsubprog Hsubprog').
  destruct n. 
  destruct hyp as (hyp & Habsurd).
  specialize (hyp (refl_equal 0)).
  specialize (hyp0 (refl_equal 0)).
  eapply sem_partial2; eassumption.

  destruct hyp as (Habsurd & hyp).
  specialize (hyp (gt_Sn_O n)).
  specialize (hypsucc (gt_Sn_O n)).
  eapply sem_partial2; eassumption.
Qed.

Tactic Notation (at level 0) "instr_brzind" constr(start) constr(start') constr(offset) 
constr(Hsubprog) constr(Hsubprog') :=
	first [eapply brzind_partial | eapply brzind_total] ;
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset |instrextract start' offset | idtac |
	 rewrite_start start start' offset ].

Lemma boolNatZero : forall n , (false = bool_of_nat n) -> (n=0).
Proof.
  induction n.
  constructor.
  simpl; intros.
  inversion H.
Qed.

Lemma boolNatNonZero : forall n , (true = bool_of_nat n) -> (exists n', n=S n').
Proof.
  induction n.
  simpl; intros.
  inversion H.
  simpl ; intros.
  exists n; reflexivity.
Qed.

Lemma brzindo : forall pf pf' p p' start start' a a' b b' cst l l' n n' Rpre,
  pf start =  Some( i_brz (s_indo cst a) (s_cst l)) ->
  pf' start' = Some( i_brz (s_indo cst a') (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (bool_of_nat n = bool_of_nat n') ->
  ((bool_of_nat n = false)  -> judgment (({a,a'}|-> {b,b'}) ## ({b+cst,b'+cst}|->{n,n'}) ## Rpre) p p' l l') ->
  ((bool_of_nat n = true) -> judgment (({a,a'}|-> {b,b'}) ## ({b+cst,b'+cst}|->{n,n'}) ## Rpre) p p' (start+1) (start'+1)) ->
  judgment (({a,a'}|-> {b,b'}) ## ({b+cst,b'+cst}|->{n,n'}) ##  Rpre) p p' start start'.
Proof.
  intros until Rpre; intros Hinstr1 Hinstr1' Hsubprog Hsubprog' Hnn' Hnext_eq Hnext_neq.
  unfold_judgment start start' Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  assert (H0:s a = b /\ s' a' = b' ); [ eapply ptsto_stateval ; apply Hpre| idtac].
  destruct H0 as (Ha,Ha').
  assert (H1:s(s a+cst) = n /\ s'(s' a'+cst) = n');
  [focus_hyp 1;rewrite Ha,Ha'; eapply ptsto_stateval ;apply Hpre | idtac].
  destruct H1 as (Hb,Hb'); rewrite Ha,Ha' in *.
  destruct n.
  hreplace Hnext_eq with (Hnext_eq (refl_equal false)).
  unfold judgment, derivation,Perp in Hnext_eq; simpl in Hnext_eq.
  unfold perpOf in Hnext_eq.
  hreplace Hnext_eq with  (Hnext_eq k s s' j Hineq).
  destruct Hnext_eq as (Hnext,Hnext').
  interp_cond.
  simpl in Hnn'.
  assert (Hn' := boolNatZero n' Hnn').
  rewrite Hn' in *.

  conclude using pretac (rewrite Hb in Hterm) (rewrite Hb' in Hterm) posttac (rewrite Ha',Hb') (rewrite Ha,Hb) with Hm Hm' k0 Hnext Hnext' Hsubprog_start Hsubprog_start'. 

  hreplace Hnext_neq with (Hnext_neq (refl_equal true)).
  unfold judgment, derivation,Perp in Hnext_neq; simpl in Hnext_neq.
  unfold perpOf in Hnext_neq.
  hreplace Hnext_neq with (Hnext_neq k s s' j Hineq).
  destruct Hnext_neq as (Hnext,Hnext').
  interp_cond.
  simpl in Hnn'.
  assert (Hn' := boolNatNonZero n' Hnn').
  destruct Hn' as [n0' Hn'].
  rewrite Hn' in *.
  replace (start+1) with (S start) in * by omega.
  replace (start'+1) with (S start') in * by omega.
  conclude using pretac (rewrite Hb in Hterm) (rewrite Hb' in Hterm) posttac (rewrite Ha',Hb') (rewrite Ha,Hb)  with Hm Hm' k0 Hnext Hnext' Hsubprog_start Hsubprog_start'. 
Qed.


Tactic Notation (at level 0) "instr_brzindo" constr(start) constr(start') constr(offset) 
constr(Hsubprog) constr(Hsubprog') constr(Hbooleq) :=
	eapply brzindo; try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset |instrextract start' offset | apply Hbooleq | idtac |
	 rewrite_start start start' offset ].

(*  instruction EQTST *)

Lemma sem_eqtst : forall pf pf' p p' start start' a1' a1 b1 a2 a2' b2 l l' aux aux' Rpre,
  pf start = Some(i_eqtst (s_ind a1) (s_imm a2 ) (s_cst l)) ->
  pf' start' = Some(i_eqtst (s_ind a1') (s_imm a2' ) (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  ((b1=b2 -> forall s1 s1' j, Interp ({a1,a1'}|-> {aux,aux'} ## {aux,aux'}|->{b1,b1} ## {a2,a2'} |-> {b2,b2} ## Rpre) s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, l) /\
    sem_instr (p' start') s1' start' = Some (s1', l')) /\
   (b1<>b2 -> forall s1 s1' j, Interp ({a1,a1'}|-> {aux,aux'} ## {aux,aux'}|->{b1,b1} ## {a2,a2'} |-> {b2,b2} ## Rpre) s1 s1' p p' j ->
    sem_instr (p start) s1 start = Some (s1, start+1) /\
    sem_instr (p' start') s1' start' = Some (s1', start'+1))).
Proof.
  intros until Rpre.
  intros Hinstr1 Hinstr1' Hsubprog Hsubprog'.
  compare b1 b2; split; intro Hn;
    [idtac | absurd (b1=b2); assumption | absurd (b1=b2); assumption | idtac];
  intros until j; intros Hinterp;
  assert (HstateA :s1 a1 = aux /\ s1' a1' = aux');
  try (eapply ptsto_stateval; apply Hinterp);
  destruct HstateA as (HstateA & HstateA');
  focus_hyp 1;
  assert (HstateB :s1 aux = b1 /\ s1' aux' = b1);
  try (eapply ptsto_stateval; apply Hinterp);
  destruct HstateB as (HstateB & HstateB');
  focus_hyp 2;
  assert (HstateB2 :s1 a2 = b2 /\ s1' a2' = b2);
  try (eapply ptsto_stateval; apply Hinterp);
  destruct HstateB2 as (HstateB2 & HstateB2');
  unfold_sem start start' Hinstr1 Hinstr1' Hsubprog Hsubprog';
  rewrite Hsubprog_start, Hsubprog_start'; simpl;
  rewrite HstateA, HstateA'; simpl;
  rewrite HstateB, HstateB'; simpl;
  rewrite HstateB2, HstateB2'; simpl.

  assert (Heq : true = beq_nat b1 b2 ) ; [rewrite Hn;apply beq_nat_refl | idtac].
  rewrite <- Heq.
  intuition.

  assert (Hneq : false = beq_nat b1 b2 ); [apply beq_nat_neq;omega | idtac]. 
  rewrite <- Hneq.
  omega_replace (start+1) with (S start) ; omega_replace (start'+1)
  with (S start') ; intuition.
Qed.
  
Lemma eqtst_total : forall pf pf' p p' start start' lpost lpost' a1' a1 b1 a2 a2' b2 l l' aux aux' Rpost Rpre,
  pf start = Some(i_eqtst (s_ind a1) (s_imm a2 ) (s_cst l)) ->
  pf' start' = Some(i_eqtst (s_ind a1') (s_imm a2' ) (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (b1=b2 -> explicit_judgment ({a1,a1'}|-> {aux,aux'} ## {aux,aux'}|->{b1,b1} ## {a2,a2'} |-> {b2,b2} ## Rpre) Rpost p p' l l' lpost lpost') ->
  (b1<>b2 -> explicit_judgment ({a1,a1'}|-> {aux,aux'} ## {aux,aux'}|->{b1,b1} ## {a2,a2'} |-> {b2,b2} ## Rpre) Rpost p p' (start+1) (start'+1) lpost lpost') ->
  explicit_judgment ({a1,a1'}|-> {aux,aux'} ## {aux,aux'}|->{b1,b1} ## {a2,a2'} |-> {b2,b2} ## Rpre) Rpost p p' start start' lpost lpost'.
Proof.
  intros until Rpre;  intros Hinstr1 Hinstr1' Hsubprog Hsubprog' Heq Hneq.
  assert (hyp := sem_eqtst pf pf' p p' start start' a1' a1 b1 a2 a2' b2 l l' aux aux' Rpre Hinstr1 Hinstr1' Hsubprog Hsubprog').
  compare b1 b2; intros Hcomp; try solve [absurd (b1=b2); assumption].
  destruct hyp as (hyp & Habsurd).
  specialize (hyp Hcomp).
  specialize (Heq Hcomp).
  eapply sem_total2; eassumption.
  destruct hyp as (Habsurd & hyp).
  specialize (hyp Hcomp).
  specialize (Hneq Hcomp).
  eapply sem_total2; eassumption.
Qed.

Lemma eqtst_partial : forall pf pf' p p' start start' a1' a1 b1 a2 a2' b2 l l' aux aux' Rpre,
  pf start = Some(i_eqtst (s_ind a1) (s_imm a2 ) (s_cst l)) ->
  pf' start' = Some(i_eqtst (s_ind a1') (s_imm a2' ) (s_cst l')) ->
  PfunctionExtendsFunction p pf  -> 
  PfunctionExtendsFunction p' pf'  -> 
  (b1=b2  -> judgment (({a1,a1'}|-> {aux,aux'}) ## ({aux,aux'}|->{b1,b1}) ## ({a2,a2'} |-> {b2,b2})## Rpre) p p' l l') ->
  (b1<>b2 -> judgment (({a1,a1'}|->{aux,aux'}) ## ({aux,aux'}|->{b1,b1}) ## ({a2,a2'} |-> {b2,b2})## Rpre) p p' (start+1) (start'+1)) ->
  judgment (({a1,a1'}|->{aux,aux'}) ## ({aux,aux'} |-> {b1,b1}) ## ({a2,a2'} |-> {b2,b2}) ##  Rpre) p p' start start'.
Proof.
  intros until Rpre;  intros Hinstr1 Hinstr1' Hsubprog Hsubprog' hyp0 hypsucc.
  assert (hyp := sem_eqtst pf pf' p p' start start' a1' a1 b1 a2 a2' b2 l l' aux aux' Rpre Hinstr1 Hinstr1' Hsubprog Hsubprog').
  compare b1 b2; intro Heq.

  destruct hyp as (hyp & Habsurd);
  specialize (hyp Heq);
  specialize (hyp0 Heq);
  eapply sem_partial2; eassumption.

  destruct hyp as (Habsurd & hyp).
  specialize (hyp Heq);
  specialize (hypsucc Heq);
  eapply sem_partial2; eassumption.
Qed.

Tactic Notation (at level 0) "instr_eqtst" constr(start) constr(start') constr(offset) 
constr(Hsubprog) constr(Hsubprog') :=
	first [eapply eqtst_total | eapply eqtst_partial];
        try apply Hsubprog ; try apply Hsubprog' ;
	[instrextract start offset |instrextract start' offset | idtac |
	 rewrite_start start start' offset ].

