
Require Export programs.
Require Export Relation_Definitions.
Require Export Ring.
Require Export List.

Implicit Arguments nullPfunction [].
Implicit Arguments join_disj_comm [A B].


Definition Astate := nat.
Definition Bstate := nat.

Record statePred : Type := mkStatePred {
  stateof :> (Pfunction Astate Bstate) -> program -> nat -> Prop;
  statePred_cond : forall s1 s2 p k1 k2, 
    stateof s1 p k1 -> 
    subPfunction s1 s2 -> 
    (k2 <= k1) -> 
    stateof s2 p k2}.

Implicit Arguments statePred_cond [].

Definition statePredEq (P Q: statePred) := forall s p k , P s p k <-> Q s p k.

Lemma statePredEq_refl : forall (P : statePred), statePredEq P P.
Proof.
  firstorder.
Qed.

Lemma statePredEq_comm : forall (P Q : statePred), statePredEq P Q -> statePredEq Q P.
Proof.
  firstorder.
Qed.


Lemma statePredEq_trans : forall (P Q R : statePred), statePredEq P Q -> statePredEq Q R -> statePredEq P R.
Proof.
  unfold statePredEq; intros; simpl.
  assert (foo := H0 s p k).
  assert (foo2 := H s p k).
  firstorder.
Qed.


Hint Resolve statePredEq_refl statePredEq_trans.

Add Relation statePred statePredEq
   reflexivity proved by statePredEq_refl
   symmetry proved by statePredEq_comm
   transitivity proved by statePredEq_trans
 as statePredEqRel.

Add Morphism stateof with signature statePredEq ==> (@eq (Pfunction Astate Bstate)) ==> 
(@eq program) ==> (@eq nat) ==> iff  as stateofcompat.
Proof.
  firstorder.
Qed.

Definition statePredLeq (P Q : statePred) := forall s p k , P s p k -> Q s p k.

Lemma statePredLeq_refl : forall P, statePredLeq P P.
Proof.
  firstorder.
Qed.

Lemma statePredLeq_trans : forall P Q R,  statePredLeq P Q -> statePredLeq Q R -> statePredLeq P R.
Proof.
  firstorder.
Qed.

Hint Resolve statePredLeq_refl statePredLeq_trans.

Add Relation statePred statePredLeq
   reflexivity proved by statePredLeq_refl
   transitivity proved by statePredLeq_trans
 as statePredLeqRel.

Add Morphism statePredLeq with signature statePredEq ==> statePredEq ==> iff
as ppredleqcompat.
Proof.
  unfold statePredEq, statePredLeq; intuition.
  assert (foo := H s p k ); assert (foo2 := H0 s p k).
  firstorder.
  assert (foo := H s p k ); assert (foo2 := H1 s p k).
  firstorder.
Qed.

Lemma statePredEq_Leq: forall P Q, statePredEq P Q -> statePredLeq P Q.
Proof.
  intros; unfold statePredEq.
  firstorder.
Qed.

(*-------------*)
(* statePredTensor *)
(*-------------*)

Definition statePredTensor (P1 P2 : statePred) : statePred.
Proof. 
  intros P1 P2.
  refine (mkStatePred (fun f => fun p => fun k => exists f1 : (Pfunction Astate Bstate), exists f2 : (Pfunction Astate Bstate), 
                                         disjointPfunction f1 f2 /\  f = (joinPfunction f1 f2) /\  P1 f1 p k /\ P2 f2 p k)
                _).
  intros.
  witness f3 in H.
  witness f4 in H.
  deconstruct H as [Hdisj34 [Hjoin134 [HP13 HP24]]].
  unfold subPfunction in H0.
  witness fdiff in H0.
  exists (joinPfunction f3 fdiff).
  exists f4.
  rewrite Hjoin134 in H0.
  destruct H0.
  assert (Hdisj := disj_join_left f3 f4 fdiff). 
  split.
  rewrite  (disj_join_left f3 fdiff f4);  split; [intuition | idtac] . 
  rewrite (disjointPfunction_comm fdiff f4); intuition.
  split.
  rewrite  <- (disj_join_assoc); [idtac | intuition | rewrite (disjointPfunction_comm fdiff f4); intuition | intuition].
  rewrite (join_disj_comm fdiff f4); [idtac | rewrite (disjointPfunction_comm fdiff f4); intuition ].
  rewrite  (disj_join_assoc); [idtac | intuition | intuition | intuition].
  assumption.
  split.
  eapply (statePred_cond P1).   
  apply HP13.
  apply Pfunction_ext; intuition.
  assumption.
  apply (statePred_cond P2 f4 f4 p k1 k2); [assumption | apply subPfunction_refl | assumption].
Defined.

Add Morphism statePredTensor with signature statePredEq ==> statePredEq ==> statePredEq
as ppredtensorcompat.
Proof.
  unfold statePredEq, statePredTensor; simpl; firstorder.
  exists x1; exists x2.
  firstorder.
  exists x1; exists x2.
  firstorder.
Qed.

Add Morphism statePredTensor with signature statePredLeq ++> statePredLeq ++> statePredLeq
as ppredtensorcompat2.
Proof.
  unfold statePredLeq; unfold statePredTensor; simpl.
  firstorder.
Qed.

