(* states.v

   Define a state as a function of type nat -> nat
   Define update of a state

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research
*)

Require Export partial_functions.

Implicit Arguments nullPfunction [].

Definition state := nat -> nat.

Definition Pstate := Pfunction nat nat.

Definition nullPstate := nullPfunction nat nat.

Definition update (s:state) (n:nat) (v:nat) : state := 
  fun m => if beq_nat n m then v else s m.

Lemma extends_ineq : forall n (s:Pstate), 
                  extendsPfunction (fun i => if bleq_nat i n then s i else None) s.
Proof.
  unfold extendsPfunction.
  intros until y.
  assert (case :=  le_gt_dec x n).
  destruct case.
  replace (bleq_nat x n) with true;
  [auto | apply bleq_nat_true;  omega].
  replace (bleq_nat x n) with false;
  [intro dummy; inversion dummy | apply bleq_nat_false ;assumption].
Qed.

Lemma extends_eq : forall n (s:Pstate), 
                  extendsPfunction (fun i => if beq_nat i n then s i else None) s.
Proof.
  unfold extendsPfunction.
  intros until y.
  assert (case :=  eq_nat_dec x n).
  destruct case.
  replace (beq_nat x n) with true;
  [auto | rewrite e; apply beq_nat_refl].
  replace (beq_nat x n) with false;
  [intro dummy; inversion dummy |  apply beq_nat_neq;  assumption].
Qed.

Fixpoint dumpstateupto (n:nat) (s:state) {struct n} : list (nat * nat) :=
match n with
 | 0 => (0, s 0) :: nil
 | S m => (dumpstateupto m s) ++ ((S m, s (S m)) :: nil)
end.




