(* testing.v
   This contains a testharness and a couple of simple sanity tests for the code in

   Abstracting Allocation : The New New Thing
   Nick Benton

   (c) 2006 Microsoft Research
*)

Require Import programs.
Require Import allocator_code.
Require Import client_code.

Fixpoint execute_bounded (k:nat) (p:program) (s:state) (l:nat) {struct k} : state * nat :=
match k with
 | 0 => (s,l)
 | S j => match sem_instr (p l) s l with
            | None => (s,l)
            | Some (s',l') => execute_bounded j p s' l'
          end
end.

Definition harnessinstructions (test init fact n : nat) :=
       Prog_add 1 ( i_move (d_imm 0) (s_cst (test+2)))
      (Prog_add 2 (i_branch (s_cst init))
      (Prog_add 3 (i_move (d_imm 5) (s_cst n))
      (Prog_add 4 (i_move (d_imm 6) (s_cst (test+5)))
      (Prog_add 5 (i_branch (s_cst fact))
      (Prog_add 6 (i_halt) empty_prog))))).

Definition testfrag  : program_fragment :=
let init := 0
in let alloc := 10
in let dealloc := 40
in let fact := 50
in let test := 100
in joinPfunction (fragfromprog (factinstructions fact alloc dealloc) fact)
   (joinPfunction (fragfromprog (allocinstructions alloc) alloc)
   (joinPfunction (fragfromprog (initinstructions init) init)
   (joinPfunction (fragfromprog (deallocinstructions dealloc) dealloc)
                            (fragfromprog (harnessinstructions test init fact 5) test)))).

Definition initialstate : state := fun n=>0.

Definition testprogram : program := halting_extension testfrag.

Fixpoint dumpstateupto (n:nat) (s:state) {struct n} : list (nat * nat) :=
match n with
 | 0 => (0, s 0) :: nil
 | S m => (dumpstateupto m s) ++ ((S m, s (S m)) :: nil)
end.

Definition testresult :=
let (finalstate, finalpc) := (execute_bounded 1000 testprogram initialstate 100)
in (dumpstateupto 30 finalstate, finalpc).

Eval compute in testresult.

Definition testfrag1  : program_fragment :=
let init := 0
in let alloc := 10
in let dealloc := 40
in let fact := 50
in let test := 100
in joinPfunction (fragfromprog (factinstructions fact alloc dealloc) fact)
   (joinPfunction (fragfromprog (ntallocinstructions alloc) alloc)
   (joinPfunction (fragfromprog (ntinitinstructions init) init)
   (joinPfunction (fragfromprog (ntdeallocinstructions dealloc) dealloc)
                            (fragfromprog (harnessinstructions test init fact 5) test)))).
   
Definition testprogram1 : program := halting_extension testfrag1.

Definition testresult1 :=
let (finalstate, finalpc) := (execute_bounded 1000 testprogram1 initialstate 100)
in (dumpstateupto 40 finalstate, finalpc).


Eval compute in testresult1.

