(* utility.v

   Generally useful stuff that doesn't belong anywhere else for

   Compiling Functional Types to Relational Specifications for Low Level Imperative Code
   Nick Benton and Nicolas Tabareau

   (c) 2007 Microsoft Research

   This also exports a bunch of pervasive stuff from the standard library
*)

Require Export Arith.
Require Export EqNat.
Require Export Bool.
Require Export BoolEq.
Require Export Omega.
Require Export Bool_nat.
Require Export List.
Require Export Max.
Require Export Min.
Require Export Setoid.

Typeclasses eauto:= dfs 10000. 

Lemma generalise_recursion : forall (P:nat->Prop), (forall k , P k) <-> (forall k k0, k0<=k -> P k0).
Proof.
  intuition; apply (H k k); trivial. 
Qed.

Ltac dcase x := generalize (refl_equal x); pattern x at -1; case x.

Notation "[ x ; .. ; y ]" := (cons x .. (cons y nil) ..).

Tactic Notation "hreplace" constr(h) "with" constr(t) := let id:=fresh in (assert (id := t); clear h; rename id into h).

Tactic Notation "omega_replace" constr(m) "with" constr(n) := replace m with n by omega.

Tactic Notation "omega_replace" constr(m) "with" constr(n) "in" ident(H) := replace m with n in H by omega.


(*Tactic Notation "deconstruct" hyp(h) := _deconstruct h.*)
Tactic Notation "deconstruct" hyp(h) "as" simple_intropattern(ip) := generalize h; clear h; intros ip.

Tactic Notation "witness" ident(p) "in" hyp(h) := elim h; clear h; intros p h.

Axiom Extensionality : forall (A B : Type) (f g : A->B), (forall a:A, f a = g a) <-> f =g.

Lemma strictPosSucc: forall n, n>0 -> exists m , S m =n.
Proof.
  intros.
  induction n.
  destruct H.
  exists n; trivial.
  exists m; trivial.
  exists n; trivial.
Qed.  

Ltac unfold_plus:=
  try repeat rewrite <- plus_n_Sm;
  try rewrite plus_0_r;
  try rewrite plus_0_l.

Tactic Notation (at level 0) "unfold_plus" "in" constr(H) := try repeat rewrite <- plus_n_Sm in H;
  try rewrite plus_0_r in H;
  try rewrite plus_0_l in H.

Lemma beq_nat_neq :forall m n, m<>n ->  false = (beq_nat m n).
     double induction m n; simpl in |-*.
    intros.
    tauto.
    firstorder.
    firstorder.
    intros.
    set(Hn := (H0 n)).
    assert (n1 <> n).
    omega.
    apply (Hn H2).
Qed.

(* bleq_nat, ble_nat ... *)


Fixpoint bleq_nat (n m: nat) {struct n} : bool :=
  match n with 
    | 0 => true
    | S n1 => match m with
		| 0 => false
		| S m1 => bleq_nat n1 m1
	      end
  end.


Fixpoint ble_nat (n m : nat) {struct n} : bool :=
  match n with 
    | 0 => match m with
	     | 0 => false
	     | S _ => true
	    end
    | S n1 => match m with
		| 0 => false
		| S m1 => ble_nat n1 m1
	      end
  end.


Lemma ble_nat_false : forall m n, m >=n -> false = ble_nat m n.
  double induction m n.
  firstorder.
  firstorder.
  firstorder.
  firstorder.
Qed.

Lemma bleq_nat_false : forall m n, m > n -> false = bleq_nat m n.
double induction m n.
 simpl.
 intro; absurd (0>0).
 omega.
 assumption.

 intros.
 simpl.
 absurd (0> S n0).
 omega.
 assumption.

 intros.
 simpl.
 tauto.

 intros.
 simpl.
 apply H0.
 omega.
Qed.

Lemma bleq_nat_true : forall m n, m <= n -> true = bleq_nat m n.
double induction m n; intros; simpl.
tauto.
tauto.
absurd (S n <=0).
 omega.
 assumption.
apply H0.
omega.
Qed.

Lemma true_bleq_nat : forall m n, true = bleq_nat m n -> m <= n.
double induction m n.
simpl.
omega.
simpl; intros.
omega.
simpl; intros.
discriminate H0.
simpl; intros.
intuition.
Qed.

Lemma false_bleq_nat : forall m n, false = bleq_nat m n -> m > n.
intros.
assert (~ m <= n).
intro Habs; generalize (bleq_nat_true m n Habs); intro.
rewrite <- H0 in H; discriminate H.
omega.
Qed.

Lemma ble_nat_true : forall m n, m < n -> true = ble_nat m n.
double induction m n.
intro; absurd (0<0).
omega.
assumption.

intros.
simpl.
tauto.

intros.
simpl.
absurd (S n < 0).
omega.
assumption.

intros.
simpl.
apply H0.
omega.
Qed.

Lemma booleq : forall (a b : bool), ((true = a) <-> (true = b)) -> (a = b).
intros; dcase a.
intuition.
intuition.
dcase b.
rewrite H0 in H2.
intro; symmetry; apply H2; symmetry; assumption.
tauto.
Qed.

Lemma split_conj : forall A B C, (A -> (B/\C)) <-> ((A->B) /\ (A->C)).
Proof.
  intuition.
Qed.

Lemma min_gt : forall m n k, k < m -> k < n -> k < min m n.
Proof.
  double induction m n; simpl in *; intuition.
  destruct k.
  omega.
  apply lt_n_S.
  eapply H0; intuition.
Qed.


